﻿/*
 * MediaPlayer.cs
 * Copyright (c) 2007-2009 kbinani
 *
 * This file is part of Boare.Lib.Media.
 *
 * Boare.Lib.Media is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Lib.Media is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Boare.Lib.Media {

    public class MediaPlayer {

        // 構造体で再現した共用体
        [StructLayout( LayoutKind.Explicit )]
        struct DoubleWord {
            [FieldOffset( 0 )]
            public System.UInt32 Value;

            [FieldOffset( 0 )]
            public System.UInt16 LowWord;
            [FieldOffset( 2 )]
            public System.UInt16 HighWord;
        }

        private string m_filename = "";
        const int FALSE = 0;
        const int TRUE = 1;
        bool pausing = false;
        uint m_volume;
        bool mute = false;
        string m_alias = "";

        StringBuilder sb = new StringBuilder( 32 );  // mciSendString() のReturn String 格納用

        public MediaPlayer() {
            w_waveOutGetVolume( 0, ref m_volume );
            DoubleWord tmp = new DoubleWord();
            tmp.Value = m_volume;
        }

        public void ApplyVolume() {
            if ( !mute ) {
                //mciSendString( "setaudio \"" + filename + "\" volume to " + volume_.ToString(), null, 0, 0 );
                w_waveOutSetVolume( 0, m_volume );
            }
        }


        int m_speed = 1000;
        public int Speed {
            get {
                return m_speed;
            }
            set {
                m_speed = value;
                //w_mciSendString( "set \"" + m_filename + "\" speed " + m_speed, null, 0, 0 );
                w_mciSendString( "set " + m_alias + " speed " + m_speed, null, 0, 0 );
            }
        }


        public int Volume {
            get {
                //mciSendString( "status \"" + filename + "\" volume", sb, sb.Capacity, 0 );
                w_waveOutGetVolume( 0, ref m_volume );
                DoubleWord vol = new DoubleWord();
                vol.Value = m_volume;
                return (vol.HighWord + vol.LowWord) / 2;
            }
            set {
                DoubleWord tmp = new DoubleWord();
                tmp.Value = m_volume;
                int sum = (tmp.HighWord + tmp.LowWord);
                float k = 2.0f * value / sum;
                if ( sum == 0 ) {
                    tmp.HighWord = (ushort)value;
                    tmp.LowWord = (ushort)value;
                } else {
                    tmp.HighWord = (ushort)(k * tmp.HighWord);
                    tmp.LowWord = (ushort)(k * tmp.LowWord);
                }
                m_volume = tmp.Value;
            }
        }


        public void EnterMute() {
            mute = true;
            //mciSendString( "setaudio \"" + filename + "\" volume to " + 0, null, 0, 0 );
            w_waveOutSetVolume( 0, 0 );
        }


        public void ExitMute() {
            mute = false;
            //mciSendString( "setaudio \"" + filename + "\" volume to " + volume_.ToString(), null, 0, 0 );
            w_waveOutSetVolume( 0, m_volume );
        }


        [DllImport( "winmm.dll" )]
        private extern static long waveOutGetVolume( uint hwo, ref uint pdwVolume );


        [DllImport( "winmm.dll" )]
        private extern static long waveOutSetVolume( uint hwo, uint dwVolume );


        private long w_waveOutGetVolume( uint hwo, ref uint pdwVolume ) {
            long io_status = 0;
            try {
                io_status = waveOutGetVolume( hwo, ref pdwVolume );
            } catch {
                return 1;
            }
            return io_status;
        }


        private long w_waveOutSetVolume( uint hwo, uint dwVolume ) {
            long io_status = 0;
            try {
                io_status = waveOutSetVolume( hwo, dwVolume );
            } catch {
                return 1;
            }
            return io_status;
        }


        public string SoundLocation {
            get {
                return m_filename;
            }
        }

        [DllImport( "winmm.dll" )]
        extern static int mciSendString( string s1, StringBuilder s2, int i1, int i2 );
        // 返り値 = 0 @正常終了
        // s1     = Command String
        // s2     = Return String
        // i1     = Return String Size
        // i2     = Callback Hwnd


        /// <summary>
        /// エラー処理付のmciSendStringラッパー
        /// </summary>
        /// <param name="s1"></param>
        /// <param name="s2"></param>
        /// <param name="i1"></param>
        /// <param name="i2"></param>
        /// <returns></returns>
        static int w_mciSendString( string command, StringBuilder s2, int i1, int i2 ) {
            int io_status = 0;
            try {
                io_status = mciSendString( command, s2, i1, i2 );
            } catch {
                return 1;
            }
            if ( io_status == 0 ) {
                return 0;
            } else {
                return 1;
            }
        }


        public int Load( string filename ) {
            if ( this.SoundLocation != "" ) {
                Close();
            }
            this.m_filename = filename;
            m_alias = bocoree.misc.getmd5( m_filename );
            return mciSendString( "open \"" + filename + "\" alias " + m_alias, null, 0, 0 );
            //return w_mciSendString( "open \"" + filename + "\"", null, 0, 0 );
        }


        public int PlayFrom( double time ) {
            long position = (long)(time * 1000);
            if ( mute ) {
                EnterMute();
            } else {
                ExitMute();
            }
            //return w_mciSendString( "play \"" + m_filename + "\" from " + position.ToString(), null, 0, 0 );
            return w_mciSendString( "play " + m_alias + " from " + position.ToString(), null, 0, 0 );
        }


        public int Close() {
            if ( this.SoundLocation == "" ) {
                return FALSE;
            }
            Stop();
            //return w_mciSendString( "close \"" + m_filename + "\"", null, 0, 0 );
            return w_mciSendString( "close " + m_alias, null, 0, 0 );
        }


        public int Play() {
            if ( this.SoundLocation == "" ) {
                return FALSE;
            }
            if ( mute ) {
                EnterMute();
            } else {
                ExitMute();
            }
            if ( pausing ) {
                //return w_mciSendString( "resume \"" + m_filename + "\"", null, 0, 0 );
                return w_mciSendString( "resume " + m_alias, null, 0, 0 );
            } else {
                //return w_mciSendString( "play \"" + m_filename + "\"", null, 0, 0 );
                return w_mciSendString( "play " + m_alias, null, 0, 0 );
            }
        }


        public int Seek( double pos_second ) {
            if ( this.SoundLocation == "" ) {
                return FALSE;
            }
            long position = (long)(pos_second * 1000.0);
            //int ret = w_mciSendString( "seek \"" + m_filename + "\" to " + position, sb, sb.Capacity, 0 );
            int ret = w_mciSendString( "seek " + m_alias + " to " + position, sb, sb.Capacity, 0 );
            return ret;
        }


        public int Pause() {
            if ( this.SoundLocation == "" ) {
                return FALSE;
            }
            pausing = true;
            //return w_mciSendString( "pause \"" + m_filename + "\"", null, 0, 0 );
            return w_mciSendString( "pause " + m_alias, null, 0, 0 );
        }

        /*public int resume() {
            // 一時停止解除
            return mciSendString( "resume " + DEVICE, null, 0, 0 );
        }*/

        public int GetPosition() {
            // play 中のファイルの位置を ms単位 で返す
            if ( this.SoundLocation == "" ) {
                return -1;
            }
            //w_mciSendString( "status \"" + m_filename + "\" position", sb, sb.Capacity, 0 );
            w_mciSendString( "status " + m_alias + " position", sb, sb.Capacity, 0 );
            int pos;
            try {
                pos = int.Parse( sb.ToString() );
            } catch {
                pos = -1;
            }
            return pos;
        }

        public int GetLength() {
            // open 中のファイルの曲長を ms単位 で返す
            if ( this.SoundLocation == "" ) {
                return -1;
            }
            //w_mciSendString( "status \"" + m_filename + "\" length", sb, sb.Capacity, 0 );
            w_mciSendString( "status " + m_alias + " length", sb, sb.Capacity, 0 );
            return int.Parse( sb.ToString() );
        }

        public int Stop() {
            //return w_mciSendString( "stop \"" + m_filename + "\"", null, 0, 0 );
            return w_mciSendString( "stop " + m_alias, null, 0, 0 );
        }
    }

}
