﻿/*
 * AppManager.cs
 * Copyright (c) 2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;
using System.Windows.Forms;
using System.Xml.Serialization;

using Boare.Lib.Vsq;
/*
-ビブラートを、ダイアログを出さずにピアノロールと同じ画面で調整
-アクセント・ディケイパラメータを、ダイアログを出さずにピアノロールと同じ画面で調整
-コントロールカーブのテンプレートの登録＆貼付け機能
-Mac、Linuxでの編集作業を可能にする
-ショートカットキーのカスタマイズ
-選択したノートのプロパティを一括変更する機能
-選択したノートを指定したクロック数分一括シフトする機能
-MIDIキーボード（？用語があってるか不明）を使えるようにする
-前回レンダリングしたWAVEファイルをリサイクルする（起動ごとにレンダリングするのは面倒？）
-横軸を変更するショートカットキー
-音符の時間位置を秒指定で入力できるモード
-コントロールカーブのリアルタイムかつアナログ的な入力（ジョイスティックorゲームコントローラ）
*/

namespace Boare.Cadencii {

    delegate void BSimpleDelegate<T>( T arg );

    partial class AppManager : Form {
        private static int m_base_tempo = 480000;
        /// <summary>
        /// エディタの設定
        /// </summary>
        public static EditorConfig EditorConfig = new EditorConfig();
        /// <summary>
        /// EditorConfigクラス用のシリアライザ
        /// </summary>
        public static XmlSerializer XmlSerializerEditorConfig = new XmlSerializer( typeof( EditorConfig ) );
        /// <summary>
        /// List&lt;BezierCurves&gt;用のシリアライザ
        /// </summary>
        public static XmlSerializer XmlSerializerListBezierCurves = new XmlSerializer( typeof( List<BezierCurves> ) );
        /// <summary>
        /// コピーorカットなどで複製されたエントリ
        /// </summary>
        public static List<VsqEvent> CopiedEvent = null;
        /// <summary>
        /// コピーorカットで複製されたテンポ
        /// </summary>
        public static List<TempoTableEntry> CopiedTempo = null;
        /// <summary>
        /// コピーorカットで複製された拍子
        /// </summary>
        public static List<TimeSigTableEntry> CopiedTimesig = null;
        /// <summary>
        /// コピーorカットで複製されたカーブ
        /// </summary>
        public static List<BPPair> CopiedCurve = null;
        /// <summary>
        /// コピーorカットで複製されたカーブの種類
        /// </summary>
        public static VsqCurveType CopiedCurveType = VsqCurveType.DYN;
        /// <summary>
        /// 複数種類のコントロールカーブのコピー用。
        /// </summary>
        public static Dictionary<VsqCurveType, List<BPPair>> CopiedCurveWhole = null;
        /// <summary>
        /// 複数種類のコントロールカーブがコピーされた状態かどうかを表すフラグ
        /// </summary>
        public static bool IsWholeCurveCopied = false;

        #region Static Readonly Fields
        public static readonly Color[] s_HILIGHT = new Color[] { 
            Color.FromArgb( 181, 220, 16 ),
            Color.FromArgb( 231, 244, 49 ),
            Color.FromArgb( 252, 230, 29 ),
            Color.FromArgb( 247, 171, 20 ),
            Color.FromArgb( 249, 94, 17 ),
            Color.FromArgb( 234, 27, 47 ),
            Color.FromArgb( 175, 20, 80 ),
            Color.FromArgb( 183, 24, 149 ),
            Color.FromArgb( 105, 22, 158 ),
            Color.FromArgb( 22, 36, 163 ),
            Color.FromArgb( 37, 121, 204 ),
            Color.FromArgb( 29, 179, 219 ),
            Color.FromArgb( 24, 239, 239 ),
            Color.FromArgb( 25, 206, 175 ),
            Color.FromArgb( 23, 160, 134 ),
            Color.FromArgb( 79, 181, 21 ) };
        public static readonly Color[] s_RENDER = new Color[]{
            Color.FromArgb( 19, 143, 52 ),
            Color.FromArgb( 158, 154, 18 ),
            Color.FromArgb( 160, 143, 23 ),
            Color.FromArgb( 145, 98, 15 ),
            Color.FromArgb( 142, 52, 12 ),
            Color.FromArgb( 142, 19, 37 ),
            Color.FromArgb( 96, 13, 47 ),
            Color.FromArgb( 117, 17, 98 ),
            Color.FromArgb( 62, 15, 99 ),
            Color.FromArgb( 13, 23, 84 ),
            Color.FromArgb( 25, 84, 132 ),
            Color.FromArgb( 20, 119, 142 ),
            Color.FromArgb( 19, 142, 139 ),
            Color.FromArgb( 17, 122, 102 ),
            Color.FromArgb( 13, 86, 72 ),
            Color.FromArgb( 43, 91, 12 ) };
        #endregion

        #region Private Static Fields
        private static SolidBrush s_hilight_brush = new SolidBrush( Color.CornflowerBlue );
        private static object s_locker;
        #endregion

        public const string CONFIG_PATH = "config.xml";
        public delegate void MainFormClosedEventHandler( EditorManager manager );
        public delegate void EditorConfigChangedEventHandler( EditorManager manager );

        private static List<EditorManager> s_editor_managers = null;
        private object m_locker = new object();

        public AppManager() {
            InitializeComponent();
            s_editor_managers = new List<EditorManager>();
        }

        public static void SaveConfig() {
            EditorConfig.Save( Path.Combine( Application.StartupPath, CONFIG_PATH ) );
        }

        public static string _VERSION {
            get {
                string prefix = "";
#if DEBUG
                prefix = "(build: debug)";
#else
                prefix = "(build: release)";
#endif
                return GetAssemblyFileVersion( typeof( AppManager ) ) + " " + prefix;
            }
        }

        public static string GetAssemblyFileVersion( Type t ) {
            Assembly a = Assembly.GetAssembly( t );
            AssemblyFileVersionAttribute afva = (AssemblyFileVersionAttribute)Attribute.GetCustomAttribute( a, typeof( AssemblyFileVersionAttribute ) );
            return afva.Version;
        }

        public static string GetAssemblyNameAndFileVersion( Type t ) {
            Assembly a = Assembly.GetAssembly( t );
            AssemblyFileVersionAttribute afva = (AssemblyFileVersionAttribute)Attribute.GetCustomAttribute( a, typeof( AssemblyFileVersionAttribute ) );
            return a.GetName().Name + " v" + afva.Version;
        }

        static AppManager() {
            s_locker = new object();

            string config_file = Path.Combine( Application.StartupPath, CONFIG_PATH );
            LoadEditorConfig( config_file );

            KeySoundPlayer.SingerProgramChange = EditorConfig.PreviewSingerProgramChange;
            KeySoundPlayer.Init();

            #region Apply User Dictionary Configuration
            SymbolTable.LoadDictionary();
            List<BKeyValuePair<string, bool>> current = new List<BKeyValuePair<string, bool>>();
            for ( int i = 0; i < SymbolTable.Count; i++ ) {
                current.Add( new BKeyValuePair<string, bool>( SymbolTable.GetSymbolTable( i ).Name, false ) );
            }
            List<BKeyValuePair<string, bool>> config_data = new List<BKeyValuePair<string, bool>>();
            for ( int i = 0; i < EditorConfig.UserDictionaries.Count; i++ ) {
                string[] spl = EditorConfig.UserDictionaries[i].Split( "\t".ToCharArray(), 2 );
                config_data.Add( new BKeyValuePair<string, bool>( spl[0], (spl[1] == "T" ? true : false) ) );
#if DEBUG
                Common.DebugWriteLine( "    " + spl[0] + "," + spl[1] );
#endif
            }
            List<KeyValuePair<string, bool>> common = new List<KeyValuePair<string, bool>>();
            for ( int i = 0; i < config_data.Count; i++ ) {
                for ( int j = 0; j < current.Count; j++ ) {
                    if ( config_data[i].Key == current[j].Key ) {
                        current[j].Value = true; //こっちのboolは、AppManager.EditorConfigのUserDictionariesにもKeyが含まれているかどうかを表すので注意
                        common.Add( new KeyValuePair<string, bool>( config_data[i].Key, config_data[i].Value ) );
                        break;
                    }
                }
            }
            for ( int i = 0; i < current.Count; i++ ) {
                if ( !current[i].Value ) {
                    common.Add( new KeyValuePair<string, bool>( current[i].Key, false ) );
                }
            }
            SymbolTable.ChangeOrder( common.ToArray() );
            #endregion

            Boare.Lib.AppUtil.Messaging.LoadMessages();
            Boare.Lib.AppUtil.Messaging.Language = EditorConfig.Language;
        }

        [STAThread]
        static void Main() {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault( false );
            try {
                Application.Run( new AppManager() );
            } catch ( Exception ex ) {
                using ( StreamWriter sw = new StreamWriter( Path.Combine( Application.StartupPath, "error.log" ) ) ) {
                    sw.WriteLine( ex.StackTrace );
                }
            }
        }

        public static SolidBrush HilightBrush {
            get {
                return s_hilight_brush;
            }
        }

        public static Color HilightColor {
            get {
                return s_hilight_brush.Color;
            }
            set {
                s_hilight_brush = new SolidBrush( value );
            }
        }

        /// <summary>
        /// ベースとなるテンポ。
        /// </summary>
        public static int BaseTempo {
            get {
                return m_base_tempo;
            }
            set {
                m_base_tempo = value;
            }
        }

        public static void LoadEditorConfig( string config_file ) {
            if ( File.Exists( config_file ) ) {
                FileStream fs = null;
                try {
                    fs = new FileStream( config_file, FileMode.Open );
                    EditorConfig = (EditorConfig)XmlSerializerEditorConfig.Deserialize( fs );
                } catch {
                    FileStream fs2 = null;
                    try {
                        fs2 = new FileStream( config_file, FileMode.Open );
                        BinaryFormatter bf = new BinaryFormatter();
                        EditorConfig = (EditorConfig)bf.Deserialize( fs2 );
                    } catch {
                        EditorConfig = new EditorConfig();
                    } finally {
                        if ( fs2 != null ) {
                            fs2.Close();
                        }
                    }
                } finally {
                    if ( fs != null ) {
                        fs.Close();
                    }
                }
            }
            if ( EditorConfig == null ) {
                EditorConfig = new EditorConfig();
            }
        }

        private void FormRoot_Load( object sender, EventArgs e ) {
            StartNewForm();
        }

        public void StartNewForm() {
            lock ( m_locker ) {
                EditorManager em = new EditorManager();
                em.SetFormRoot( this );
                string s = s_editor_managers.Count + ":" + DateTime.Now.ToBinary();
                string id = bocoree.misc.getmd5( s );
                em.ID = id;
                em.MainFormClosed += new MainFormClosedEventHandler( em_MainFormClosed );
                em.EditorConfigChanged += new EditorConfigChangedEventHandler( em_EditorConfigChanged );
                s_editor_managers.Add( em );
                Thread t = new Thread( new ParameterizedThreadStart( EditorStart ) );
                t.SetApartmentState( ApartmentState.STA );
                t.Start( em );
            }
        }

        private void em_EditorConfigChanged( EditorManager manager ) {
            for ( int i = 0; i < s_editor_managers.Count; i++ ) {
                if ( s_editor_managers[i].ID != manager.ID ) {
                    s_editor_managers[i].SendApplyEditorConfigMessage();
                }
            }
        }

        private void em_MainFormClosed( EditorManager sender ) {
            for ( int i = 0; i < s_editor_managers.Count; i++ ) {
                if ( s_editor_managers[i].Equals( sender ) ) {
                    s_editor_managers.RemoveAt( i );
                    break;
                }
            }
            if ( s_editor_managers.Count == 0 ) {
                this.Invoke( new EventHandler( WindowClose ) );
                KeySoundPlayer.Finalize();
                VSTiProxy.Finalize();
            }
        }

        private void WindowClose( object sender, EventArgs e ) {
            this.Close();
        }

        private void EditorStart( object manager ) {
            EditorManager em = (EditorManager)manager;
            Application.Run( new FormMain( em ) );
        }
    }

}
