﻿/*
 * Common.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.CodeDom.Compiler;
using System.Reflection;
using System.Windows.Forms;
using Microsoft.CSharp;

using bocoree;
using Boare.Lib.Vsq;
using Boare.Lib.AppUtil;

namespace Boare.Cadencii {

    delegate void EditVsqScriptDelegate( VsqFile vsq );

    static class Common {
        public static string _( string id ) {
            return Messaging.GetMessage( id );
        }

        public static EditVsqScriptDelegate LoadScript( string script, EditorConfig editor_config, out string err_msg ) {
#if DEBUG
            Console.WriteLine( "Common.LoadScript(string)" );
#endif
            string code = "";
            code += "using System;";
            code += "using System.IO;";
            code += "using Boare.Lib.Vsq;";
            code += "using Boare.Cadencii;";
            code += "using bocoree;";
            code += "using Boare.Lib.Media;";
            code += "using Boare.Lib.AppUtil;";
            code += "using System.Windows.Forms;";
            code += "using System.Collections.Generic;";
            code += "namespace Boare.CadenciiScript{";
            code += script;
            code += "}";
            CSharpCodeProvider provider = new CSharpCodeProvider();
            string path = Application.StartupPath;
            CompilerParameters parameters = new CompilerParameters( new string[] {
                Path.Combine( path, "Boare.Lib.Vsq.dll" ),
                Path.Combine( path, "Cadencii.exe" ),
                Path.Combine( path, "Boare.Lib.Media.dll" ),
                Path.Combine( path, "Boare.Lib.AppUtil.dll" ),
                Path.Combine( path, "bocoree.dll" ) } );
            parameters.ReferencedAssemblies.Add( "System.Windows.Forms.dll" );
            parameters.ReferencedAssemblies.Add( "System.dll" );
            parameters.GenerateInMemory = true;
            parameters.GenerateExecutable = false;
            parameters.IncludeDebugInformation = true;
            CompilerResults results = provider.CompileAssemblyFromSource( parameters, code );
            err_msg = "";
            if ( results.Errors.Count != 0 ) {
                for ( int i = 0; i < results.Errors.Count; i++ ) {
                    err_msg += _( "line" ) + ":" + (results.Errors[i].Line - 1) + " " + results.Errors[i].ErrorText + "\n";
                }
                if ( results.Errors.HasErrors ) {
                    return null;
                }
            }
            Assembly testAssembly = results.CompiledAssembly;
            EditVsqScriptDelegate ret = null;
            if ( testAssembly != null ) {
                foreach ( Type implemented in testAssembly.GetTypes() ) {
                    MethodInfo tmi = implemented.GetMethod( "Edit", new Type[] { typeof( VsqFile ) } );
                    if ( tmi != null ) {
                        ret = (EditVsqScriptDelegate)Delegate.CreateDelegate( typeof( EditVsqScriptDelegate ), tmi );
                    }
                    /*MethodInfo init = implemented.GetMethod( "Init" );
                    if ( init != null ) {
                        InitVsqScriptDelegate ivsd = (InitVsqScriptDelegate)Delegate.CreateDelegate( typeof( InitVsqScriptDelegate ), init );
                        ivsd.Invoke( common_config, editor_config );
                    }*/
                }
            }
            return ret;
        }


        /// <summary>
        /// fontを使って文字を描画したとき，文字の縦方向の中心線と，描画時に指定した座標との間にどれだけのオフセットが生じるかを調べる
        /// </summary>
        /// <param name="font"></param>
        /// <returns></returns>
        public static unsafe int GetStringOffset( Font font ) {
            SizeF str_size = MeasureString( "Qjp", font );
            if ( str_size.Width <= 0 || str_size.Height <= 0 ) {
                return 0;
            }
            int draw_pos = (int)(str_size.Height / 2);
            Bitmap test = new Bitmap( (int)str_size.Width * 2, (int)str_size.Height * 2, System.Drawing.Imaging.PixelFormat.Format24bppRgb );
            using ( Graphics g = Graphics.FromImage( test ) ) {
                g.DrawString(
                    "Qjp",
                    font,
                    Brushes.Black,
                    new PointF( str_size.Width / 2, draw_pos ) );
            }
            BitmapData bdat = test.LockBits(
                new Rectangle( 0, 0, test.Width, test.Height ),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb );
            int stride = bdat.Stride;
            byte* dat = (byte*)bdat.Scan0;
            bool found = false;
            int firsty = draw_pos;
            for ( int y = 0; y < bdat.Height; y++ ) {
                int pos = y * stride - 3;
                for ( int x = 0; x < bdat.Width; x++ ) {
                    pos += 3;
                    if ( dat[pos] == 0 ) {
                        found = true;
                        break;
                    }
                }
                if ( found ) {
                    firsty = y;
                    break;
                }
            }
            found = false;
            int lasty = draw_pos;
            for ( int y = bdat.Height - 1; y >= 0; y-- ) {
                int pos = y * stride - 3;
                for ( int x = 0; x < bdat.Width; x++ ) {
                    pos++;
                    if ( dat[pos] == 0 ) {
                        found = true;
                        break;
                    }
                }
                if ( found ) {
                    lasty = y;
                    break;
                }
            }
            return (lasty + firsty) / 2 - draw_pos;
        }

        public static SizeF MeasureString( string item, Font font ) {
            using ( Bitmap test = new Bitmap( 10, 10 ) )
            using ( Graphics g = Graphics.FromImage( test ) ) {
                return g.MeasureString( item, font );
            }
        }
        
        /// <summary>
        /// 文字列itemをfontを用いて描画したとき、幅widthピクセルに収まるようにitemを調節したものを返します。
        /// 例えば"1 Voice"→"1 Voi..."ナド。
        /// </summary>
        /// <param name="item"></param>
        /// <param name="font"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static string TrimString( string item, Font font, int width ) {
            string edited = item;
            int delete_count = item.Length;
            while ( true ) {
                SizeF measured = MeasureString( edited, font );
                if ( measured.Width <= width ) {
                    return edited;
                }
                delete_count -= 1;
                if ( delete_count > 0 ) {
                    edited = item.Substring( 0, delete_count ) + "...";
                } else {
                    return edited;
                }
            }
        }

        
        public static void DebugWriteLine( string message ) {
#if DEBUG
            Console.WriteLine( message );
#endif
        }
    }

}