﻿/*
 * FormMainUtil.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.IO;
using System.Media;
using System.Windows.Forms;
using System.Reflection;
using System.Threading;

using Boare.Lib.AppUtil;
using Boare.Lib.Media;
using Boare.Lib.Vsq;
using bocoree;

namespace Boare.Cadencii {

    partial class FormMain {
        /// <summary>
        /// プレイカーソルが見えるようスクロールする
        /// </summary>
        private void EnsureCursorVisible() {
            // カーソルが画面内にあるかどうか検査
            int clock_left = ClockFromXCoord( 68 );
            int clock_right = ClockFromXCoord( pictPianoRoll.Width );
            int uwidth = clock_right - clock_left;
            if ( m_manager.CurrentClock < clock_left || clock_right < m_manager.CurrentClock ) {
                int cl_new_center = (m_manager.CurrentClock / uwidth) * uwidth + uwidth / 2;
                float f_draft = cl_new_center - (pictPianoRoll.Width / 2 + 34 - 70) / ScaleX;
                if ( f_draft < 0f ) {
                    f_draft = 0;
                }
                int draft = (int)(f_draft);
                if ( draft < hScroll.Minimum ) {
                    draft = hScroll.Minimum;
                } else if ( hScroll.Maximum < draft ) {
                    draft = hScroll.Maximum;
                }
                if ( hScroll.Value != draft ) {
                    m_draw_start_index[m_manager.Selected - 1] = 0;
                    hScroll.Value = draft;
                }
            }
        }

        private void ProcessSpecialShortcutKey( PreviewKeyDownEventArgs e ) {
#if DEBUG
            Common.DebugWriteLine( "ProcessSpecialShortcutKey" );
            Common.DebugWriteLine( "    e.KeyCode=" + e.KeyCode );
#endif
            if ( m_input_textbox == null ) {
                if ( e.KeyCode == Keys.Return || e.KeyCode == Keys.Space ) {
                    m_manager.Playing = !m_manager.Playing;
                } else if ( (e.KeyCode & Keys.OemPeriod) == Keys.OemPeriod ) {
                    if ( m_manager.Playing ) {
                        m_manager.Playing = false;
                    } else {
                        if ( !m_manager.StartMarkerEnabled ) {
                            m_manager.CurrentClock = 0;
                        } else {
                            m_manager.CurrentClock = m_manager.StartMarker;
                        }
                        RefreshScreen();
                    }
                } else if ( (e.KeyCode & Keys.Add) == Keys.Add || (e.KeyCode & Keys.Oemplus) == Keys.Oemplus || (e.KeyCode & Keys.Right) == Keys.Right ) {
                    Forward();
                    EnsureCursorVisible();
                } else if ( (e.KeyCode & Keys.Subtract) == Keys.Subtract || (e.KeyCode & Keys.OemMinus) == Keys.OemMinus || (e.KeyCode & Keys.Left) == Keys.Left ) {
                    Rewind();
                    EnsureCursorVisible();
                }
            }
            return;
            #region OLD CODES DO NOT REMOVE
            /*if ( AppManager.EditorConfig.Platform == Platform.Macintosh ) {
                if ( AppManager.EditorConfig.CommandKeyAsControl ) {
                    #region menuStripMain
                    if ( e.Alt && e.KeyCode == Keys.N && menuFileNew.Enabled ) {
                        this.menuFileNew_Click( menuFileNew, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.O && menuFileOpen.Enabled ) {
                        this.menuFileOpen_Click( menuFileOpen, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.S && menuFileSave.Enabled ) {
                        this.menuFileSave_Click( menuFileSave, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.Q && menuFileQuit.Enabled ) {
                        this.menuFileQuit_Click( menuFileQuit, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.Z && menuEditUndo.Enabled ) {
                        this.menuEditUndo_Click( menuEditUndo, null );
                        return;
                    } else if ( e.Alt && e.Shift && e.KeyCode == Keys.Z && menuEditRedo.Enabled ) {
                        this.menuEditRedo_Click( this.menuEditRedo, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.X && this.menuEditCut.Enabled ) {
                        this.menuEditCut_Click( this.menuEditCut, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.C && this.menuEditCopy.Enabled ) {
                        this.menuEditCopy_Click( this.menuEditCopy, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.V && this.menuEditPaste.Enabled ) {
                        this.menuEditPaste_Click( this.menuEditPaste, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.A && this.menuEditSelectAll.Enabled ) {
                        this.menuEditSelectAll_Click( this.menuEditSelectAll, null );
                        return;
                    } else if ( e.Alt && e.Shift && this.menuEditSelectAllEvents.Enabled ) {
                        this.menuEditSelectAllEvents_Click( this.menuEditSelectAllEvents, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.V && this.menuHiddenEditPaste.Enabled ) {
                        this.menuHiddenEditPaste_Click( this.menuHiddenEditPaste, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.W && this.menuHiddenEditFlipToolPointerPencil.Enabled ) {
                        this.menuHiddenEditFlipToolPointerPencil_Click( this.menuHiddenEditFlipToolPointerPencil, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.E && this.menuHiddenEditFlipToolPointerEraser.Enabled ) {
                        this.menuHiddenEditFlipToolPointerEraser_Click( this.menuHiddenEditFlipToolPointerEraser, null );
                        return;
                    } else if ( (e.KeyCode & Keys.Clear) == Keys.Clear && e.Alt && e.Shift && this.menuHiddenVisualForwardParameter.Enabled ) {
                        this.menuHiddenVisualForwardParameter_Click( this.menuHiddenVisualForwardParameter, null );
                        return;
                    } else if ( (e.KeyCode & Keys.LButton) == Keys.LButton && (e.KeyCode & Keys.LineFeed) == Keys.LineFeed && e.Alt && e.Shift && this.menuHiddenVisualBackwardParameter.Enabled ) {
                        this.menuHiddenVisualBackwardParameter_Click( this.menuHiddenVisualBackwardParameter, null );
                        return;
                    } else if ( (e.KeyCode & Keys.Clear) == Keys.Clear && e.Alt && this.menuHiddenTrackNext.Enabled ) {
                        this.menuHiddenTrackNext_Click( this.menuHiddenTrackNext, null );
                        return;
                    } else if ( (e.KeyCode & Keys.LButton) == Keys.LButton && (e.KeyCode & Keys.LineFeed) == Keys.LineFeed && e.Alt && this.menuHiddenTrackBack.Enabled ) {
                        this.menuHiddenTrackBack_Click( this.menuHiddenTrackBack, null );
                        return;
                    }
                    #endregion

                    #region cMenuPiano
                    if ( e.Alt && e.KeyCode == Keys.Z && cMenuPianoUndo.Enabled ) {
                        this.cMenuPianoUndo_Click( this.cMenuPianoUndo, null );
                        return;
                    } else if ( e.Alt && e.Shift && e.KeyCode == Keys.Z && this.cMenuPianoRedo.Enabled ) {
                        this.cMenuPianoRedo_Click( this.cMenuPianoRedo, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.X && this.cMenuPianoCut.Enabled ) {
                        this.cMenuPianoCut_Click( this.cMenuPianoCut, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.C && this.cMenuPianoCopy.Enabled ) {
                        this.cMenuPianoCopy_Click( this.cMenuPianoCopy, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.A && cMenuPianoSelectAll.Enabled ) {
                        this.cMenuPianoSelectAll_Click( this.cMenuPianoSelectAll, null );
                        return;
                    } else if ( e.Alt && e.Shift && e.KeyCode == Keys.A && cMenuPianoSelectAllEvents.Enabled ) {
                        this.cMenuPianoSelectAllEvents_Click( this.cMenuPianoSelectAllEvents, null );
                        return;
                    }
                    #endregion

                    #region cMenuTrackSelector
                    if ( e.Alt && e.KeyCode == Keys.Z && cMenuTrackSelectorUndo.Enabled ) {
                        this.cMenuTrackSelectorUndo_Click( this.cMenuTrackSelectorUndo, null );
                        return;
                    } else if ( e.Alt && e.Shift && e.KeyCode == Keys.Z && this.cMenuTrackSelectorRedo.Enabled ) {
                        this.cMenuTrackSelectorRedo_Click( this.cMenuTrackSelectorRedo, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.X && this.cMenuTrackSelectorCut.Enabled ) {
                        this.cMenuTrackSelectorCut_Click( this.cMenuTrackSelectorCut, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.C && this.cMenuTrackSelectorCopy.Enabled ) {
                        this.cMenuTrackSelectorCopy_Click( this.cMenuTrackSelectorCopy, null );
                        return;
                    } else if ( e.Alt && e.KeyCode == Keys.V && this.cMenuTrackSelectorPaste.Enabled ) {
                        this.cMenuTrackSelectorPaste_Click( this.cMenuTrackSelectorPaste, null );
                        return;
                    } else if ( e.Alt && e.Shift && e.KeyCode == Keys.A && this.cMenuTrackSelectorSelectAll.Enabled ) {
                        this.cMenuTrackSelectorSelectAll_Click( this.cMenuTrackSelectorSelectAll, null );
                        return;
                    }
                    #endregion
                } else {
                    bool RButton = (e.KeyCode & Keys.RButton) == Keys.RButton;
                    bool Clear = (e.KeyCode & Keys.Clear) == Keys.Clear;
                    bool Return = (e.KeyCode & Keys.Return) == Keys.Return;
                    bool Pause = (e.KeyCode & Keys.Pause) == Keys.Pause;
                    bool FinalMode = (e.KeyCode & Keys.FinalMode) == Keys.FinalMode;
                    bool Cancel = (e.KeyCode & Keys.Cancel) == Keys.Cancel;
                    bool CapsLock = (e.KeyCode & Keys.CapsLock) == Keys.CapsLock;
                    bool LButton = (e.KeyCode & Keys.LButton) == Keys.LButton;
                    bool JunjaMode = (e.KeyCode & Keys.JunjaMode) == Keys.JunjaMode;
                    bool LineFeed = (e.KeyCode & Keys.LineFeed) == Keys.LineFeed;
                    bool ControlKey = (e.KeyCode & Keys.ControlKey) == Keys.ControlKey;
                    bool XButton1 = (e.KeyCode & Keys.XButton1) == Keys.XButton1;

                    #region menuStripMain
                    if ( RButton && Clear && (e.KeyCode & Keys.N) == Keys.N && menuFileNew.Enabled ) {
                        this.menuFileNew_Click( menuFileNew, null );
                        return;
                    } else if ( RButton && Return && (e.KeyCode & Keys.O) == Keys.O && menuFileOpen.Enabled ) {
                        this.menuFileOpen_Click( menuFileOpen, null );
                        return;
                    } else if ( Pause && (e.KeyCode & Keys.S) == Keys.S && menuFileSave.Enabled ) {
                        this.menuFileSave_Click( menuFileSave, null );
                        return;
                    } else if ( ControlKey && (e.KeyCode & Keys.Q) == Keys.Q && menuFileQuit.Enabled ) {
                        this.menuFileQuit_Click( menuFileQuit, null );
                        return;
                    } else if ( RButton && FinalMode && (e.KeyCode & Keys.Z) == Keys.Z && menuEditUndo.Enabled ) {
                        this.menuEditUndo_Click( menuEditUndo, null );
                        return;
                    } else if ( RButton && FinalMode && e.Shift && (e.KeyCode & Keys.Z) == Keys.Z && menuEditRedo.Enabled ) {
                        this.menuEditRedo_Click( this.menuEditRedo, null );
                        return;
                    } else if ( FinalMode && (e.KeyCode & Keys.X) == Keys.X && this.menuEditCut.Enabled ) {
                        this.menuEditCut_Click( this.menuEditCut, null );
                        return;
                    } else if ( Cancel && (e.KeyCode & Keys.C) == Keys.C && this.menuEditCopy.Enabled ) {
                        this.menuEditCopy_Click( this.menuEditCopy, null );
                        return;
                    } else if ( RButton && CapsLock && (e.KeyCode & Keys.V) == Keys.V && this.menuEditPaste.Enabled ) {
                        this.menuEditPaste_Click( this.menuEditPaste, null );
                        return;
                    } else if ( LButton && (e.KeyCode & Keys.A) == Keys.A && this.menuEditSelectAll.Enabled ) {
                        this.menuEditSelectAll_Click( this.menuEditSelectAll, null );
                        return;
                    } else if ( LButton && e.Shift && (e.KeyCode & Keys.A) == Keys.A && this.menuEditSelectAllEvents.Enabled ) {
                        this.menuEditSelectAllEvents_Click( this.menuEditSelectAllEvents, null );
                        return;
                    } else if ( RButton && CapsLock && (e.KeyCode & Keys.V) == Keys.V && this.menuHiddenEditPaste.Enabled ) {
                        this.menuHiddenEditPaste_Click( this.menuHiddenEditPaste, null );
                        return;
                    } else if ( JunjaMode && (e.KeyCode & Keys.W) == Keys.W && this.menuHiddenEditFlipToolPointerPencil.Enabled ) {
                        this.menuHiddenEditFlipToolPointerPencil_Click( this.menuHiddenEditFlipToolPointerPencil, null );
                        return;
                    } else if ( XButton1 && (e.KeyCode & Keys.E) == Keys.E && this.menuHiddenEditFlipToolPointerEraser.Enabled ) {
                        this.menuHiddenEditFlipToolPointerEraser_Click( this.menuHiddenEditFlipToolPointerEraser, null );
                        return;
                    } else if ( Clear && e.Control && e.Shift && this.menuHiddenVisualForwardParameter.Enabled ) {
                        this.menuHiddenVisualForwardParameter_Click( this.menuHiddenVisualForwardParameter, null );
                        return;
                    } else if ( LButton && LineFeed && e.Control && e.Shift && this.menuHiddenVisualBackwardParameter.Enabled ) {
                        this.menuHiddenVisualBackwardParameter_Click( this.menuHiddenVisualBackwardParameter, null );
                        return;
                    } else if ( Clear && e.Control && this.menuHiddenTrackNext.Enabled ) {
                        this.menuHiddenTrackNext_Click( this.menuHiddenTrackNext, null );
                        return;
                    } else if ( LButton && LineFeed && e.Control && this.menuHiddenTrackBack.Enabled ) {
                        this.menuHiddenTrackBack_Click( this.menuHiddenTrackBack, null );
                        return;
                    }
                    #endregion

                    #region cMenuPiano
                    if ( RButton && FinalMode && (e.KeyCode & Keys.Z) == Keys.Z && cMenuPianoUndo.Enabled ) {
                        this.cMenuPianoUndo_Click( this.cMenuPianoUndo, null );
                        return;
                    } else if ( RButton && FinalMode && e.Shift && (e.KeyCode & Keys.Z) == Keys.Z && this.cMenuPianoRedo.Enabled ) {
                        this.cMenuPianoRedo_Click( this.cMenuPianoRedo, null );
                        return;
                    } else if ( FinalMode && (e.KeyCode & Keys.X) == Keys.X && this.cMenuPianoCut.Enabled ) {
                        this.cMenuPianoCut_Click( this.cMenuPianoCut, null );
                        return;
                    } else if ( Cancel && (e.KeyCode & Keys.C) == Keys.C && this.cMenuPianoCopy.Enabled ) {
                        this.cMenuPianoCopy_Click( this.cMenuPianoCopy, null );
                        return;
                    } else if ( LButton && (e.KeyCode & Keys.A) == Keys.A && cMenuPianoSelectAll.Enabled ) {
                        this.cMenuPianoSelectAll_Click( this.cMenuPianoSelectAll, null );
                        return;
                    } else if ( LButton && e.Shift && (e.KeyCode & Keys.A) == Keys.A && cMenuPianoSelectAllEvents.Enabled ) {
                        this.cMenuPianoSelectAllEvents_Click( this.cMenuPianoSelectAllEvents, null );
                        return;
                    }
                    #endregion

                    #region cMenuTrackSelector
                    if ( RButton && FinalMode && (e.KeyCode & Keys.Z) == Keys.Z && cMenuTrackSelectorUndo.Enabled ) {
                        this.cMenuTrackSelectorUndo_Click( this.cMenuTrackSelectorUndo, null );
                        return;
                    } else if ( RButton && FinalMode && e.Shift && (e.KeyCode & Keys.Z) == Keys.Z && this.cMenuTrackSelectorRedo.Enabled ) {
                        this.cMenuTrackSelectorRedo_Click( this.cMenuTrackSelectorRedo, null );
                        return;
                    } else if ( FinalMode && (e.KeyCode & Keys.X) == Keys.X && this.cMenuTrackSelectorCut.Enabled ) {
                        this.cMenuTrackSelectorCut_Click( this.cMenuTrackSelectorCut, null );
                        return;
                    } else if ( Cancel && (e.KeyCode & Keys.C) == Keys.C && this.cMenuTrackSelectorCopy.Enabled ) {
                        this.cMenuTrackSelectorCopy_Click( this.cMenuTrackSelectorCopy, null );
                        return;
                    } else if ( RButton && CapsLock && (e.KeyCode & Keys.V) == Keys.V && this.cMenuTrackSelectorPaste.Enabled ) {
                        this.cMenuTrackSelectorPaste_Click( this.cMenuTrackSelectorPaste, null );
                        return;
                    } else if ( LButton && e.Shift && (e.KeyCode & Keys.A) == Keys.A && this.cMenuTrackSelectorSelectAll.Enabled ) {
                        this.cMenuTrackSelectorSelectAll_Click( this.cMenuTrackSelectorSelectAll, null );
                        return;
                    }
                    #endregion
                }
            } else {
                #region menuStripMain
                if ( e.Control && e.KeyCode == Keys.N && menuFileNew.Enabled ) {
                    this.menuFileNew_Click( menuFileNew, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.O && menuFileOpen.Enabled ) {
                    this.menuFileOpen_Click( menuFileOpen, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.S && menuFileSave.Enabled ) {
                    this.menuFileSave_Click( menuFileSave, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.Q && menuFileQuit.Enabled ) {
                    this.menuFileQuit_Click( menuFileQuit, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.Z && menuEditUndo.Enabled ) {
                    this.menuEditUndo_Click( menuEditUndo, null );
                    return;
                } else if ( e.Control && e.Shift && e.KeyCode == Keys.Z && menuEditRedo.Enabled ) {
                    this.menuEditRedo_Click( this.menuEditRedo, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.X && this.menuEditCut.Enabled ) {
                    this.menuEditCut_Click( this.menuEditCut, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.C && this.menuEditCopy.Enabled ) {
                    this.menuEditCopy_Click( this.menuEditCopy, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.V && this.menuEditPaste.Enabled ) {
                    this.menuEditPaste_Click( this.menuEditPaste, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.A && this.menuEditSelectAll.Enabled ) {
                    this.menuEditSelectAll_Click( this.menuEditSelectAll, null );
                    return;
                } else if ( e.Control && e.Shift && this.menuEditSelectAllEvents.Enabled ) {
                    this.menuEditSelectAllEvents_Click( this.menuEditSelectAllEvents, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.V && this.menuHiddenEditPaste.Enabled ) {
                    this.menuHiddenEditPaste_Click( this.menuHiddenEditPaste, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.W && this.menuHiddenEditFlipToolPointerPencil.Enabled ) {
                    this.menuHiddenEditFlipToolPointerPencil_Click( this.menuHiddenEditFlipToolPointerPencil, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.E && this.menuHiddenEditFlipToolPointerEraser.Enabled ) {
                    this.menuHiddenEditFlipToolPointerEraser_Click( this.menuHiddenEditFlipToolPointerEraser, null );
                    return;
                } else if ( e.Control && e.Alt && (e.KeyCode & Keys.PageDown) == Keys.PageDown && this.menuHiddenVisualForwardParameter.Enabled ) {
                    this.menuHiddenVisualForwardParameter_Click( this.menuHiddenVisualForwardParameter, null );
                    return;
                } else if ( e.Control && e.Alt && (e.KeyCode & Keys.PageUp) == Keys.PageUp && this.menuHiddenVisualBackwardParameter.Enabled ) {
                    this.menuHiddenVisualBackwardParameter_Click( this.menuHiddenVisualBackwardParameter, null );
                    return;
                } else if ( e.Control && (e.KeyCode & Keys.PageDown) == Keys.PageDown && this.menuHiddenTrackNext.Enabled ) {
                    this.menuHiddenTrackNext_Click( this.menuHiddenTrackNext, null );
                    return;
                } else if ( e.Control && (e.KeyCode & Keys.PageUp) == Keys.PageUp && this.menuHiddenTrackBack.Enabled ) {
                    this.menuHiddenTrackBack_Click( this.menuHiddenTrackBack, null );
                    return;
                }
                #endregion

                #region cMenuPiano
                if ( e.Control && e.KeyCode == Keys.Z && cMenuPianoUndo.Enabled ) {
                    this.cMenuPianoUndo_Click( this.cMenuPianoUndo, null );
                    return;
                } else if ( e.Control && e.Shift && e.KeyCode == Keys.Z && this.cMenuPianoRedo.Enabled ) {
                    this.cMenuPianoRedo_Click( this.cMenuPianoRedo, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.X && this.cMenuPianoCut.Enabled ) {
                    this.cMenuPianoCut_Click( this.cMenuPianoCut, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.C && this.cMenuPianoCopy.Enabled ) {
                    this.cMenuPianoCopy_Click( this.cMenuPianoCopy, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.A && cMenuPianoSelectAll.Enabled ) {
                    this.cMenuPianoSelectAll_Click( this.cMenuPianoSelectAll, null );
                    return;
                } else if ( e.Control && e.Shift && e.KeyCode == Keys.A && cMenuPianoSelectAllEvents.Enabled ) {
                    this.cMenuPianoSelectAllEvents_Click( this.cMenuPianoSelectAllEvents, null );
                    return;
                }
                #endregion

                #region cMenuTrackSelector
                if ( e.Control && e.KeyCode == Keys.Z && cMenuTrackSelectorUndo.Enabled ) {
                    this.cMenuTrackSelectorUndo_Click( this.cMenuTrackSelectorUndo, null );
                    return;
                } else if ( e.Control && e.Shift && e.KeyCode == Keys.Z && this.cMenuTrackSelectorRedo.Enabled ) {
                    this.cMenuTrackSelectorRedo_Click( this.cMenuTrackSelectorRedo, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.X && this.cMenuTrackSelectorCut.Enabled ) {
                    this.cMenuTrackSelectorCut_Click( this.cMenuTrackSelectorCut, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.C && this.cMenuTrackSelectorCopy.Enabled ) {
                    this.cMenuTrackSelectorCopy_Click( this.cMenuTrackSelectorCopy, null );
                    return;
                } else if ( e.Control && e.KeyCode == Keys.V && this.cMenuTrackSelectorPaste.Enabled ) {
                    this.cMenuTrackSelectorPaste_Click( this.cMenuTrackSelectorPaste, null );
                    return;
                } else if ( e.Control && e.Shift && e.KeyCode == Keys.A && this.cMenuTrackSelectorSelectAll.Enabled ) {
                    this.cMenuTrackSelectorSelectAll_Click( this.cMenuTrackSelectorSelectAll, null );
                    return;
                }
                #endregion
            }*/
            #endregion
        }

        private void SetHScrollRange( int draft_length ) {
            const int _ARROWS = 40; // 両端の矢印の表示幅px（おおよその値）
            if ( draft_length > hScroll.Maximum ) {
                hScroll.Maximum = draft_length;
            }
            int large_change = (int)((pictPianoRoll.Width - 68) / (float)ScaleX);
            int box_width = (int)((hScroll.Width - _ARROWS) * large_change / (float)(hScroll.Maximum + large_change));
            if ( box_width < AppManager.EditorConfig.MinimumScrollHandleWidth ) {
                box_width = AppManager.EditorConfig.MinimumScrollHandleWidth;
                large_change = (int)((float)hScroll.Maximum * (float)box_width / (float)(hScroll.Width - _ARROWS - box_width));
            }
            if ( large_change > 0 ) {
                hScroll.LargeChange = large_change;
            }
        }

        private void SetVScrollRange( int draft_length ) {
            const int _ARROWS = 40; // 両端の矢印の表示幅px（おおよその値）
            if ( draft_length > vScroll.Maximum ) {
                vScroll.Maximum = draft_length;
            }
            int large_change = (int)pictPianoRoll.Height;
            int box_width = (int)((vScroll.Height - _ARROWS) * large_change / (float)(vScroll.Maximum + large_change));
            if ( box_width < AppManager.EditorConfig.MinimumScrollHandleWidth ) {
                box_width = AppManager.EditorConfig.MinimumScrollHandleWidth;
                large_change = (int)((float)vScroll.Maximum * (float)box_width / (float)(vScroll.Width - _ARROWS - box_width));
            }
            if ( large_change > 0 ) {
                vScroll.LargeChange = large_change;
            }
        }

        private static void ApplyRendererRecurse( Control control, ToolStripProfessionalRenderer renderer ) {
            ToolStrip ts = control as ToolStrip;
            if ( ts != null ) {
                if ( renderer == null ) {
                    ts.RenderMode = ToolStripRenderMode.System;
                } else {
                    ts.Renderer = renderer;
                }
            }
            MenuStrip ms = control as MenuStrip;
            if ( ms != null ) {
                if ( renderer == null ) {
                    ms.RenderMode = ToolStripRenderMode.System;
                } else {
                    ms.Renderer = renderer;
                }
            }
            foreach ( Control c in control.Controls ) {
                ApplyRendererRecurse( c, renderer );
            }
        }

        private void RefreshScreenCore() {
            pictPianoRoll.Refresh();
            picturePositionIndicator.Refresh();
            trackSelector.Refresh();
        }

        private void RefreshScreen() {
            if ( !bgWorkScreen.IsBusy ) {
                bgWorkScreen.RunWorkerAsync();
            }
        }

        public void FlipMixerDialogVisible( bool visible ) {
            m_mixer_dlg.Visible = visible;
            AppManager.EditorConfig.MixerVisible = visible;
            menuVisualMixer.Checked = visible;
        }

        /// <summary>
        /// メニューのショートカットキーを、プラットフォームに応じて変更します
        /// </summary>
        private void ApplyPlatformDependentShortCut() {
            if ( AppManager.EditorConfig.Platform == Platform.Macintosh ) {
                string _CO = "";
                //if ( AppManager.EditorConfig.CommandKeyAsControl ) {
                _CO = new string( '\x2318', 1 );
                //} else {
                //_CO = "^";
                //}
                string _SHIFT = "⇧";
                //if ( AppManager.EditorConfig.CommandKeyAsControl ) {
                #region menuStripMain
                menuFileNew.ShortcutKeys = Keys.Alt | Keys.N;
                menuFileOpen.ShortcutKeys = Keys.Alt | Keys.O;
                menuFileSave.ShortcutKeys = Keys.Alt | Keys.S;
                menuFileQuit.ShortcutKeys = Keys.Alt | Keys.Q;

                menuEditUndo.ShortcutKeys = Keys.Alt | Keys.Z;
                menuEditRedo.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.Z;
                menuEditCut.ShortcutKeys = Keys.Alt | Keys.X;
                menuEditCopy.ShortcutKeys = Keys.Alt | Keys.C;
                menuEditPaste.ShortcutKeys = Keys.Alt | Keys.V;
                menuEditSelectAll.ShortcutKeys = Keys.Alt | Keys.A;
                menuEditSelectAllEvents.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.A;

                menuHiddenEditPaste.ShortcutKeys = Keys.Alt | Keys.V;
                menuHiddenEditFlipToolPointerPencil.ShortcutKeys = Keys.Alt | Keys.W;
                menuHiddenEditFlipToolPointerEraser.ShortcutKeys = Keys.Alt | Keys.E;
                menuHiddenVisualForwardParameter.ShortcutKeys = Keys.Clear | Keys.Alt | Keys.Shift;
                menuHiddenVisualBackwardParameter.ShortcutKeys = Keys.LButton | Keys.LineFeed | Keys.Alt | Keys.Shift;
                menuHiddenTrackNext.ShortcutKeys = Keys.Clear | Keys.Alt;
                menuHiddenTrackBack.ShortcutKeys = Keys.LButton | Keys.LineFeed | Keys.Alt;
                #endregion

                #region cMenuPiano
                cMenuPianoUndo.ShortcutKeys = Keys.Alt | Keys.Z;
                cMenuPianoRedo.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.Z;
                cMenuPianoCut.ShortcutKeys = Keys.Alt | Keys.X;
                cMenuPianoCopy.ShortcutKeys = Keys.Alt | Keys.C;
                cMenuPianoSelectAll.ShortcutKeys = Keys.Alt | Keys.A;
                cMenuPianoSelectAllEvents.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.A;
                #endregion

                #region cMenuTrackSelector
                cMenuTrackSelectorUndo.ShortcutKeys = Keys.Alt | Keys.Z;
                cMenuTrackSelectorRedo.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.Z;
                cMenuTrackSelectorCut.ShortcutKeys = Keys.Alt | Keys.X;
                cMenuTrackSelectorCopy.ShortcutKeys = Keys.Alt | Keys.C;
                cMenuTrackSelectorPaste.ShortcutKeys = Keys.Alt | Keys.V;
                cMenuTrackSelectorSelectAll.ShortcutKeys = Keys.Alt | Keys.Shift | Keys.A;
                #endregion
                /*} else {
                    #region menuStripMain
                    menuFileNew.ShortcutKeys = Keys.RButton | Keys.Clear | Keys.N;
                    menuFileOpen.ShortcutKeys = Keys.RButton | Keys.Return | Keys.O;
                    menuFileSave.ShortcutKeys = Keys.Pause | Keys.S;
                    menuFileQuit.ShortcutKeys = Keys.ControlKey | Keys.Q;

                    menuEditUndo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Z;
                    menuEditRedo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Shift | Keys.Z;
                    menuEditCut.ShortcutKeys = Keys.FinalMode | Keys.X;
                    menuEditCopy.ShortcutKeys = Keys.Cancel | Keys.C;
                    menuEditPaste.ShortcutKeys = Keys.RButton | Keys.CapsLock | Keys.V;
                    menuEditSelectAll.ShortcutKeys = Keys.LButton | Keys.A;
                    menuEditSelectAllEvents.ShortcutKeys = Keys.LButton | Keys.Shift | Keys.A;

                    menuHiddenEditPaste.ShortcutKeys = Keys.RButton | Keys.CapsLock | Keys.V;
                    menuHiddenEditFlipToolPointerPencil.ShortcutKeys = Keys.JunjaMode | Keys.W;
                    menuHiddenEditFlipToolPointerEraser.ShortcutKeys = Keys.XButton1 | Keys.E;
                    menuHiddenVisualForwardParameter.ShortcutKeys = Keys.Clear | Keys.Control | Keys.Shift;
                    menuHiddenVisualBackwardParameter.ShortcutKeys = Keys.LButton | Keys.LineFeed | Keys.Control | Keys.Shift;
                    menuHiddenTrackNext.ShortcutKeys = Keys.Clear | Keys.Control;
                    menuHiddenTrackBack.ShortcutKeys = Keys.LButton | Keys.LineFeed | Keys.Control;
                    #endregion

                    #region cMenuPiano
                    cMenuPianoUndo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Z;
                    cMenuPianoRedo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Shift | Keys.Z;
                    cMenuPianoCut.ShortcutKeys = Keys.FinalMode | Keys.X;
                    cMenuPianoCopy.ShortcutKeys = Keys.Cancel | Keys.C;
                    cMenuPianoSelectAll.ShortcutKeys = Keys.LButton | Keys.A;
                    cMenuPianoSelectAllEvents.ShortcutKeys = Keys.LButton | Keys.Shift | Keys.A;
                    #endregion

                    #region cMenuTrackSelector
                    cMenuTrackSelectorUndo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Z;
                    cMenuTrackSelectorRedo.ShortcutKeys = Keys.RButton | Keys.FinalMode | Keys.Shift | Keys.Z;
                    cMenuTrackSelectorCut.ShortcutKeys = Keys.FinalMode | Keys.X;
                    cMenuTrackSelectorCopy.ShortcutKeys = Keys.Cancel | Keys.C;
                    cMenuTrackSelectorPaste.ShortcutKeys = Keys.RButton | Keys.CapsLock | Keys.V;
                    cMenuTrackSelectorSelectAll.ShortcutKeys = Keys.LButton | Keys.Shift | Keys.A;
                    #endregion
                }*/
                menuFileNew.ShortcutKeyDisplayString = _CO + "N";
                menuFileOpen.ShortcutKeyDisplayString = _CO + "O";
                menuFileSave.ShortcutKeyDisplayString = _CO + "S";
                menuFileQuit.ShortcutKeyDisplayString = _CO + "Q";

                menuEditUndo.ShortcutKeyDisplayString = _CO + "Z";
                menuEditRedo.ShortcutKeyDisplayString = _SHIFT + _CO + "Z";
                menuEditCut.ShortcutKeyDisplayString = _CO + "X";
                menuEditCopy.ShortcutKeyDisplayString = _CO + "C";
                menuEditPaste.ShortcutKeyDisplayString = _CO + "V";
                menuEditSelectAll.ShortcutKeyDisplayString = _CO + "A";
                menuEditSelectAllEvents.ShortcutKeyDisplayString = _SHIFT + _CO + "A";

                cMenuPianoUndo.ShortcutKeyDisplayString = _CO + "Z";
                cMenuPianoRedo.ShortcutKeyDisplayString = _SHIFT + _CO + "Z";
                cMenuPianoCut.ShortcutKeyDisplayString = _CO + "X";
                cMenuPianoCopy.ShortcutKeyDisplayString = _CO + "C";
                cMenuPianoSelectAll.ShortcutKeyDisplayString = _CO + "A";
                cMenuPianoSelectAllEvents.ShortcutKeyDisplayString = _SHIFT + _CO + "A";

                cMenuTrackSelectorUndo.ShortcutKeyDisplayString = _CO + "Z";
                cMenuTrackSelectorRedo.ShortcutKeyDisplayString = _SHIFT + _CO + "Z";
                cMenuTrackSelectorCut.ShortcutKeyDisplayString = _CO + "X";
                cMenuTrackSelectorCopy.ShortcutKeyDisplayString = _CO + "C";
                cMenuTrackSelectorPaste.ShortcutKeyDisplayString = _CO + "V";
                cMenuTrackSelectorSelectAll.ShortcutKeyDisplayString = _SHIFT + _CO + "A";
            } else {
                #region menuStripMain
                menuFileNew.ShortcutKeys = Keys.Control | Keys.N;
                menuFileOpen.ShortcutKeys = Keys.Control | Keys.O;
                menuFileSave.ShortcutKeys = Keys.Control | Keys.S;
                menuFileQuit.ShortcutKeys = Keys.Control | Keys.Q;

                menuEditUndo.ShortcutKeys = Keys.Control | Keys.Z;
                menuEditRedo.ShortcutKeys = Keys.Control | Keys.Shift | Keys.Z;
                menuEditCut.ShortcutKeys = Keys.Control | Keys.X;
                menuEditCopy.ShortcutKeys = Keys.Control | Keys.C;
                menuEditPaste.ShortcutKeys = Keys.Control | Keys.V;
                menuEditSelectAll.ShortcutKeys = Keys.Control | Keys.A;
                menuEditSelectAllEvents.ShortcutKeys = Keys.Control | Keys.Shift | Keys.A;

                menuHiddenEditPaste.ShortcutKeys = Keys.Control | Keys.V;
                menuHiddenEditFlipToolPointerPencil.ShortcutKeys = Keys.Control | Keys.W;
                menuHiddenEditFlipToolPointerEraser.ShortcutKeys = Keys.Control | Keys.E;
                menuHiddenVisualForwardParameter.ShortcutKeys = Keys.Control | Keys.Alt | Keys.PageDown;
                menuHiddenVisualBackwardParameter.ShortcutKeys = Keys.Control | Keys.Alt | Keys.PageUp;
                menuHiddenTrackNext.ShortcutKeys = Keys.Control | Keys.PageDown;
                menuHiddenTrackBack.ShortcutKeys = Keys.Control | Keys.PageUp;
                #endregion

                #region cMenuPiano
                cMenuPianoUndo.ShortcutKeys = Keys.Control | Keys.Z;
                cMenuPianoRedo.ShortcutKeys = Keys.Control | Keys.Shift | Keys.Z;
                cMenuPianoCut.ShortcutKeys = Keys.Control | Keys.X;
                cMenuPianoCopy.ShortcutKeys = Keys.Control | Keys.C;
                cMenuPianoSelectAll.ShortcutKeys = Keys.Control | Keys.A;
                cMenuPianoSelectAllEvents.ShortcutKeys = Keys.Control | Keys.Shift | Keys.A;
                #endregion

                #region cMenuTrackSelector
                cMenuTrackSelectorUndo.ShortcutKeys = Keys.Control | Keys.Z;
                cMenuTrackSelectorRedo.ShortcutKeys = Keys.Control | Keys.Shift | Keys.Z;
                cMenuTrackSelectorCut.ShortcutKeys = Keys.Control | Keys.X;
                cMenuTrackSelectorCopy.ShortcutKeys = Keys.Control | Keys.C;
                cMenuTrackSelectorPaste.ShortcutKeys = Keys.Control | Keys.V;
                cMenuTrackSelectorSelectAll.ShortcutKeys = Keys.Control | Keys.Shift | Keys.A;
                #endregion
            }
        }

        /// <summary>
        /// ソングポジションを1小節進めます
        /// </summary>
        private void Forward() {
            bool playing = m_manager.Playing;
            if ( playing ) {
                m_manager.Playing = false;
            }
            int current = m_manager.VsqFile.GetBarCountFromClock( m_manager.CurrentClock ) + 1;
            int new_clock = m_manager.VsqFile.GetClockFromBarCount( current );
            if ( new_clock <= hScroll.Maximum + (pictPianoRoll.Width - 68) / ScaleX ) {
                m_manager.CurrentClock = new_clock;
                m_manager.Playing = playing;
                RefreshScreen();
            }
        }

        /// <summary>
        /// ソングポジションを1小節戻します
        /// </summary>
        private void Rewind() {
            bool playing = m_manager.Playing;
            if ( playing ) {
                m_manager.Playing = false;
            }
            int current = m_manager.VsqFile.GetBarCountFromClock( m_manager.CurrentClock );
            if ( current > 0 ) {
                current--;
            }
            int new_clock = m_manager.VsqFile.GetClockFromBarCount( current );
            m_manager.CurrentClock = new_clock;
            m_manager.Playing = playing;
            RefreshScreen();
        }

        /// <summary>
        /// cMenuPianoの固定長音符入力の各メニューのチェック状態をm_pencil_modeを元に更新します
        /// </summary>
        public void UpdateCMenuPianoFixed() {
            cMenuPianoFixed01.CheckState = CheckState.Unchecked;
            cMenuPianoFixed02.CheckState = CheckState.Unchecked;
            cMenuPianoFixed04.CheckState = CheckState.Unchecked;
            cMenuPianoFixed08.CheckState = CheckState.Unchecked;
            cMenuPianoFixed16.CheckState = CheckState.Unchecked;
            cMenuPianoFixed32.CheckState = CheckState.Unchecked;
            cMenuPianoFixed64.CheckState = CheckState.Unchecked;
            cMenuPianoFixedOff.CheckState = CheckState.Unchecked;
            cMenuPianoFixedTriplet.CheckState = CheckState.Unchecked;
            cMenuPianoFixedDotted.CheckState = CheckState.Unchecked;
            switch ( m_pencil_mode.Mode ) {
                case PencilModeEnum.L1:
                    cMenuPianoFixed01.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L2:
                    cMenuPianoFixed02.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L4:
                    cMenuPianoFixed04.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L8:
                    cMenuPianoFixed08.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L16:
                    cMenuPianoFixed16.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L32:
                    cMenuPianoFixed32.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.L64:
                    cMenuPianoFixed64.CheckState = CheckState.Indeterminate;
                    break;
                case PencilModeEnum.Off:
                    cMenuPianoFixedOff.CheckState = CheckState.Indeterminate;
                    break;
            }
            cMenuPianoFixedTriplet.CheckState = (m_pencil_mode.Triplet) ? CheckState.Indeterminate : CheckState.Unchecked;
            cMenuPianoFixedDotted.CheckState = (m_pencil_mode.Dot) ? CheckState.Indeterminate : CheckState.Unchecked;
        }

        public bool DetermineCaptureForcus() {
            if ( (m_versioninfo == null || (m_versioninfo != null && m_versioninfo.IsDisposed)) &&
                 !m_mixer_dlg.Visible &&
                 (m_form_script_editor == null || (m_form_script_editor != null && !m_form_script_editor.Visible)) ) {
                return true;
            }
            return false;
        }

        private void ClearTempWave() {
            string tmppath = m_manager.TempWaveDir;
            if ( m_sound_player != null && m_sound_player.SoundLocation != "" ) {
#if DEBUG
                Common.DebugWriteLine( "FormMain+ClearTempWave()" );
                Common.DebugWriteLine( "    m_sound_player.SoundLocation=" + m_sound_player.SoundLocation );
#endif
                int ret = m_sound_player.Close();
#if DEBUG
                Common.DebugWriteLine( "    close result=" + ret );
#endif
            }
            for ( int i = 1; i <= 16; i++ ) {
                string file = Path.Combine( tmppath, i + ".wav" );
                if ( File.Exists( file ) ) {
                    for ( int error = 0; error < 100; error++ ) {
                        try {
                            File.Delete( file );
                            break;
                        } catch ( Exception ex ) {
#if DEBUG
                            bocoree.debug.PushLog( "FormMain+ClearTempWave()" );
                            bocoree.debug.PushLog( "    ex=" + ex.ToString() );
                            bocoree.debug.PushLog( "    error_count=" + error );
#endif
                            System.Threading.Thread.Sleep( 100 );
                        }
                    }
                }
            }
        }

        private void Render( int[] tracks ) {
            string tmppath = m_manager.TempWaveDir;
            if ( !Directory.Exists( tmppath ) ) {
                Directory.CreateDirectory( tmppath );
            }
            string[] files = new string[tracks.Length];
            for ( int i = 0; i < tracks.Length; i++ ) {
                files[i] = Path.Combine( tmppath, tracks[i] + ".wav" );
            }
            using ( FormSynthesize dlg = new FormSynthesize( m_manager, m_manager.VsqFile, AppManager.EditorConfig.PreSendTime, tracks, files, (int)m_manager.VsqFile.TotalClocks ) ) {
                if ( dlg.ShowDialog() == DialogResult.OK ) {
                    m_manager.VsqFile.Tracks[m_manager.Selected].ResetEditedArea();
                }
                int[] finished = dlg.Finished;
                for ( int i = 0; i < finished.Length; i++ ) {
                    m_manager.setRenderRequired( finished[i], false );
                }
            }
        }

        private void PlayPreviewSound( int note ) {
            if ( !KeySoundPlayer.IsBusy ) {
                KeySoundPlayer.Play( note );
            }
        }

        private void MouseHoverEventGenerator( object arg ) {
            int note = (int)arg;
            if ( AppManager.EditorConfig.MouseHoverTime > 0 ) {
                Thread.Sleep( AppManager.EditorConfig.MouseHoverTime );
            }
            Invoke( new BSimpleDelegate<int>( PlayPreviewSound ), note );
        }

        public void WaitForPreviewSinglePrepared() {
#if DEBUG
            Common.DebugWriteLine( "WaitForPreviewSinglePrepared" );
#endif
            DateTime wait_start = DateTime.Now;
            while ( !m_preview_single_prepared ) {
                double elapsed = DateTime.Now.Subtract( wait_start ).TotalSeconds;
                if ( elapsed > _WAIT_LIMIT ) {
                    break;
                }
                Application.DoEvents();
            }
            if ( m_preview_single_prepared ) {
                SoundPlayer sp = new SoundPlayer( m_preview_single_file );
                sp.Play();
            }
            m_preview_single_prepared = false;
        }

        private static string StringFromQuantizeMode( QuantizeMode quantize_mode ) {
            switch ( quantize_mode ) {
                case QuantizeMode.off:
                    return _( "Off" );
                case QuantizeMode.p4:
                    return "1/4";
                case QuantizeMode.p8:
                    return "1/8";
                case QuantizeMode.p16:
                    return "1/16";
                case QuantizeMode.p32:
                    return "1/32";
                case QuantizeMode.p64:
                    return "1/64";
                default:
                    return "";
            }
        }

        public static string _( string id ) {
            return Messaging.GetMessage( id );
        }

        public void ApplyLanguage() {
            try {
                openVsqDialog.Filter = _( "VSQ Format(*.vsq)|*.vsq" ) + "|" + _( "Original Format(*.evsq)|*.evsq" ) + "|" + _( "All Files(*.*)|*.*" );
            } catch {
                openVsqDialog.Filter = "VSQ Format(*.vsq)|*.vsq|Original Format(*.evsq)|*.evsq|All Files(*.*)|*.*";
            }
            try {
                saveVsqDialog.Filter = _( "VSQ Format(*.vsq)|*.vsq" ) + "|" + _( "Original Format(*.evsq)|*.evsq" ) + "|" + _( "MIDI Format(*.mid)|*.mid" );
            } catch {
                saveVsqDialog.Filter = "VSQ Format(*.vsq)|*.vsq|Original Format(*.evsq)|*.evsq|MIDI Format(*.mid)|*.mid";
            }
            
            #region main menu
            menuFile.Text = _( "File" ) + "(&F)";
            menuFileNew.Text = _( "New" ) + "(&N)";
            menuFileOpen.Text = _( "Open" ) + "(&O)";
            menuFileSave.Text = _( "Save" ) + "(&S)";
            menuFileSaveNamed.Text = _( "Save As" ) + "(&A)";
            menuFileImport.Text = _( "Import" ) + "(&I)";
            menuFileExport.Text = _( "Export" ) + "(&E)";
            menuFileRecent.Text = _( "Recent Files" ) + "(&R)";
            menuFileQuit.Text = _( "Quit" ) + "(&Q)";

            menuEdit.Text = _( "Edit" ) + "(&E)";
            menuEditUndo.Text = _( "Undo" ) + "(&U)";
            menuEditRedo.Text = _( "Redo" ) + "(&R)";
            menuEditCut.Text = _( "Cut" ) + "(&T)";
            menuEditCopy.Text = _( "Copy" ) + "(&C)";
            menuEditPaste.Text = _( "Paste" ) + "(&P)";
            menuEditDelete.Text = _( "Delete" ) + "(&D)";
            menuEditAutoNormalizeMode.Text = _( "Auto Normalize Mode" ) + "(&N)";
            menuEditSelectAll.Text = _( "Select All" ) + "(&A)";
            menuEditSelectAllEvents.Text = _( "Select All Events" ) + "(&E)";

            menuVisual.Text = _( "View" ) + "(&V)";
            menuVisualToolbar.Text = _( "Toolbar" ) + "(&T)";
            menuVisualControlTrack.Text = _( "Control Track" ) + "(&C)";
            menuVisualMixer.Text = _( "Mixer" ) + "(&X)";
            menuVisualGridline.Text = _( "Grid Line" ) + "(&G)";
            menuVisualStartMarker.Text = _( "Start Marker" ) + "(&S)";
            menuVisualEndMarker.Text = _( "End Marker" ) + "(&E)";
            menuVisualLyrics.Text = _( "Lyrics/Phoneme" ) + "(&L)";
            menuVisualNoteProperty.Text = _( "Note Expression/Vibrato" ) + "(&N)";
            menuVisualToolbarEdit.Text = _( "Edit" ) + "(&E)";
            menuVisualToolbarMeasure.Text = _( "Measure" ) + "(&M)";
            menuVisualToolbarTransport.Text = _( "Transport" ) + "(&P)";
            menuVisualToolbarReset.Text = _( "Reset Toolbar" ) + "(&R)";

            menuJob.Text = _( "Job" ) + "(&J)";
            menuJobNormalize.Text = _( "Normalize Notes" ) + "(&N)";
            menuJobInsertBar.Text = _( "Insert Bars" ) + "(&I)";
            menuJobDeleteBar.Text = _( "Delete Bars" ) + "(&D)";
            menuJobRandomize.Text = _( "Randomize" ) + "(&R)";
            menuJobConnect.Text = _( "Connect Notes" ) + "(&C)";
            menuJobLyric.Text = _( "Insert Lyrics" ) + "(&L)";
            menuJobRewire.Text = _( "Import ReWire Host Tempo" ) + "(&T)";
            menuJobReloadVsti.Text = _( "Reload VSTi" ) + "(&R)";

            menuTrack.Text = _( "Track" ) + "(&T)";
            menuTrackOn.Text = _( "Track On" ) + "(&K)";
            menuTrackAdd.Text = _( "Add Track" ) + "(&A)";
            menuTrackCopy.Text = _( "Copy Track" ) + "(&C)";
            menuTrackChangeName.Text = _( "Rename Track" ) + "(&R)";
            menuTrackDelete.Text = _( "Delete Track" ) + "(&D)";
            menuTrackRenderCurrent.Text = _( "Render Current Track" ) + "(&T)";
            menuTrackRenderAll.Text = _( "Render All Tracks" ) + "(&S)";
            menuTrackOverlay.Text = _( "Overlay" ) + "(&O)";

            menuLyric.Text = _( "Lyrics" ) + "(&L)";
            menuLyricExpressionProperty.Text = _( "Note Expression Property" ) + "(&E)";
            menuLyricVibratoProperty.Text = _( "Note Vibrato Property" ) + "(&V)";
            menuLyricSymbol.Text = _( "Phoneme Transformation" ) + "(&T)";
            menuLyricDictionary.Text = _( "User Word Dictionary" ) + "(&C)";

            menuScript.Text = _( "Script" ) + "(&C)";
            menuScriptRunEditor.Text = _( "Run Script Editor" ) + "(&E)";

            menuSetting.Text = _( "Setting" ) + "(&S)";
            menuSettingPreference.Text = _( "Preference" ) + "(&P)";
            menuSettingDefaultSingerStyle.Text = _( "Singing Style Defaults" ) + "(&D)";
            menuSettingPositionQuantize.Text = _( "Quantize" ) + "(&Q)";
            menuSettingPositionQuantizeOff.Text = _( "Off" );
            menuSettingPositionQuantizeTriplet.Text = _( "Triplet" );
            menuSettingLengthQuantize.Text = _( "Length" ) + "(&L)";
            menuSettingLengthQuantizeOff.Text = _( "Off" );
            menuSettingLengthQuantizeTriplet.Text = _( "Triplet" );
            menuSettingSingerProperty.Text = _( "Singer Properties" ) + "(&S)";
            menuSettingBgmSetSound.Text = _( "Set Sound File" ) + "(&S)";
            menuSettingBgmMute.Text = _( "Mute" ) + "(&M)";

            menuWindow.Text = _( "Window" ) + "(&W)";
            menuWindowNew.Text = _( "New Window" ) + "(&N)";

            menuHelp.Text = _( "Help" ) + "(&H)";
            menuHelpAbout.Text = _( "About Cadencii" ) + "(&A)";
            menuHelpCheckLatestRelease.Text = _( "Check Latest Release" ) + "(&C)";
            #endregion

            #region cMenuPiano
            cMenuPianoPointer.Text = _( "Arrow" ) + "(&A)";
            cMenuPianoPencil.Text = _( "Pencil" ) + "(&W)";
            cMenuPianoEraser.Text = _( "Eraser" ) + "(&E)";
            cMenuPianoCurve.Text = _( "Curve" ) + "(&V)";

            cMenuPianoFixed.Text = _( "Note Fixed Length" ) + "(&N)";
            cMenuPianoFixedTriplet.Text = _( "Triplet" );
            cMenuPianoFixedOff.Text = _( "Off" );
            cMenuPianoFixedDotted.Text = _( "Dot" );
            cMenuPianoQuantize.Text = _( "Quantize" ) + "(&Q)";
            cMenuPianoQuantizeTriplet.Text = _( "Triplet" );
            cMenuPianoQuantizeOff.Text = _( "Off" );
            cMenuPianoLength.Text = _( "Length" ) + "(&L)";
            cMenuPianoLengthTriplet.Text = _( "Triplet" );
            cMenuPianoLengthOff.Text = _( "Off" );
            cMenuPianoGrid.Text = _( "Show/Hide Grid Line" ) + "(&S)";

            cMenuPianoUndo.Text = _( "Undo" ) + "(&U)";
            cMenuPianoRedo.Text = _( "Redo" ) + "(&R)";

            cMenuPianoCut.Text = _( "Cut" ) + "(&T)";
            cMenuPianoPaste.Text = _( "Paste" ) + "(&P)";
            cMenuPianoCopy.Text = _( "Copy" ) + "(&C)";
            cMenuPianoDelete.Text = _( "Delete" ) + "(&D)";

            cMenuPianoSelectAll.Text = _( "Select All" ) + "(&A)";
            cMenuPianoSelectAllEvents.Text = _( "Select All Events" ) + "(&E)";

            cMenuPianoExpressionProperty.Text = _( "Note Expression Property" ) + "(&P)";
            cMenuPianoVibratoProperty.Text = _( "Note Vibrato Property" );
            cMenuPianoImportLyric.Text = _( "Insert Lyrics" ) + "(&P)";
            #endregion

            #region cMenuTrackTab
            cMenuTrackTabTrackOn.Text = _( "Track On" ) + "(&K)";
            cMenuTrackTabAdd.Text = _( "Add Track" ) + "(&A)";
            cMenuTrackTabCopy.Text = _( "Copy Track" ) + "(&C)";
            cMenuTrackTabChangeName.Text = _( "Rename Track" ) + "(&R)";
            cMenuTrackTabDelete.Text = _( "Delete Track" ) + "(&D)";

            cMenuTrackTabRenderCurrent.Text = _( "Render Current Track" ) + "(&T)";
            cMenuTrackTabRenderAll.Text = _( "Render All Tracks" ) + "(&S)";
            cMenuTrackTabOverlay.Text = _( "Overlay" ) + "(&O)";
            #endregion

            #region cMenuTrackSelector
            cMenuTrackSelectorPointer.Text = _( "Arrow" ) + "(&A)";
            cMenuTrackSelectorPencil.Text = _( "Pencil" ) + "(&W)";
            cMenuTrackSelectorLine.Text = _( "Line" ) + "(&L)";
            cMenuTrackSelectorEraser.Text = _( "Eraser" ) + "(&E)";
            cMenuTrackSelectorCurve.Text = _( "Curve" ) + "(&V)";

            cMenuTrackSelectorUndo.Text = _( "Undo" ) + "(&U)";
            cMenuTrackSelectorRedo.Text = _( "Redo" ) + "(&R)";

            cMenuTrackSelectorCut.Text = _( "Cut" ) + "(&T)";
            cMenuTrackSelectorCopy.Text = _( "Copy" ) + "(&C)";
            cMenuTrackSelectorPaste.Text = _( "Paste" ) + "(&P)";
            cMenuTrackSelectorDelete.Text = _( "Delete" ) + "(&D)";

            cMenuTrackSelectorSelectAll.Text = _( "Select All Events" ) + "(&E)";
            #endregion
        }

        private void ImportLyric() {
#if DEBUG
            Common.DebugWriteLine( "ImportLyric" );
#endif
            int start = 0;
            int selectedid = m_manager.LastSelectedEventID;
            for ( int i = 0; i < m_manager.VsqFile.Tracks[m_manager.Selected].Events.Count; i++ ) {
                if ( selectedid == m_manager.VsqFile.Tracks[m_manager.Selected].Events[i].InternalID ) {
                    start = i;
                    break;
                }
            }
            int count = m_manager.VsqFile.Tracks[m_manager.Selected].Events.Count - 1 - start + 1;
#if DEBUG
            Common.DebugWriteLine( "    count=" + count );
#endif
            using ( FormImportLyric dlg = new FormImportLyric( count ) ) {
                if ( dlg.ShowDialog() == DialogResult.OK ) {
                    string[] phrases = dlg.GetLetters();
                    int min = Math.Min( count, phrases.Length );
                    string[] new_phrases = new string[min];
                    string[] new_symbols = new string[min];
                    for ( int i = 0; i < min; i++ ) {
                        new_phrases[i] = phrases[i];
                        SymbolTable.Attatch( phrases[i], out new_symbols[i] );
                    }
                    VsqID[] new_ids = new VsqID[min];
                    int[] ids = new int[min];
                    for ( int i = start; i < start + min; i++ ) {
                        new_ids[i - start] = (VsqID)m_manager.VsqFile.Tracks[m_manager.Selected].Events[i].ID.Clone();
                        new_ids[i - start].LyricHandle.L0.Phrase = new_phrases[i - start];
                        new_ids[i - start].LyricHandle.L0.PhoneticSymbol = new_symbols[i - start];
                        ids[i - start] = m_manager.VsqFile.Tracks[m_manager.Selected].Events[i].InternalID;
                    }
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandEventChangeIDContaintsRange( m_manager.Selected, ids, new_ids ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    this.Refresh();
                }
            }
        }

        private void NoteVibratoProperty() {
            SelectedEventEntry item = m_manager.LastSelectedEvent;
            if ( item == null ) {
                return;
            }

            VsqEvent ev = item.Original;
            using ( FormVibratoConfig dlg = new FormVibratoConfig( ev.ID.VibratoHandle, ev.ID.Length, AppManager.EditorConfig.DefaultVibratoLength ) ) {
                if ( dlg.ShowDialog() == DialogResult.OK ) {
                    VsqEvent edited = (VsqEvent)ev.Clone();
                    if ( dlg.VibratoHandle != null ) {
                        edited.ID.VibratoHandle = (VibratoHandle)dlg.VibratoHandle.Clone();
                        edited.ID.VibratoDelay = ev.ID.Length - dlg.VibratoHandle.Length;
                    } else {
                        edited.ID.VibratoHandle = null;
                    }
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandEventChangeIDContaints( m_manager.Selected, ev.InternalID, (VsqID)edited.ID.Clone() ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    RefreshScreen();
                }
            }
        }

        private void NoteExpressionProperty() {
            SelectedEventEntry item = m_manager.LastSelectedEvent;
            if ( item == null ) {
                return;
            }

            VsqEvent ev = item.Original;
            using ( FormSingerStyleConfig dlg = new FormSingerStyleConfig( _( "Expression control property" ), false ) ) {
                dlg.PMBendDepth = ev.ID.PMBendDepth;
                dlg.PMBendLength = ev.ID.PMBendLength;
                dlg.PMbPortamentoUse = ev.ID.PMbPortamentoUse;
                dlg.DEMdecGainRate = ev.ID.DEMdecGainRate;
                dlg.DEMaccent = ev.ID.DEMaccent;
                if ( dlg.ShowDialog() == DialogResult.OK ) {
                    VsqEvent edited = (VsqEvent)ev.Clone();
                    edited.ID.PMBendDepth = dlg.PMBendDepth;
                    edited.ID.PMBendLength = dlg.PMBendLength;
                    edited.ID.PMbPortamentoUse = dlg.PMbPortamentoUse;
                    edited.ID.DEMdecGainRate = dlg.DEMdecGainRate;
                    edited.ID.DEMaccent = dlg.DEMaccent;
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandEventChangeIDContaints( m_manager.Selected, ev.InternalID, (VsqID)edited.ID.Clone() ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    RefreshScreen();
                }
            }
        }

        private int NewHScrollValueFromWheelDelta( int delta ) {
            double new_val = (double)hScroll.Value - delta * AppManager.EditorConfig.WheelOrder / (5.0 * ScaleX);
            if ( new_val < 0.0 ) {
                new_val = 0;
            }
            int draft = (int)new_val;
            if ( draft > hScroll.Maximum ) {
                draft = hScroll.Maximum;
            } else if ( draft < hScroll.Minimum ) {
                draft = hScroll.Minimum;
            }
            return draft;
        }

        #region 音符の編集関連
        private void SelectAll() {
            m_manager.ClearSelectedEvent();
            m_manager.ClearSelectedTempo();
            m_manager.ClearSelectedTimesig();
            int min = int.MaxValue;
            int max = int.MinValue;
            foreach ( VsqEvent ve in m_manager.VsqFile.Tracks[m_manager.Selected].Events ) {
                m_manager.AddSelectedEvent( ve.InternalID );
                min = Math.Min( min, ve.Clock );
                max = Math.Max( max, ve.Clock + ve.ID.Length );
            }
            int premeasure = m_manager.VsqFile.PreMeasureClocks;
            foreach ( VsqCurveType vct in Enum.GetValues( typeof( VsqCurveType ) ) ) {
                if ( vct == VsqCurveType.VEL ) {
                    continue;
                }
                int count = m_manager.VsqFile.Tracks[m_manager.Selected][vct].Count;
                if ( count >= 1 ) {
                    int max_key = m_manager.VsqFile.Tracks[m_manager.Selected][vct].Keys[count - 1];
                    max = Math.Max( max, m_manager.VsqFile.Tracks[m_manager.Selected][vct][max_key] );
                    for ( int i = 0; i < count; i++ ) {
                        int key = m_manager.VsqFile.Tracks[m_manager.Selected][vct].Keys[i];
                        if ( key >= premeasure ) {
                            min = Math.Min( min, key );
                            break;
                        }
                    }
                }
            }
            if ( min < premeasure ) {
                min = premeasure;
            }
            if ( min < max ) {
                int stdx = StartToDrawX;
                min = XCoordFromClocks( min ) + stdx;
                max = XCoordFromClocks( max ) + stdx;
                m_manager.SelectedRegion = new SelectedRegion( min );
                m_manager.SelectedRegion.SetEnd( max );
                m_manager.SelectedRegionEnabled = true;
            }
        }


        private void SelectAllEvent() {
            m_manager.ClearSelectedTempo();
            m_manager.ClearSelectedTimesig();
            m_manager.ClearSelectedEvent();
            foreach ( VsqEvent ev in m_manager.VsqFile.Tracks[m_manager.Selected].Events ) {
                m_manager.AddSelectedEvent( ev.InternalID );
            }
            this.Refresh();
        }


        private void DeleteEvent() {
#if DEBUG
            Common.DebugWriteLine( "DeleteEvent()" );
            Common.DebugWriteLine( "   (m_input_textbox==null)=" + (m_input_textbox == null) );
#endif
            if ( m_input_textbox == null ) {
                if ( m_manager.SelectedEvent.Count > 0 ) {
                    List<int> ids = new List<int>();
                    foreach ( KeyValuePair<int, SelectedEventEntry> ev in m_manager.SelectedEvent ) {
                        ids.Add( ev.Key );
                    }
                    VsqCommand run = VsqCommand.GCommandEventDeleteRange( m_manager.Selected, ids.ToArray() );
                    if ( m_manager.SelectedRegionEnabled ) {
                        VsqFileEx work = (VsqFileEx)m_manager.VsqFile.Clone();
                        work.Execute( run );
                        int stdx = StartToDrawX;
                        int start_clock = ClockFromXCoord( m_manager.SelectedRegion.Start - stdx );
                        int end_clock = ClockFromXCoord( m_manager.SelectedRegion.End - stdx );
                        List<List<BPPair>> curves = new List<List<BPPair>>();
                        List<VsqCurveType> types = new List<VsqCurveType>();
                        foreach ( VsqCurveType vct in Enum.GetValues( typeof( VsqCurveType ) ) ) {
                            if ( vct == VsqCurveType.VEL ) {
                                continue;
                            }
                            List<BPPair> t = new List<BPPair>();
                            t.Add( new BPPair( start_clock, work.Tracks[m_manager.Selected][vct][start_clock] ) );
                            t.Add( new BPPair( end_clock, work.Tracks[m_manager.Selected][vct][end_clock] ) );
                            curves.Add( t );
                            types.Add( vct );
                        }
                        VsqCommand delete_curve = VsqCommand.GCommandTrackEditCurveRange( m_manager.Selected, types.ToArray(), curves.ToArray() );
                        work.Execute( delete_curve );
                        CadenciiCommand run2 = new CadenciiCommand( VsqCommand.GCommandReplace( work ) );
                        m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                        Edited = true;
                    } else {
                        CadenciiCommand run2 = new CadenciiCommand( run );
                        m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                        Edited = true;
                        m_manager.ClearSelectedEvent();
                    }
                    this.Refresh();
                } else if ( m_manager.SelectedTempo.Count > 0 ) {
                    List<int> clocks = new List<int>();
                    foreach ( KeyValuePair<int, SelectedTempoEntry> ev in m_manager.SelectedTempo ) {
                        clocks.Add( ev.Key );
                    }
                    int[] dum = new int[clocks.Count];
                    for ( int i = 0; i < dum.Length; i++ ) {
                        dum[i] = -1;
                    }
                    CadenciiCommand run = new CadenciiCommand( VsqCommand.GCommandUpdateTempoRange( clocks.ToArray(), clocks.ToArray(), dum ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    m_manager.ClearSelectedTempo();
                    this.Refresh();
                } else if ( m_manager.SelectedTimesig.Count > 0 ) {
#if DEBUG
                    Common.DebugWriteLine( "    Timesig" );
#endif
                    int[] barcounts = new int[m_manager.SelectedTimesig.Count];
                    int[] numerators = new int[m_manager.SelectedTimesig.Count];
                    int[] denominators = new int[m_manager.SelectedTimesig.Count];
                    int count = -1;
                    foreach ( KeyValuePair<int, SelectedTimesigEntry> ev in m_manager.SelectedTimesig ) {
                        count++;
                        barcounts[count] = ev.Key;
                        numerators[count] = -1;
                        denominators[count] = -1;
                    }
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandUpdateTimesigRange( barcounts, barcounts, numerators, denominators ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    m_manager.ClearSelectedTimesig();
                    this.Refresh();
                }
                if ( trackSelector.SelectedRegionEnabled ) {
#if DEBUG
                    Common.DebugWriteLine( "    Curve" );
#endif
                    KeyValuePair<int, int> t = trackSelector.SelectedRegion;
                    int start = t.Key;
                    int end = t.Value;
                    List<BPPair> edit = new List<BPPair>();
                    int value = m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve][start];
                    edit.Add( new BPPair( start, value ) );
                    edit.Add( new BPPair( end, value ) );
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandTrackEditCurve( m_manager.Selected, trackSelector.SelectedCurve, edit ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    trackSelector.DisableSelectedRegion();
                    RefreshScreen();
                }
            }
        }

        private void PasteEvent() {
#if DEBUG
            Common.DebugWriteLine( "PasteEvent" );
#endif
            int clock = m_manager.CurrentClock;
            int unit = GetUnitClocksOfQuantize( AppManager.EditorConfig.PositionQuantize, AppManager.EditorConfig.PositionQuantizeTriplet );
            int odd = clock % unit;
            clock -= odd;
            if ( odd > unit / 2 ) {
                clock += unit;
            }

            if ( AppManager.CopiedEvent != null ) {
                if ( clock < m_manager.VsqFile.PreMeasureClocks ) {
                    return;
                }
                int min_clock = AppManager.CopiedEvent[0].Clock;
                foreach ( VsqEvent item in AppManager.CopiedEvent ) {
                    min_clock = Math.Min( min_clock, item.Clock );
                }
                int dclock = clock - min_clock;
                List<VsqEvent> paste = new List<VsqEvent>();
                for ( int i = 0; i < AppManager.CopiedEvent.Count; i++ ) {
                    VsqEvent item = (VsqEvent)AppManager.CopiedEvent[i].Clone();
                    item.Clock = AppManager.CopiedEvent[i].Clock + dclock;
                    paste.Add( item );
                }
                VsqCommand run = VsqCommand.GCommandEventAddRange( m_manager.Selected, paste.ToArray() );
                if ( AppManager.IsWholeCurveCopied ) {
                    VsqFile work = (VsqFile)m_manager.VsqFile.Clone();
                    work.Execute( run );
                    int start_clock = int.MaxValue;
                    int end_clock = int.MinValue;
                    foreach ( KeyValuePair<VsqCurveType, List<BPPair>> item in AppManager.CopiedCurveWhole ) {
                        if ( item.Value.Count > 0 ) {
                            start_clock = Math.Min( start_clock, item.Value[0].Clock );
                            end_clock = Math.Max( end_clock, item.Value[item.Value.Count - 1].Clock );
                        }
                    }
                    int dclock2 = clock - start_clock;
                    List<List<BPPair>> edits = new List<List<BPPair>>();
                    List<VsqCurveType> targets = new List<VsqCurveType>();
                    foreach ( VsqCurveType vct in AppManager.CopiedCurveWhole.Keys ) {
                        targets.Add( vct );
                        List<BPPair> t = new List<BPPair>();
                        foreach ( BPPair bpp in AppManager.CopiedCurveWhole[vct] ) {
                            if ( start_clock <= bpp.Clock && bpp.Clock < end_clock ) {
                                t.Add( new BPPair( bpp.Clock + dclock2, bpp.Value ) );
                            }
                        }
                        t.Add( new BPPair( end_clock + dclock2, work.Tracks[m_manager.Selected][vct][end_clock + dclock2] ) );
                        edits.Add( t );
                    }
                    VsqCommand replace_curve = VsqCommand.GCommandTrackEditCurveRange( m_manager.Selected, targets.ToArray(), edits.ToArray() );
                    work.Execute( replace_curve );
                    VsqTrack work_track = (VsqTrack)work.Tracks[m_manager.Selected].Clone();
                    CadenciiCommand run2 = new CadenciiCommand( VsqCommand.GCommandTrackReplace( m_manager.Selected, work_track ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                    Edited = true;
                } else {
                    CadenciiCommand run2 = new CadenciiCommand( run );
                    m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                    Edited = true;
                }
                this.Refresh();
            }
            if ( AppManager.CopiedTempo != null ) {
                int min_clock = AppManager.CopiedTempo[0].Clock;
                foreach ( TempoTableEntry tte in AppManager.CopiedTempo ) {
                    min_clock = Math.Min( min_clock, tte.Clock );
                }
                int dclock = clock - min_clock;
                int[] clocks = new int[AppManager.CopiedTempo.Count];
                int[] tempos = new int[AppManager.CopiedTempo.Count];
                for ( int i = 0; i < AppManager.CopiedTempo.Count; i++ ) {
                    clocks[i] = AppManager.CopiedTempo[i].Clock + dclock;
                    tempos[i] = AppManager.CopiedTempo[i].Tempo;
                }
                CadenciiCommand run = new CadenciiCommand( VsqCommand.GCommandUpdateTempoRange( clocks, clocks, tempos ) );
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                Edited = true;
                this.Refresh();
            }
            if ( AppManager.CopiedTimesig != null ) {
                int bar_count = m_manager.VsqFile.GetBarCountFromClock( clock );
                int min_barcount = AppManager.CopiedTimesig[0].BarCount;
                foreach ( TimeSigTableEntry tste in AppManager.CopiedTimesig ) {
                    min_barcount = Math.Min( min_barcount, tste.BarCount );
                }
                int dbarcount = bar_count - min_barcount;
                int[] barcounts = new int[AppManager.CopiedTimesig.Count];
                int[] numerators = new int[AppManager.CopiedTimesig.Count];
                int[] denominators = new int[AppManager.CopiedTimesig.Count];
                for ( int i = 0; i < AppManager.CopiedTimesig.Count; i++ ) {
                    barcounts[i] = AppManager.CopiedTimesig[i].BarCount + dbarcount;
                    numerators[i] = AppManager.CopiedTimesig[i].Numerator;
                    denominators[i] = AppManager.CopiedTimesig[i].Denominator;
                }
                CadenciiCommand run = new CadenciiCommand(
                    VsqCommand.GCommandUpdateTimesigRange( barcounts, barcounts, numerators, denominators ) );
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                Edited = true;
                this.Refresh();
            } else if ( AppManager.CopiedCurve != null ) {
                if ( trackSelector.SelectedCurve != VsqCurveType.VEL ) {
                    KeyValuePair<int, int> kvp = trackSelector.SelectedRegion;
                    int min = Math.Min( kvp.Key, kvp.Value );
                    int max = Math.Max( kvp.Key, kvp.Value );
                    //int start = m_config.VsqFile.Tracks[m_config.Selected].MetaText[trackSelector.SelectedCurve][min];
                    int dclock = m_manager.CurrentClock - min;
                    int end = m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve][max + dclock];
                    List<BPPair> edit = new List<BPPair>();
                    for ( int i = 0; i < AppManager.CopiedCurve.Count; i++ ) {
#if DEBUG
                        Common.DebugWriteLine( "    clock=" + (AppManager.CopiedCurve[i].Clock + dclock) );
#endif
                        edit.Add( new BPPair( AppManager.CopiedCurve[i].Clock + dclock, AppManager.CopiedCurve[i].Value ) );
                    }
                    bool end_found = false;
#if DEBUG
                    Common.DebugWriteLine( "    ---------------------------------------" );
#endif
                    edit[edit.Count - 1].Value = end;
                    bool changed = true;
                    while ( changed ) {
                        changed = false;
                        for ( int i = 0; i < edit.Count - 1; i++ ) {
                            if ( edit[i].Clock > edit[i + 1].Clock ) {
                                BPPair tmp = new BPPair( edit[i].Clock, edit[i].Value );
                                edit[i].Clock = edit[i + 1].Clock;
                                edit[i].Value = edit[i + 1].Value;
                                edit[i + 1].Clock = tmp.Clock;
                                edit[i + 1].Value = tmp.Value;
                                changed = true;
                            }
                        }
                    }
                    CadenciiCommand run = new CadenciiCommand(
                        VsqCommand.GCommandTrackEditCurve( m_manager.Selected, trackSelector.SelectedCurve, edit ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run ) );
                    Edited = true;
                    trackSelector.DisableSelectedRegion();
                    RefreshScreen();
                }
            }
        }

        private void CopyEvent() {
#if DEBUG
            Common.DebugWriteLine( "CopyEvent" );
#endif
            if ( m_manager.SelectedEvent.Count > 0 ) {
                AppManager.CopiedTempo = null;
                AppManager.CopiedTimesig = null;
                if ( AppManager.CopiedEvent != null ) {
                    AppManager.CopiedEvent.Clear();
                } else {
                    AppManager.CopiedEvent = new List<VsqEvent>();
                }
                foreach ( KeyValuePair<int, SelectedEventEntry> item in m_manager.SelectedEvent ) {
                    AppManager.CopiedEvent.Add( (VsqEvent)item.Value.Original.Clone() );
                }
            }
            if ( m_manager.SelectedTempo.Count > 0 ) {
                AppManager.CopiedEvent = null;
                AppManager.CopiedTimesig = null;
                if ( AppManager.CopiedTempo != null ) {
                    AppManager.CopiedTempo.Clear();
                } else {
                    AppManager.CopiedTempo = new List<TempoTableEntry>();
                }
                foreach ( KeyValuePair<int, SelectedTempoEntry> item in m_manager.SelectedTempo ) {
                    AppManager.CopiedTempo.Add( (TempoTableEntry)item.Value.Original.Clone() );
                }
            }
            if ( m_manager.SelectedTimesig.Count > 0 ) {
                AppManager.CopiedEvent = null;
                AppManager.CopiedTempo = null;
                if ( AppManager.CopiedTimesig != null ) {
                    AppManager.CopiedTimesig.Clear();
                } else {
                    AppManager.CopiedTimesig = new List<TimeSigTableEntry>();
                }
                foreach ( KeyValuePair<int, SelectedTimesigEntry> item in m_manager.SelectedTimesig ) {
                    AppManager.CopiedTimesig.Add( (TimeSigTableEntry)item.Value.Original.Clone() );
                }
            }
#if DEBUG
            Common.DebugWriteLine( "    trackSelector.SelectedRegionEnabled=" + trackSelector.SelectedRegionEnabled );
#endif
            if ( trackSelector.SelectedRegionEnabled ) {
                AppManager.IsWholeCurveCopied = false;
                if ( AppManager.CopiedCurve != null ) {
                    AppManager.CopiedCurve.Clear();
                } else {
                    AppManager.CopiedCurve = new List<BPPair>();
                }
                KeyValuePair<int, int> t = trackSelector.SelectedRegion;
                int start = t.Key;
                int end = t.Value;
                foreach ( int clock in m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve].GetKeyClockEnumerator() ) {
                    if ( start <= clock && clock <= end ) {
                        AppManager.CopiedCurve.Add( new BPPair( clock, m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve][clock] ) );
                    }
                }
                AppManager.CopiedCurveType = trackSelector.SelectedCurve;
            } else if ( m_manager.SelectedRegionEnabled ) {
                AppManager.IsWholeCurveCopied = true;
                int stdx = StartToDrawX;
                int start_clock = ClockFromXCoord( m_manager.SelectedRegion.Start - stdx );
                int end_clock = ClockFromXCoord( m_manager.SelectedRegion.End - stdx );
                if ( AppManager.CopiedCurveWhole != null ) {
                    AppManager.CopiedCurveWhole.Clear();
                } else {
                    AppManager.CopiedCurveWhole = new Dictionary<VsqCurveType, List<BPPair>>();
                }
                foreach ( VsqCurveType vct in Enum.GetValues( typeof( VsqCurveType ) ) ) {
                    if ( vct == VsqCurveType.VEL ) {
                        continue;
                    }
                    List<BPPair> tmp = new List<BPPair>();
                    tmp.Add( new BPPair( start_clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][start_clock] ) );
                    foreach ( int clock in m_manager.VsqFile.Tracks[m_manager.Selected][vct].GetKeyClockEnumerator() ) {
                        if ( start_clock < clock && clock < end_clock ) {
                            tmp.Add( new BPPair( clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][clock] ) );
                        }
                    }
                    tmp.Add( new BPPair( end_clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][end_clock] ) );
                    AppManager.CopiedCurveWhole.Add( vct, tmp );
                }
            }
        }

        private void CutEvent() {
            if ( m_manager.SelectedEvent.Count > 0 ) {
                AppManager.CopiedTempo = null;
                AppManager.CopiedTimesig = null;
                if ( AppManager.CopiedEvent != null ) {
                    AppManager.CopiedEvent.Clear();
                } else {
                    AppManager.CopiedEvent = new List<VsqEvent>();
                }
                List<int> ids = new List<int>();
                foreach ( KeyValuePair<int, SelectedEventEntry> item in m_manager.SelectedEvent ) {
                    AppManager.CopiedEvent.Add( (VsqEvent)item.Value.Original.Clone() );
                    ids.Add( item.Key );
                }
                VsqCommand run = VsqCommand.GCommandEventDeleteRange( m_manager.Selected, ids.ToArray() );
                if ( m_manager.SelectedRegionEnabled ) {
                    AppManager.IsWholeCurveCopied = true;
                    int stdx = StartToDrawX;
                    int start_clock = ClockFromXCoord( m_manager.SelectedRegion.Start - stdx );
                    int end_clock = ClockFromXCoord( m_manager.SelectedRegion.End - stdx );
                    if ( AppManager.CopiedCurveWhole != null ) {
                        AppManager.CopiedCurveWhole.Clear();
                    } else {
                        AppManager.CopiedCurveWhole = new Dictionary<VsqCurveType, List<BPPair>>();
                    }
                    foreach ( VsqCurveType vct in Enum.GetValues( typeof( VsqCurveType ) ) ) {
                        if ( vct == VsqCurveType.VEL ) {
                            continue;
                        }
                        List<BPPair> tmp = new List<BPPair>();
                        tmp.Add( new BPPair( start_clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][start_clock] ) );
                        foreach ( int clock in m_manager.VsqFile.Tracks[m_manager.Selected][vct].GetKeyClockEnumerator() ) {
                            if ( start_clock < clock && clock < end_clock ) {
                                tmp.Add( new BPPair( clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][clock] ) );
                            }
                        }
                        tmp.Add( new BPPair( end_clock, m_manager.VsqFile.Tracks[m_manager.Selected][vct][end_clock] ) );
                        AppManager.CopiedCurveWhole.Add( vct, tmp );
                    }
                    VsqFile work = (VsqFile)m_manager.VsqFile.Clone();
                    work.Execute( run );
                    List<List<BPPair>> curves = new List<List<BPPair>>();
                    List<VsqCurveType> types = new List<VsqCurveType>();
                    foreach ( VsqCurveType vct in Enum.GetValues( typeof( VsqCurveType ) ) ) {
                        if ( vct == VsqCurveType.VEL ) {
                            continue;
                        }
                        List<BPPair> t = new List<BPPair>();
                        t.Add( new BPPair( start_clock, work.Tracks[m_manager.Selected][vct][start_clock] ) );
                        t.Add( new BPPair( end_clock, work.Tracks[m_manager.Selected][vct][end_clock] ) );
                        curves.Add( t );
                        types.Add( vct );
                    }
                    VsqCommand delete_curve = VsqCommand.GCommandTrackEditCurveRange( m_manager.Selected, types.ToArray(), curves.ToArray() );
                    work.Execute( delete_curve );
                    CadenciiCommand run2 = new CadenciiCommand( VsqCommand.GCommandReplace( work ) );
                    m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                    Edited = true;
                } else {
                    AppManager.IsWholeCurveCopied = false;
                    CadenciiCommand run2 = new CadenciiCommand( run );
                    m_manager.Register( m_manager.VsqFile.Execute( run2 ) );
                    m_manager.ClearSelectedEvent();
                    Edited = true;
                }
                RefreshScreen();
            } else if ( m_manager.SelectedTempo.Count > 0 ) {
                AppManager.CopiedEvent = null;
                AppManager.CopiedTimesig = null;
                if ( AppManager.CopiedTempo != null ) {
                    AppManager.CopiedTempo.Clear();
                } else {
                    AppManager.CopiedTempo = new List<TempoTableEntry>();
                }

                int count = -1;
                int[] dum = new int[m_manager.SelectedTempo.Count];
                int[] clocks = new int[m_manager.SelectedTempo.Count];
                foreach ( KeyValuePair<int, SelectedTempoEntry> item in m_manager.SelectedTempo ) {
                    count++;
                    AppManager.CopiedTempo.Add( (TempoTableEntry)item.Value.Original.Clone() );
                    dum[count] = -1;
                    clocks[count] = item.Value.Original.Clock;
                }
                CadenciiCommand run = new CadenciiCommand( VsqCommand.GCommandUpdateTempoRange( clocks, clocks, dum ) );
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                Edited = true;
                this.Refresh();
            } else if ( m_manager.SelectedTimesig.Count > 0 ) {
                AppManager.CopiedEvent = null;
                AppManager.CopiedTempo = null;
                if ( AppManager.CopiedTimesig != null ) {
                    AppManager.CopiedTimesig.Clear();
                } else {
                    AppManager.CopiedTimesig = new List<TimeSigTableEntry>();
                }
                int[] barcounts = new int[m_manager.SelectedTimesig.Count];
                int[] numerators = new int[m_manager.SelectedTimesig.Count];
                int[] denominators = new int[m_manager.SelectedTimesig.Count];
                int count = -1;
                foreach ( KeyValuePair<int, SelectedTimesigEntry> item in m_manager.SelectedTimesig ) {
                    AppManager.CopiedTimesig.Add( (TimeSigTableEntry)item.Value.Original.Clone() );
                    count++;
                    barcounts[count] = item.Value.Original.BarCount;
                    numerators[count] = -1;
                    denominators[count] = -1;
                }
                CadenciiCommand run = new CadenciiCommand(
                    VsqCommand.GCommandUpdateTimesigRange( barcounts, barcounts, numerators, denominators ) );
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                Edited = true;
                this.Refresh();
            } else if ( trackSelector.SelectedRegionEnabled ) {
                if ( AppManager.CopiedCurve != null ) {
                    AppManager.CopiedCurve.Clear();
                } else {
                    AppManager.CopiedCurve = new List<BPPair>();
                }
                KeyValuePair<int, int> t = trackSelector.SelectedRegion;
                int start = t.Key;
                int end = t.Value;
                foreach ( int clock in m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve].GetKeyClockEnumerator() ) {
                    if ( start <= clock && clock <= end ) {
                        AppManager.CopiedCurve.Add( new BPPair( clock, m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve][clock] ) );
                    }
                }
                List<BPPair> edit = new List<BPPair>();
                int value = m_manager.VsqFile.Tracks[m_manager.Selected][trackSelector.SelectedCurve][start];
                edit.Add( new BPPair( start, value ) );
                edit.Add( new BPPair( end, value ) );
                AppManager.CopiedCurveType = trackSelector.SelectedCurve;
                CadenciiCommand run = new CadenciiCommand(
                    VsqCommand.GCommandTrackEditCurve( m_manager.Selected, trackSelector.SelectedCurve, edit ) );
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                Edited = true;
                trackSelector.DisableSelectedRegion();
                this.Refresh();
            }
        }
        #endregion

        #region トラックの編集関連
        private void TrackCopyCore() {
            VsqTrack track = (VsqTrack)m_manager.VsqFile.Tracks[m_manager.Selected].Clone();
            track.Name += " (1)";
            CadenciiCommand run = new CadenciiCommand(
                VsqCommand.GCommandAddTrack( track, m_manager.VsqFile.Mixer.Slave[m_manager.Selected - 1], m_manager.VsqFile.Tracks.Count ) );
            m_manager.Register( m_manager.VsqFile.Execute( run ) );
            Edited = true;
            m_mixer_dlg.UpdateStatus();
            RefreshScreen();
        }

        private void ChangeTrackNameCore() {
            if ( m_txtbox_track_name != null ) {
                if ( !m_txtbox_track_name.IsDisposed ) {
                    m_txtbox_track_name.Dispose();
                }
                m_txtbox_track_name = null;
            }
            m_txtbox_track_name = new TextBoxEx();
            m_txtbox_track_name.Visible = false;
            int selector_width = trackSelector.SelectorWidth;
            int x = 67 + (m_manager.Selected - 1) * selector_width;
            m_txtbox_track_name.Location = new Point( x, trackSelector.Height - TrackSelector._OFFSET_TRACK_TAB + 1 );
            m_txtbox_track_name.BorderStyle = BorderStyle.None;
            m_txtbox_track_name.Text = m_manager.VsqFile.Tracks[m_manager.Selected].Name;
            m_txtbox_track_name.KeyUp += new KeyEventHandler( m_txtbox_track_name_KeyUp );
            m_txtbox_track_name.Size = new Size( selector_width, TrackSelector._OFFSET_TRACK_TAB );
            m_txtbox_track_name.Parent = trackSelector;
            m_txtbox_track_name.Visible = true;
            m_txtbox_track_name.Focus();
            m_txtbox_track_name.SelectAll();
        }


        private void DeleteTrackCore() {
            int selected = m_manager.Selected;
            if ( MessageBox.Show(
                    string.Format( _( "Do you wish to remove track? {0} : \"{1}\"" ), selected, m_manager.VsqFile.Tracks[selected].Name ),
                    _APP_NAME,
                    MessageBoxButtons.YesNo,
                    MessageBoxIcon.Question ) == DialogResult.Yes ) {
                CadenciiCommand run = new CadenciiCommand( VsqCommand.GCommandDeleteTrack( m_manager.Selected ) );
                if ( m_manager.Selected >= 2 ) {
                    m_manager.Selected -= 1;
                }
                m_manager.Register( m_manager.VsqFile.Execute( run ) );
                UpdateDrawObjectList();
                Edited = true;
                m_mixer_dlg.UpdateStatus();
                RefreshScreen();
            }
        }

        private void AddTrackCore() {
            int i = m_manager.VsqFile.Tracks.Count;
            string name = "Voice" + i;
            string singer = "Miku";
            CadenciiCommand run = new CadenciiCommand(
                VsqCommand.GCommandAddTrack( new VsqTrack( name, singer ), new VsqMixerEntry( 0, 0, 0, 0 ), i ) );
            m_manager.Register( m_manager.VsqFile.Execute( run ) );
            UpdateDrawObjectList();
            Edited = true;
            m_manager.Selected = i;
            m_mixer_dlg.UpdateStatus();
            RefreshScreen();
        }
        #endregion

        /// <summary>
        /// length, positionの各Quantizeモードに応じて、表示状態を更新します
        /// </summary>
        private void ApplyQuantizeMode() {
            cMenuPianoQuantize04.CheckState = CheckState.Unchecked;
            cMenuPianoQuantize08.CheckState = CheckState.Unchecked;
            cMenuPianoQuantize16.CheckState = CheckState.Unchecked;
            cMenuPianoQuantize32.CheckState = CheckState.Unchecked;
            cMenuPianoQuantize64.CheckState = CheckState.Unchecked;
            cMenuPianoQuantizeOff.CheckState = CheckState.Unchecked;

            stripDDBtnQuantize04.CheckState = CheckState.Unchecked;
            stripDDBtnQuantize08.CheckState = CheckState.Unchecked;
            stripDDBtnQuantize16.CheckState = CheckState.Unchecked;
            stripDDBtnQuantize32.CheckState = CheckState.Unchecked;
            stripDDBtnQuantize64.CheckState = CheckState.Unchecked;
            stripDDBtnQuantizeOff.CheckState = CheckState.Unchecked;

            menuSettingPositionQuantize04.CheckState = CheckState.Unchecked;
            menuSettingPositionQuantize08.CheckState = CheckState.Unchecked;
            menuSettingPositionQuantize16.CheckState = CheckState.Unchecked;
            menuSettingPositionQuantize32.CheckState = CheckState.Unchecked;
            menuSettingPositionQuantize64.CheckState = CheckState.Unchecked;
            menuSettingPositionQuantizeOff.CheckState = CheckState.Unchecked;

            stripDDBtnQuantize.Text = "QUANTIZE " + StringFromQuantizeMode( AppManager.EditorConfig.PositionQuantize );
            switch ( AppManager.EditorConfig.PositionQuantize ) {
                case QuantizeMode.p4:
                    cMenuPianoQuantize04.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantize04.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantize04.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p8:
                    cMenuPianoQuantize08.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantize08.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantize08.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p16:
                    cMenuPianoQuantize16.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantize16.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantize16.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p32:
                    cMenuPianoQuantize32.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantize32.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantize32.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p64:
                    cMenuPianoQuantize64.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantize64.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantize64.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.off:
                    cMenuPianoQuantizeOff.CheckState = CheckState.Indeterminate;
                    stripDDBtnQuantizeOff.CheckState = CheckState.Indeterminate;
                    menuSettingPositionQuantizeOff.CheckState = CheckState.Indeterminate;
                    break;
            }
            cMenuPianoQuantizeTriplet.Checked = AppManager.EditorConfig.PositionQuantizeTriplet;
            stripDDBtnQuantizeTriplet.Checked = AppManager.EditorConfig.PositionQuantizeTriplet;
            menuSettingPositionQuantizeTriplet.Checked = AppManager.EditorConfig.PositionQuantizeTriplet;

            cMenuPianoLength04.CheckState = CheckState.Unchecked;
            cMenuPianoLength08.CheckState = CheckState.Unchecked;
            cMenuPianoLength16.CheckState = CheckState.Unchecked;
            cMenuPianoLength32.CheckState = CheckState.Unchecked;
            cMenuPianoLength64.CheckState = CheckState.Unchecked;
            cMenuPianoLengthOff.CheckState = CheckState.Unchecked;

            stripDDBtnLength04.CheckState = CheckState.Unchecked;
            stripDDBtnLength08.CheckState = CheckState.Unchecked;
            stripDDBtnLength16.CheckState = CheckState.Unchecked;
            stripDDBtnLength32.CheckState = CheckState.Unchecked;
            stripDDBtnLength64.CheckState = CheckState.Unchecked;
            stripDDBtnLengthOff.CheckState = CheckState.Unchecked;

            menuSettingLengthQuantize04.CheckState = CheckState.Unchecked;
            menuSettingLengthQuantize08.CheckState = CheckState.Unchecked;
            menuSettingLengthQuantize16.CheckState = CheckState.Unchecked;
            menuSettingLengthQuantize32.CheckState = CheckState.Unchecked;
            menuSettingLengthQuantize64.CheckState = CheckState.Unchecked;
            menuSettingLengthQuantizeOff.CheckState = CheckState.Unchecked;

            stripDDBtnLength.Text = "LENGTH " + StringFromQuantizeMode( AppManager.EditorConfig.LengthQuantize );
            switch ( AppManager.EditorConfig.LengthQuantize ) {
                case QuantizeMode.p4:
                    cMenuPianoLength04.CheckState = CheckState.Indeterminate;
                    stripDDBtnLength04.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantize04.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p8:
                    cMenuPianoLength08.CheckState = CheckState.Indeterminate;
                    stripDDBtnLength08.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantize08.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p16:
                    cMenuPianoLength16.CheckState = CheckState.Indeterminate;
                    stripDDBtnLength16.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantize16.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p32:
                    cMenuPianoLength32.CheckState = CheckState.Indeterminate;
                    stripDDBtnLength32.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantize32.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.p64:
                    cMenuPianoLength64.CheckState = CheckState.Indeterminate;
                    stripDDBtnLength64.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantize64.CheckState = CheckState.Indeterminate;
                    break;
                case QuantizeMode.off:
                    cMenuPianoLengthOff.CheckState = CheckState.Indeterminate;
                    stripDDBtnLengthOff.CheckState = CheckState.Indeterminate;
                    menuSettingLengthQuantizeOff.CheckState = CheckState.Indeterminate;
                    break;
            }
            cMenuPianoLengthTriplet.Checked = AppManager.EditorConfig.LengthQuantizeTriplet;
            stripDDBtnLengthTriplet.Checked = AppManager.EditorConfig.LengthQuantizeTriplet;
            menuSettingLengthQuantizeTriplet.Checked = AppManager.EditorConfig.LengthQuantizeTriplet;
        }

        /// <summary>
        /// 現在選択されている編集ツールに応じて、メニューのチェック状態を更新します
        /// </summary>
        private void ApplySelectedTool() {
            switch ( m_manager.SelectedTool ) {
                case EditTool.Arrow:
                    cMenuPianoPointer.CheckState = CheckState.Indeterminate;
                    cMenuPianoPencil.CheckState = CheckState.Unchecked;
                    cMenuPianoEraser.CheckState = CheckState.Unchecked;

                    cMenuTrackSelectorPointer.CheckState = CheckState.Indeterminate;
                    cMenuTrackSelectorPencil.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorLine.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorEraser.CheckState = CheckState.Unchecked;

                    stripBtnPointer.Checked = true;
                    stripBtnPencil.Checked = false;
                    stripBtnLine.Checked = false;
                    stripBtnEraser.Checked = false;
                    break;
                case EditTool.Pencil:
                    cMenuPianoPointer.CheckState = CheckState.Unchecked;
                    cMenuPianoPencil.CheckState = CheckState.Indeterminate;
                    cMenuPianoEraser.CheckState = CheckState.Unchecked;

                    cMenuTrackSelectorPointer.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorPencil.CheckState = CheckState.Indeterminate;
                    cMenuTrackSelectorLine.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorEraser.CheckState = CheckState.Unchecked;

                    stripBtnPointer.Checked = false;
                    stripBtnPencil.Checked = true;
                    stripBtnLine.Checked = false;
                    stripBtnEraser.Checked = false;
                    break;
                case EditTool.Eraser:
                    cMenuPianoPointer.CheckState = CheckState.Unchecked;
                    cMenuPianoPencil.CheckState = CheckState.Unchecked;
                    cMenuPianoEraser.CheckState = CheckState.Indeterminate;

                    cMenuTrackSelectorPointer.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorPencil.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorLine.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorEraser.CheckState = CheckState.Indeterminate;

                    stripBtnPointer.Checked = false;
                    stripBtnPencil.Checked = false;
                    stripBtnLine.Checked = false;
                    stripBtnEraser.Checked = true;
                    break;
                case EditTool.Line:
                    cMenuPianoPointer.CheckState = CheckState.Unchecked;
                    cMenuPianoPencil.CheckState = CheckState.Unchecked;
                    cMenuPianoEraser.CheckState = CheckState.Unchecked;

                    cMenuTrackSelectorPointer.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorPencil.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorLine.CheckState = CheckState.Indeterminate;
                    cMenuTrackSelectorEraser.CheckState = CheckState.Unchecked;

                    stripBtnPointer.Checked = false;
                    stripBtnPencil.Checked = false;
                    stripBtnLine.Checked = true;
                    stripBtnEraser.Checked = false;
                    break;
                default:
                    cMenuPianoPointer.CheckState = CheckState.Unchecked;
                    cMenuPianoPencil.CheckState = CheckState.Unchecked;
                    cMenuPianoEraser.CheckState = CheckState.Unchecked;

                    cMenuTrackSelectorPointer.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorPencil.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorLine.CheckState = CheckState.Unchecked;
                    cMenuTrackSelectorEraser.CheckState = CheckState.Unchecked;

                    stripBtnPointer.Checked = false;
                    stripBtnPencil.Checked = false;
                    stripBtnLine.Checked = false;
                    stripBtnEraser.Checked = false;
                    break;
            }
            cMenuPianoCurve.Checked = m_manager.IsCurveMode;
            cMenuTrackSelectorCurve.Checked = m_manager.IsCurveMode;
            stripBtnCurve.Checked = m_manager.IsCurveMode;
        }

        /// <summary>
        /// 画面上のマウス位置におけるクロック値を元に，_toolbar_measureの場所表示文字列を更新します．
        /// </summary>
        /// <param name="mouse_pos_x"></param>
        private void UpdatePositionViewFromMousePosition( int clock ) {
            int barcount = m_manager.VsqFile.GetBarCountFromClock( clock );
            int numerator, denominator;
            m_manager.VsqFile.GetTimesigAt( clock, out numerator, out denominator );
            int clock_per_beat = 480 / 4 * denominator;
            int barcount_clock = m_manager.VsqFile.GetClockFromBarCount( barcount );
            int beat = (clock - barcount_clock) / clock_per_beat;
            int odd = clock - barcount_clock - beat * clock_per_beat;
#if OBSOLUTE
            m_toolbar_measure.Measure = (barcount - AppManager.VsqFile.PreMeasure + 1) + " : " + (beat + 1) + " : " + odd.ToString( "000" );
#else
            stripLblMeasure.Text = (barcount - m_manager.VsqFile.PreMeasure + 1) + " : " + (beat + 1) + " : " + odd.ToString( "000" );
#endif
        }

        /// <summary>
        /// 描画すべきオブジェクトのリスト，m_draw_objectsを更新します
        /// </summary>
        private void UpdateDrawObjectList() {
            if ( m_manager.VsqFile == null ) {
                return;
            }
            if ( m_draw_objects != null ) {
                foreach ( List<DrawObject> list in m_draw_objects ) {
                    list.Clear();
                }
                m_draw_objects.Clear();
            }
            m_draw_objects = new List<List<DrawObject>>();

            int xoffset = 73;
            int yoffset = 127 * AppManager.EditorConfig.PxTrackHeight;
            float scalex = ScaleX;
            using ( Font SMALL_FONT = new Font( AppManager.EditorConfig.ScreenFontName, 8 ) ) {
                for ( int track = 1; track < m_manager.VsqFile.Tracks.Count; track++ ) {
                    List<DrawObject> tmp = new List<DrawObject>();
                    foreach ( VsqEvent ev in m_manager.VsqFile.Tracks[track].Events ) {
                        int timesig = ev.Clock;
                        if ( ev.ID.LyricHandle != null ) {
                            int length = ev.ID.Length;
                            int note = ev.ID.Note;
                            int x = (int)(timesig * scalex + xoffset);
                            int y = -note * AppManager.EditorConfig.PxTrackHeight + yoffset;
                            int lyric_width = (int)(length * scalex);
                            string lyric_jp = ev.ID.LyricHandle.L0.Phrase;
                            string lyric_en = ev.ID.LyricHandle.L0.PhoneticSymbol;
                            string title = Common.TrimString( lyric_jp + " [" + lyric_en + "]", SMALL_FONT, lyric_width );
                            int accent = ev.ID.DEMaccent;
                            int vibrato_start = x + lyric_width;
                            int vibrato_end = x;
                            int vibrato_delay = lyric_width * 2;
                            if ( ev.ID.VibratoHandle != null ) {
                                double rate = (double)ev.ID.VibratoDelay / (double)length;
                                vibrato_delay = _PX_ACCENT_HEADER + (int)((lyric_width - _PX_ACCENT_HEADER) * rate);
                            }
                            tmp.Add( new DrawObject( new Rectangle( x, y, lyric_width, AppManager.EditorConfig.PxTrackHeight ),
                                                     title,
                                                     accent,
                                                     ev.InternalID,
                                                     vibrato_delay,
                                                     false ) );
                        }
                    }

                    // 重複部分があるかどうかを判定
                    for ( int i = 0; i < tmp.Count - 1; i++ ) {
                        bool overwrapped = false;
                        for ( int j = i + 1; j < tmp.Count; j++ ) {
                            int startx = tmp[j].pxRectangle.X;
                            int endx = tmp[j].pxRectangle.X + tmp[j].pxRectangle.Width;
                            if ( startx < tmp[i].pxRectangle.X ) {
                                if ( tmp[i].pxRectangle.X < endx ) {
                                    overwrapped = true;
                                    tmp[j].Overwrapped = true;
                                    // breakできない．2個以上の重複を検出する必要があるので．
                                }
                            } else if ( tmp[i].pxRectangle.X <= startx && startx < tmp[i].pxRectangle.X + tmp[i].pxRectangle.Width ) {
                                overwrapped = true;
                                tmp[j].Overwrapped = true;
                            }
                        }
                        if ( overwrapped ) {
                            tmp[i].Overwrapped = true;
                        }
                    }
                    m_draw_objects.Add( tmp );
                }
            }
        }

        private void DrawTo( Graphics g, Size window_size, Point mouse_position ) {
            int start_draw_x = StartToDrawX;
            int start_draw_y = StartToDrawY;
            // [screen_x] = 67 + [clock] * ScaleX - StartToDrawX + 6
            // [screen_y] = -1 * ([note] - 127) * TRACK_HEIGHT - StartToDrawY
            //
            // [screen_x] = [clock] * _scalex + 73 - StartToDrawX
            // [screen_y] = -[note] * TRACK_HEIGHT + 127 * TRACK_HEIGHT - StartToDrawY
            int xoffset = 73 - start_draw_x;
            int yoffset = 127 * AppManager.EditorConfig.PxTrackHeight - start_draw_y;
            //      ↓
            // [screen_x] = [clock] * _scalex + xoffset
            // [screen_y] = -[note] * TRACK_HEIGHT + yoffset
            int y, dy;
            float scalex = m_scalex;
            float inv_scalex = 1f / scalex;

            if ( m_manager.SelectedEvent.Count > 0 && m_input_textbox != null ) {
                VsqEvent original = m_manager.LastSelectedEvent.Original;
                int event_x = (int)(original.Clock * scalex + xoffset);
                int event_y = -original.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset;
                m_input_textbox.Left = event_x + 4;
                m_input_textbox.Top = event_y + 2;
            }

            int width = window_size.Width;
            int height = window_size.Height;

            #region ピアノロール周りのスクロールバーなど
            // スクロール画面背景
            int calculated_height = height;
            if ( calculated_height > 0 ) {
                g.FillRectangle( s_brs_240_240_240,
                                 new Rectangle( 3, 0, width, calculated_height ) );
            }
            // ピアノロールとカーブエディタの境界
            g.DrawLine( s_pen_112_112_112,
                        new Point( 2, height - 1 ),
                        new Point( width - 1, height - 1 ) );
            // 横スクロールバー
            //hScroll.BarLength = (int)((width - 152) * inv_scalex);
            // 縦スクロールバー
            calculated_height = height - 14;//height - 93 - 14;
            if ( calculated_height > 0 ) {
                //vScroll.BarLength = (int)((float)vScroll.Maximum * vScroll.Height / (AppManager.EditorConfig.PxTrackHeight * 128f));
            }
            #endregion

            #region ピアノロール本体
            if ( m_manager.VsqFile != null ) {
                int odd = -1;
                y = 128 * AppManager.EditorConfig.PxTrackHeight - start_draw_y;
                dy = -AppManager.EditorConfig.PxTrackHeight;
                for ( int i = 0; i <= 127; i++ ) {
                    odd++;
                    if ( odd == 12 ) {
                        odd = 0;
                    }
                    int order = (i - odd) / 12 - 2;
                    y += dy;
                    if ( y > height ) {
                        continue;
                    } else if ( 0 > y + AppManager.EditorConfig.PxTrackHeight ) {
                        break;
                    }
                    bool note_is_whitekey = VsqNote.NoteIsWhiteKey( i );

                    #region ピアノロール背景
                    SolidBrush b = new SolidBrush( Color.Black );
                    SolidBrush border;
                    bool paint_required = true;
                    if ( order == -2 || order == -1 || (6 <= order && order <= 8) ) {
                        if ( note_is_whitekey ) {
                            b = s_brs_180_180_180;
                        } else {
                            b = s_brs_106_108_108;
                        }
                        border = s_brs_106_108_108;
                    } else if ( order == 5 || order == 0 ) {
                        if ( note_is_whitekey ) {
                            b = s_brs_212_212_212;
                        } else {
                            b = s_brs_180_180_180;
                        }
                        border = new SolidBrush( Color.FromArgb( 150, 152, 150 ) );
                    } else {
                        if ( note_is_whitekey ) {
                            paint_required = false;
                        } else {
                            b = s_brs_212_212_212;
                        }
                        border = new SolidBrush( Color.FromArgb( 210, 205, 172 ) );
                    }
                    if ( paint_required ) {
                        g.FillRectangle( b,
                                         new Rectangle( 68, y, width - 68, AppManager.EditorConfig.PxTrackHeight + 1 ) );
                    }
                    if ( odd == 0 || odd == 5 ) {
                        g.DrawLine( new Pen( border ),
                                    new Point( 68, y + AppManager.EditorConfig.PxTrackHeight ),
                                    new Point( width, y + AppManager.EditorConfig.PxTrackHeight ) );
                    }
                    #endregion

                    #region プリメジャー部分のピアノロール背景
                    int premeasure_start_x = xoffset;
                    if ( premeasure_start_x < 68 ) {
                        premeasure_start_x = 68;
                    }
                    int premeasure_end_x = (int)(m_manager.VsqFile.PreMeasureClocks * scalex + xoffset);
                    if ( premeasure_start_x >= 68 ) {
                        if ( note_is_whitekey ) {
                            g.FillRectangle( s_brs_153_153_153,
                                             new Rectangle( premeasure_start_x,
                                                            y,
                                                            premeasure_end_x - premeasure_start_x,
                                                            AppManager.EditorConfig.PxTrackHeight + 1 ) );
                        } else {
                            g.FillRectangle( s_brs_106_108_108,
                                             new Rectangle( premeasure_start_x,
                                                            y,
                                                            premeasure_end_x - premeasure_start_x,
                                                            AppManager.EditorConfig.PxTrackHeight + 1 ) );
                        }
                        if ( odd == 0 || odd == 5 ) {
                            g.DrawLine( s_pen_106_108_108,
                                        new Point( premeasure_start_x, y + AppManager.EditorConfig.PxTrackHeight ),
                                        new Point( premeasure_end_x, y + AppManager.EditorConfig.PxTrackHeight ) );
                        }
                    }
                    #endregion

                }
            }

            //ピアノロールと鍵盤部分の縦線
            g.DrawLine( s_pen_212_212_212,
                        new Point( 67, 0 ),
                        new Point( 67, height - 15 ) );
            int odd2 = -1;
            y = 128 * AppManager.EditorConfig.PxTrackHeight - start_draw_y;
            dy = -AppManager.EditorConfig.PxTrackHeight;
            for ( int i = 0; i <= 127; i++ ) {
                odd2++;
                if ( odd2 == 12 ) {
                    odd2 = 0;
                }
                y += dy;
                if ( y > height ) {
                    continue;
                } else if ( y + AppManager.EditorConfig.PxTrackHeight < 0 ) {
                    break;
                }

                #region 鍵盤部分
                g.DrawLine(
                    s_pen_212_212_212,
                    new Point( 3, y ),
                    new Point( 66, y ) );
                bool hilighted = false;
                if ( m_manager.EditMode == EditMode.AddEntry ) {
                    if ( m_adding.ID.Note == i ) {
                        hilighted = true;
                    }
                } else if ( m_manager.EditMode == EditMode.EditLeftEdge || m_manager.EditMode == EditMode.EditRightEdge ) {
                    if ( m_manager.LastSelectedEvent.Original.ID.Note == i ) {
                        hilighted = true;
                    }
                } else {
                    if ( 3 <= mouse_position.X && mouse_position.X <= width - 17 &&
                        0 <= mouse_position.Y && mouse_position.Y <= height - 1 ) {
                        if ( y <= mouse_position.Y && mouse_position.Y < y + AppManager.EditorConfig.PxTrackHeight ) {
                            hilighted = true;
                        }
                    }
                }
                if ( hilighted ) {
                    g.FillRectangle( AppManager.HilightBrush,
                                     new Rectangle( 35, y, 33, AppManager.EditorConfig.PxTrackHeight ) );
                }
                if ( odd2 == 0 || hilighted ) {
                    g.DrawString( VsqNote.NoteToString( i ),
                                  s_font_small,
                                  s_brs_072_077_098,
                                  new PointF( 42, y + 2 ) );
                }
                if ( !VsqNote.NoteIsWhiteKey( i ) ) {
                    g.FillRectangle( s_brs_125_123_124, new Rectangle( 0, y, 34, AppManager.EditorConfig.PxTrackHeight ) );
                }
                #endregion
            }
            g.ResetClip();

            g.SetClip( new Rectangle( 68, 0, width - 68, height ) );
            #region 小節ごとの線
            IEnumerable<VsqBarLineType> enumerable = m_manager.VsqFile.GetEnumerator( ClockFromXCoord( width ) );
            int dashed_line_step = GetUnitClocksOfQuantize( AppManager.EditorConfig.PositionQuantize,
                                                            AppManager.EditorConfig.PositionQuantizeTriplet );
            foreach ( VsqBarLineType blt in enumerable ) {
                int local_clock_step = 1920 / blt.LocalDenominator;
                int x = (int)(blt.Clock * scalex + xoffset);
                if ( blt.IsSeparator ) {
                    //ピアノロール上
                    g.DrawLine( s_pen_161_157_136,
                                new Point( x, 0 ),
                                new Point( x, height ) );
                } else {
                    //ピアノロール上
                    g.DrawLine( s_pen_209_204_172,
                                new Point( x, 0 ),
                                new Point( x, height ) );
                }
                if ( dashed_line_step > 1 && m_manager.GridVisible ) {
                    int numDashedLine = local_clock_step / dashed_line_step;
                    for ( int i = 1; i < numDashedLine; i++ ) {
                        int x2 = (int)((blt.Clock + i * dashed_line_step) * scalex + xoffset);
                        g.DrawLine( s_pen_dashed_209_204_172,
                                    new Point( x2, 0 ),
                                    new Point( x2, height ) );
                    }
                }
            }
            #endregion

            #region トラックのエントリを描画
            if ( m_draw_objects != null ) {
                if ( m_manager.Overlay ) {
                    // まず、選択されていないトラックの簡易表示を行う
                    for ( int i = 0; i < m_draw_objects.Count; i++ ) {
                        if ( i == m_manager.Selected - 1 ) {
                            continue;
                        }
                        int j_start = m_draw_start_index[i];
                        bool first = true;
                        for ( int j = j_start; j < m_draw_objects[i].Count; j++ ) {
                            DrawObject dobj = m_draw_objects[i][j];
                            int x = dobj.pxRectangle.X - start_draw_x;
                            y = dobj.pxRectangle.Y - start_draw_y;
                            int lyric_width = dobj.pxRectangle.Width;
                            if ( x + lyric_width < 0 ) {
                                continue;
                            } else if ( width < x ) {
                                break;
                            }
                            if ( m_manager.Playing && first ) {
                                m_draw_start_index[i] = j;
                                first = false;
                            }
                            if ( y + AppManager.EditorConfig.PxTrackHeight < 0 || y > height ) {
                                continue;
                            }
                            g.DrawLine( new Pen( AppManager.s_HILIGHT[i] ),
                                        new Point( x + 1, y + 7 ),
                                        new Point( x + lyric_width - 1, y + 7 ) );
                            g.DrawPolygon( new Pen( s_HIDDEN[i] ),
                                           new Point[] { 
                                               new Point( x, y + 7 ), 
                                               new Point( x + 1, y + 6 ),
                                               new Point( x + lyric_width - 1, y + 6 ),
                                               new Point( x + lyric_width, y + 7 ),
                                               new Point( x + lyric_width - 1, y + 8 ),
                                               new Point( x + 1, y + 8 ), 
                                               new Point( x, y + 7 ) 
                                           } );
                        }
                    }
                }

                // 選択されているトラックの表示を行う
                int selected = m_manager.Selected;
                bool show_lyrics = AppManager.EditorConfig.ShowLyric;
                bool show_exp_line = AppManager.EditorConfig.ShowExpLine;
                if ( selected >= 1 ) {
                    int j_start = m_draw_start_index[selected - 1];
                    bool first = true;
                    for ( int j = j_start; j < m_draw_objects[selected - 1].Count; j++ ) {
                        DrawObject dobj = m_draw_objects[selected - 1][j];
                        int x = dobj.pxRectangle.X - start_draw_x;
                        y = dobj.pxRectangle.Y - start_draw_y;
                        int lyric_width = dobj.pxRectangle.Width;
                        if ( x + lyric_width < 0 ) {
                            continue;
                        } else if ( width < x ) {
                            break;
                        }
                        if ( m_manager.Playing && first ) {
                            m_draw_start_index[selected - 1] = j;
                            first = false;
                        }
                        if ( y + 2 * AppManager.EditorConfig.PxTrackHeight < 0 || y > height ) {
                            continue;
                        }
                        Color id_fill;
                        if ( m_manager.SelectedEvent.Count > 0 ) {
                            bool found = m_manager.SelectedEvent.ContainsKey( dobj.InternalID );
                            if ( found ) {
                                id_fill = AppManager.HilightColor;
                            } else {
                                id_fill = Color.FromArgb( 181, 220, 86 );
                            }
                        } else {
                            id_fill = Color.FromArgb( 181, 220, 86 );
                        }
                        g.FillRectangle(
                            new SolidBrush( id_fill ),
                            new Rectangle( x, y + 1, lyric_width, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                        if ( dobj.Overwrapped ) {
                            g.DrawRectangle( s_pen_125_123_124,
                                             new Rectangle( x, y + 1, lyric_width, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                            if ( show_lyrics ) {
                                g.DrawString( dobj.Text,
                                              s_font_entry,
                                              s_brs_147_147_147,
                                              new PointF( x + 1, y + 1 ) );
                            }
                        } else {
                            g.DrawRectangle( s_pen_125_123_124,
                                             new Rectangle( x, y + 1, lyric_width, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                            if ( show_lyrics ) {
                                g.DrawString( dobj.Text,
                                              s_font_entry,
                                              Brushes.Black,
                                              new PointF( x + 1, y + 1 ) );
                            }
                            if ( show_exp_line && lyric_width > 21 ) {
                                #region 表情線
                                DrawAccentLine( g, new Point( x, y + AppManager.EditorConfig.PxTrackHeight + 1 ), dobj.Accent );
                                int vibrato_start = x + lyric_width;
                                int vibrato_end = x;
                                if ( dobj.pxVibratoDelay <= lyric_width ) {
                                    int vibrato_delay = dobj.pxVibratoDelay;
                                    int vibrato_width = dobj.pxRectangle.Width - vibrato_delay;
                                    vibrato_start = x + vibrato_delay;
                                    vibrato_end = x + vibrato_delay + vibrato_width;
                                    if ( vibrato_start - x < 21 ) {
                                        vibrato_start = x + 21;
                                    }
                                }
                                g.DrawLine( s_pen_051_051_000,
                                            new Point( x + 21, y + AppManager.EditorConfig.PxTrackHeight + 7 ),
                                            new Point( vibrato_start, y + AppManager.EditorConfig.PxTrackHeight + 7 ) );
                                if ( dobj.pxVibratoDelay <= lyric_width ) {
                                    int next_draw = vibrato_start;
                                    if ( vibrato_start < vibrato_end ) {
                                        DrawVibratoLine( g, 
                                                         new Point( vibrato_start, y + AppManager.EditorConfig.PxTrackHeight + 1 ), 
                                                         vibrato_end - vibrato_start );
                                    }
                                #endregion
                                }
                            }
                        }

                    }
                }

                // 編集中のエントリを表示
                if ( m_manager.EditMode == EditMode.AddEntry ||
                     m_manager.EditMode == EditMode.AddFixedLengthEntry ) {
                    int x = (int)(m_adding.Clock * scalex + xoffset);
                    y = -m_adding.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                    if ( m_adding.ID.Length <= 0 ) {
                        g.DrawRectangle( s_pen_dashed_171_171_171,
                                         new Rectangle( x, y, 10, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                    } else {
                        int length = (int)(m_adding.ID.Length * scalex);
                        g.DrawRectangle( s_pen_a136_000_000_000,
                                         new Rectangle( x, y, length, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                    }
                } else if ( m_manager.EditMode == EditMode.EditVibratoDelay ) {
                    int x = (int)(m_adding.Clock * scalex + xoffset);
                    y = -m_adding.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                    int length = (int)(m_adding.ID.Length * scalex);
                    g.DrawRectangle( s_pen_a136_000_000_000,
                                     new Rectangle( x, y, length, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                } else if ( (m_manager.EditMode == EditMode.MoveEntry ||
                       m_manager.EditMode == EditMode.EditLeftEdge ||
                       m_manager.EditMode == EditMode.EditRightEdge) && m_manager.SelectedEvent.Count > 0 ) {
                    foreach ( KeyValuePair<int, SelectedEventEntry> ev in m_manager.SelectedEvent ) {
                        int x = (int)(ev.Value.Editing.Clock * scalex + xoffset);
                        y = -ev.Value.Editing.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                        if ( ev.Value.Editing.ID.Length == 0 ) {
                            g.DrawRectangle( s_pen_dashed_171_171_171,
                                             new Rectangle( x, y, 10, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                        } else {
                            int length = (int)(ev.Value.Editing.ID.Length * scalex);
                            g.DrawRectangle( s_pen_a136_000_000_000,
                                             new Rectangle( x, y, length, AppManager.EditorConfig.PxTrackHeight - 1 ) );
                        }
                    }
                }
            }
            #endregion

            g.ResetClip();

            #endregion

            #region 現在のマーカー
            int marker_x = (int)(m_manager.CurrentClock * scalex + xoffset);
            if ( 67 <= marker_x && marker_x <= width ) {
                g.DrawLine( new Pen( Color.White, 2f ),
                            new Point( marker_x, 0 ),
                            new Point( marker_x, height ) );
            }
            #endregion

            #region 音符編集時の補助線
            if ( m_manager.EditMode == EditMode.AddEntry ) {
                #region EditMode.AddEntry
                int x = (int)(m_adding.Clock * scalex + xoffset);
                y = -m_adding.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                int length;
                if ( m_adding.ID.Length == 0 ) {
                    length = 10;
                } else {
                    length = (int)(m_adding.ID.Length * scalex);
                }
                x += length;
                g.DrawLine( s_pen_LU,
                            new Point( x, 0 ),
                            new Point( x, y - 1 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x, height ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, 0 ),
                            new Point( x + 1, y - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x + 1, height ) );
                #endregion
            } else if ( m_manager.EditMode == EditMode.MoveEntry ) {
                #region EditMode.MoveEntry
                if ( m_manager.SelectedEvent.Count > 0 ) {
                    VsqEvent last = m_manager.LastSelectedEvent.Editing;
                    int x = (int)(last.Clock * scalex + xoffset);
                    y = -last.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                    int length = (int)(last.ID.Length * scalex);
                    // 縦線
                    g.DrawLine( s_pen_LU,
                                new Point( x, 0 ),
                                new Point( x, y - 1 ) );
                    g.DrawLine( s_pen_LU,
                                new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                                new Point( x, height ) );
                    // 横線
                    g.DrawLine( s_pen_LU,
                                new Point( 68, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ),
                                new Point( x - 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ) );
                    g.DrawLine( s_pen_LU,
                                new Point( x + length + 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ),
                                new Point( width, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ) );
                    // 縦線
                    g.DrawLine( s_pen_RD,
                                new Point( x + 1, 0 ),
                                new Point( x + 1, y - 1 ) );
                    g.DrawLine( s_pen_RD,
                                new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                                new Point( x + 1, height ) );
                    // 横線
                    g.DrawLine( s_pen_RD,
                                new Point( 68, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ),
                                new Point( x - 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ) );
                    g.DrawLine( s_pen_RD,
                                new Point( x + length + 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ),
                                new Point( width, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ) );
                }
                #endregion
            } else if ( m_manager.EditMode == EditMode.AddFixedLengthEntry ) {
                #region EditMode.MoveEntry
                int x = (int)(m_adding.Clock * scalex + xoffset);
                y = -m_adding.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                int length = (int)(m_adding.ID.Length * scalex);
                // 縦線
                g.DrawLine( s_pen_LU,
                            new Point( x, 0 ),
                            new Point( x, y - 1 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x, height ) );
                // 横線
                g.DrawLine( s_pen_LU,
                            new Point( 68, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ),
                            new Point( x - 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x + length + 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ),
                            new Point( width, y + AppManager.EditorConfig.PxTrackHeight / 2 - 2 ) );
                // 縦線
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, 0 ),
                            new Point( x + 1, y - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x + 1, height ) );
                // 横線
                g.DrawLine( s_pen_RD,
                            new Point( 68, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ),
                            new Point( x - 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + length + 1, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ),
                            new Point( width, y + AppManager.EditorConfig.PxTrackHeight / 2 - 1 ) );
                #endregion
            } else if ( m_manager.EditMode == EditMode.EditLeftEdge ) {
                #region EditMode.EditLeftEdge
                VsqEvent last = m_manager.LastSelectedEvent.Editing;
                int x = (int)(last.Clock * scalex + xoffset);
                y = -last.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                g.DrawLine( s_pen_LU,
                            new Point( x, 0 ),
                            new Point( x, y - 1 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x, height ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, 0 ),
                            new Point( x + 1, y - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x + 1, height ) );
                #endregion
            } else if ( m_manager.EditMode == EditMode.EditRightEdge ) {
                #region EditMode.EditRightEdge
                VsqEvent last = m_manager.LastSelectedEvent.Editing;
                int x = (int)(last.Clock * scalex + xoffset);
                y = -last.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                int length = (int)(last.ID.Length * scalex);
                x += length;
                g.DrawLine( s_pen_LU,
                            new Point( x, 0 ),
                            new Point( x, y - 1 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x, height ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, 0 ),
                            new Point( x + 1, y - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x + 1, height ) );
                #endregion
            } else if ( m_manager.EditMode == EditMode.EditVibratoDelay ) {
                #region EditVibratoDelay
                int x = (int)(m_adding.Clock * scalex + xoffset);
                y = -m_adding.ID.Note * AppManager.EditorConfig.PxTrackHeight + yoffset + 1;
                g.DrawLine( s_pen_LU,
                            new Point( x, 0 ),
                            new Point( x, y - 1 ) );
                g.DrawLine( s_pen_LU,
                            new Point( x, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x, height ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, 0 ),
                            new Point( x + 1, y - 1 ) );
                g.DrawLine( s_pen_RD,
                            new Point( x + 1, y + AppManager.EditorConfig.PxTrackHeight ),
                            new Point( x + 1, height ) );
                double max_length = m_adding.ID.value - _PX_ACCENT_HEADER / scalex;
                double drate = m_adding.ID.Length / max_length;
                if ( drate > 0.99 ) {
                    drate = 1.00;
                }
                int rate = (int)(drate * 100.0);
                string percent = rate + "%";
                SizeF size = Common.MeasureString( percent, s_F9PT );
                int delay_x = (int)((m_adding.Clock + m_adding.ID.Length - m_adding.ID.value + m_adding.ID.VibratoDelay) * scalex + xoffset);
                Rectangle pxArea = new Rectangle( delay_x,
                                                  (int)(y + AppManager.EditorConfig.PxTrackHeight * 2.5),
                                                  (int)(size.Width * 1.2),
                                                  (int)(size.Height * 1.2) );
                g.FillRectangle( s_brs_192_192_192, pxArea );
                g.DrawRectangle( Pens.Black, pxArea );
                StringFormat sf = new StringFormat();
                sf.Alignment = StringAlignment.Center;
                sf.LineAlignment = StringAlignment.Center;
                g.DrawString( percent, s_F9PT, Brushes.Black, pxArea, sf );
                #endregion
            }
            #endregion

            #region 外枠
            // 左(外側)
            g.DrawLine( s_pen_160_160_160,
                        new Point( 0, 0 ),
                        new Point( 0, height ) );
            // 左(内側)
            g.DrawLine( s_pen_105_105_105,
                        new Point( 1, 0 ),
                        new Point( 1, height ) );
            #endregion

            trackSelector.ScaleX = scalex;
            trackSelector.StartToDraw = start_draw_x;
        }

        /// <summary>
        /// アクセントを表す表情線を、指定された位置を基準点として描き込みます
        /// </summary>
        /// <param name="g"></param>
        /// <param name="accent"></param>
        private void DrawAccentLine( Graphics g, Point origin, int accent ) {
            int x0 = origin.X + 1;
            int y0 = origin.Y + 10;
            int height = 4 + accent * 4 / 100;
            SmoothingMode sm = g.SmoothingMode;
            g.SmoothingMode = SmoothingMode.AntiAlias;
            Point[] ps = new Point[]{ new Point( x0, y0 ),
                                      new Point( x0 + 2, y0 ),
                                      new Point( x0 + 8, y0 - height ),
                                      new Point( x0 + 13, y0 ),
                                      new Point( x0 + 16, y0 ),
                                      new Point( x0 + 20, y0 - 4 ) };
            g.DrawLines( Pens.Black, ps );
            g.SmoothingMode = sm;
        }

        private void DrawVibratoLine( Graphics g, Point origin, int vibrato_length ) {
            int x0 = origin.X + 1;
            int y0 = origin.Y + 10;
            int clipx = origin.X + 1;
            int clip_length = vibrato_length;
            if ( clipx < 68 ) {
                clipx = 68;
                clip_length = origin.X + 1 + vibrato_length - 68;
                if ( clip_length <= 0 ) {
                    return;
                }
            }
            SmoothingMode sm = g.SmoothingMode;
            g.SmoothingMode = SmoothingMode.AntiAlias;
            const int _UWID = 10;
            int count = vibrato_length / _UWID + 1;
            Point[] _BASE = new Point[]{ new Point( x0 - _UWID, y0 - 4 ),
                                         new Point( x0 + 2 - _UWID, y0 - 7 ),
                                         new Point( x0 + 4 - _UWID, y0 - 7 ),
                                         new Point( x0 + 7 - _UWID, y0 - 1 ),
                                         new Point( x0 + 9 - _UWID, y0 - 1 ),
                                         new Point( x0 + 10 - _UWID, y0 - 4 ) };
            g.Clip = new Region( new Rectangle( clipx, origin.Y + 10 - 8, clip_length, 10 ) );
            for ( int i = 0; i < count; i++ ) {
                for ( int j = 0; j < _BASE.Length; j++ ) {
                    _BASE[j].X += _UWID;
                }
                g.DrawLines( Pens.Black, _BASE );
            }
            g.SmoothingMode = sm;
            g.ResetClip();
        }

        /// <summary>
        /// クロック数から、画面に描くべきx座標の値を取得します。
        /// </summary>
        /// <param name="clocks"></param>
        /// <returns></returns>
        int XCoordFromClocks( int clocks ) {
            return (int)(67 + clocks * ScaleX - StartToDrawX) + 6;
        }

        /// <summary>
        /// 画面のx座標からクロック数を取得します
        /// </summary>
        /// <param name="x"></param>
        /// <returns></returns>
        int ClockFromXCoord( int x ) {
            return (int)((x + StartToDrawX - 6 - 67) / ScaleX);
        }

        /// <summary>
        /// QuantizeModeの値から1/4などの文字列を取得します
        /// </summary>
        /// <param name="item"></param>
        /// <returns></returns>
        public static string QModeToString( QuantizeMode item ) {
            switch ( item ) {
                case QuantizeMode.off:
                    return "off";
                case QuantizeMode.p4:
                    return "1/4";
                case QuantizeMode.p8:
                    return "1/8";
                case QuantizeMode.p16:
                    return "1/16";
                case QuantizeMode.p32:
                    return "1/32";
                case QuantizeMode.p64:
                    return "1/64";
            }
            return "";
        }

        /// <summary>
        /// _editor_configのRecentFilesを元に，menuFileRecentのドロップダウンアイテムを更新します
        /// </summary>
        private void UpdateRecentFileMenu() {
            int added = 0;
            menuFileRecent.DropDownItems.Clear();
            if ( AppManager.EditorConfig.RecentFiles != null ) {
                for ( int i = 0; i < AppManager.EditorConfig.RecentFiles.Count; i++ ) {
                    string item = AppManager.EditorConfig.RecentFiles[i];
                    if ( item == null ) {
                        continue;
                    }
                    if ( item != "" ) {
                        string short_name = Path.GetFileName( item );
                        bool available = File.Exists( item );
                        ToolStripItem itm = (ToolStripItem)(new ToolStripMenuItem( short_name ));
                        if ( !available ) {
                            itm.ToolTipText = _( "[file not found]" ) + " ";
                        }
                        itm.ToolTipText += item;
                        itm.Tag = item;
                        itm.Enabled = available;
                        itm.Click += new EventHandler( itm_Click );
                        menuFileRecent.DropDownItems.Add( itm );
                        added++;
                    }
                }
            } else {
                AppManager.EditorConfig.PushRecentFiles( "" );
            }
            if ( added == 0 ) {
                menuFileRecent.Enabled = false;
            } else {
                menuFileRecent.Enabled = true;
            }
        }

        /// <summary>
        /// 最後に保存したときから変更されているかどうかを取得または設定します
        /// </summary>
        private bool Edited {
            get {
                return m_edited;
            }
            set {
                m_edited = value;
                string file = m_manager.FileName;
                if ( file == "" ) {
                    file = "Untitled";
                } else {
                    file = Path.GetFileNameWithoutExtension( file );
                }
                string title = _APP_NAME + " - " + file;
                if ( m_edited ) {
                    title += " *";
                }
                if ( this.Text != title ) {
                    this.Text = title;
                }
                bool redo = m_manager.IsRedoAvailable;
                bool undo = m_manager.IsUndoAvailable;
                menuEditRedo.Enabled = redo;
                menuEditUndo.Enabled = undo;
                cMenuPianoRedo.Enabled = redo;
                cMenuPianoUndo.Enabled = undo;
                cMenuTrackSelectorRedo.Enabled = redo;
                cMenuTrackSelectorUndo.Enabled = undo;
                m_manager.setRenderRequired( m_manager.Selected, true );
                if ( m_manager.VsqFile != null ) {
                    int draft = (int)m_manager.VsqFile.TotalClocks;
                    if ( draft > hScroll.Maximum ) {
                        SetHScrollRange( draft );
                    }
                }
                UpdateDrawObjectList();
            }
        }

        /// <summary>
        /// 入力用のテキストボックスを初期化します
        /// </summary>
        private void InitializeInputTextBox( string phrase, string phonetic_symbol, Point position, bool phonetic_symbol_edit_mode ) {
            DisposeInputTextBox();
            m_input_textbox = new TextBoxEx();
            m_input_textbox.Visible = false;
            m_input_textbox.BorderStyle = BorderStyle.None;
            m_input_textbox.Height = AppManager.EditorConfig.PxTrackHeight;
            m_input_textbox.Width = 80;
            m_input_textbox.BackColor = Color.White;
            m_input_textbox.ImeMode = m_last_imemode;
            m_input_textbox.KeyUp += new KeyEventHandler( m_input_textbox_KeyUp );
            m_input_textbox.KeyDown += new KeyEventHandler( m_input_textbox_KeyDown );
            m_input_textbox.ImeModeChanged += new EventHandler( m_input_textbox_ImeModeChanged );
            if ( phonetic_symbol_edit_mode ) {
                m_input_textbox.Tag = new TagLyricTextBox( phrase, true );
                m_input_textbox.Text = phonetic_symbol;
                m_input_textbox.BackColor = s_txtbox_backcolor;
            } else {
                m_input_textbox.Tag = new TagLyricTextBox( phonetic_symbol, false );
                m_input_textbox.Text = phrase;
                m_input_textbox.BackColor = Color.White;
            }
            m_input_textbox.Location = new Point( position.X + 4, position.Y + 2 );
            m_input_textbox.Font = new Font( AppManager.EditorConfig.BaseFontName, 9 );
            m_input_textbox.Parent = pictPianoRoll;
            m_input_textbox.Visible = true;
            m_input_textbox.Focus();
            m_input_textbox.SelectAll();
        }

        private void DisposeInputTextBox() {
            if ( m_input_textbox != null ) {
                if ( !m_input_textbox.IsDisposed ) {
                    m_input_textbox.Dispose();
                    TagLyricTextBox tltb = (TagLyricTextBox)m_input_textbox.Tag;
                    m_last_symbol_edit_mode = tltb.PhoneticSymbolEditMode;
                }
                m_input_textbox = null;
            }
            pictPianoRoll.Focus();
        }

        /// <summary>
        /// 歌詞入力用テキストボックスのモード（歌詞/発音記号）を切り替えます
        /// </summary>
        private void FlipInputTextBoxMode() {
            TagLyricTextBox kvp = (TagLyricTextBox)m_input_textbox.Tag;
            string new_value = m_input_textbox.Text;
            if ( !kvp.PhoneticSymbolEditMode ) {
                m_input_textbox.BackColor = s_txtbox_backcolor;
            } else {
                m_input_textbox.BackColor = Color.White;
            }
            m_input_textbox.Text = kvp.BufferText;
            m_input_textbox.Tag = new TagLyricTextBox( new_value, !kvp.PhoneticSymbolEditMode );
        }

        /// <summary>
        /// 音の高さを表すnoteから、画面に描くべきy座標を計算します
        /// </summary>
        /// <param name="note"></param>
        /// <returns></returns>
        int YCoordFromNote( int note ) {
            return -1 * (note - 127) * AppManager.EditorConfig.PxTrackHeight - StartToDrawY;
        }

        /// <summary>
        /// ピアノロール画面のy座標から、その位置における音の高さを取得します
        /// </summary>
        /// <param name="y"></param>
        /// <returns></returns>
        int NoteFromYCoord( int y ) {
            return 127 - (int)((double)(StartToDrawY + y) / (double)AppManager.EditorConfig.PxTrackHeight);
        }

        /// <summary>
        /// アンドゥ処理を行います
        /// </summary>
        public void Undo() {
            if ( m_manager.IsUndoAvailable ) {
                m_manager.Undo();
                menuEditRedo.Enabled = m_manager.IsRedoAvailable;
                menuEditUndo.Enabled = m_manager.IsUndoAvailable;
                cMenuPianoRedo.Enabled = m_manager.IsRedoAvailable;
                cMenuPianoUndo.Enabled = m_manager.IsUndoAvailable;
                cMenuTrackSelectorRedo.Enabled = m_manager.IsRedoAvailable;
                cMenuTrackSelectorUndo.Enabled = m_manager.IsUndoAvailable;
                m_mixer_dlg.UpdateStatus();
                UpdateDrawObjectList();
            }
        }

        /// <summary>
        /// リドゥ処理を行います
        /// </summary>
        public void Redo() {
            if ( m_manager.IsRedoAvailable ) {
                m_manager.Redo();
                menuEditRedo.Enabled = m_manager.IsRedoAvailable;
                menuEditUndo.Enabled = m_manager.IsUndoAvailable;
                cMenuPianoRedo.Enabled = m_manager.IsRedoAvailable;
                cMenuPianoUndo.Enabled = m_manager.IsUndoAvailable;
                cMenuTrackSelectorRedo.Enabled = m_manager.IsRedoAvailable;
                cMenuTrackSelectorUndo.Enabled = m_manager.IsUndoAvailable;
                m_mixer_dlg.UpdateStatus();
                UpdateDrawObjectList();
            }
        }

        /*// <summary>
        /// 音符長さのクオンタイズが，3連符モードかどうかを表す値を取得または設定します
        /// </summary>
        bool LengthQuantizeTriplet {
            get {
                return AppManager.EditorConfig.LengthQuantizeTriplet;
            }
            set {
                AppManager.EditorConfig.LengthQuantizeTriplet = value;
                m_toolbar_measure.LengthQuantizeTriplet = value;
                menuSettingLengthQuantizeTriplet.Checked = value;
            }
        }
        
        /// <summary>
        /// 音符位置のクオンタイズが，3連符モードかどうかを表す値を取得または設定します
        /// </summary>
        bool PositionQuantizeTriplet {
            get {
                return AppManager.EditorConfig.PositionQuantizeTriplet;
            }
            set {
                AppManager.EditorConfig.PositionQuantizeTriplet = value;
                m_toolbar_measure.PositionQuantizeTriplet = value;
                menuSettingPositionQuantizeTriplet.Checked = value;
            }
        }
        
        /// <summary>
        /// 音符長さのクオンタイズモードを取得または設定します
        /// </summary>
        QuantizeMode LengthQuantize {
            get {
                return AppManager.EditorConfig.LengthQuantize;
            }
            set {
                ToolStripMenuItem item;
                for ( int i = 0; i <= 5; i++ ) {
                    item = (ToolStripMenuItem)menuSettingLengthQuantize.DropDownItems[i];
                    item.Checked = false;
                }
                AppManager.EditorConfig.LengthQuantize = value;
                int index = (int)AppManager.EditorConfig.LengthQuantize;
                item = (ToolStripMenuItem)menuSettingLengthQuantize.DropDownItems[index];
                item.Checked = true;
                m_toolbar_measure.LengthQuantize = value;
            }
        }

        /// <summary>
        /// 音符位置のクオンタイズモードを取得または設定します
        /// </summary>
        QuantizeMode PositionQuantize {
            get {
                return AppManager.EditorConfig.PositionQuantize;
            }
            set {
                ToolStripMenuItem item;
                for ( int i = 0; i <= 5; i++ ) {
                    item = (ToolStripMenuItem)menuSettingPositionQuantize.DropDownItems[i];
                    item.Checked = false;
                }
                AppManager.EditorConfig.PositionQuantize = value;
                int index = (int)AppManager.EditorConfig.PositionQuantize;
                item = (ToolStripMenuItem)menuSettingPositionQuantize.DropDownItems[index];
                item.Checked = true;
                m_toolbar_measure.PositionQuantize = value;
            }
        }*/

        /// <summary>
        /// ピアノロールの時間方向の表示倍率(pixel/clock)
        /// </summary>
        public float ScaleX {
            get {
                return m_scalex;
            }
        }

        int StartToDrawY {
            get {
                return (int)((128 * AppManager.EditorConfig.PxTrackHeight - vScroll.Height) * (float)vScroll.Value / ((float)vScroll.Maximum));
            }
        }

        int StartToDrawX {
            get {
                return (int)(hScroll.Value * ScaleX);
            }
        }

        /*// <summary>
        /// パネルに表示されているツールバーの個数を取得します
        /// </summary>
        int ToolBarsOnPanel {
            get {
                int toolbars_on_panel = 0;
                if ( m_toolbar_edit.Enabled && (!m_toolbar_edit.IsWindowMode) ) {
                    toolbars_on_panel++;
                }
                if ( m_toolbar_measure.Enabled && (!m_toolbar_measure.IsWindowMode) ) {
                    toolbars_on_panel++;
                }
                if ( m_toolbar_transport.Enabled && (!m_toolbar_transport.IsWindowMode) ) {
                    toolbars_on_panel++;
                }
                return toolbars_on_panel;
            }
        }*/

        /*public Bitmap GetScreenShot( int width, int height ) {
            Bitmap res = new Bitmap( width, height, System.Drawing.Imaging.PixelFormat.Format24bppRgb );
            //int xoffset = picturePositionIndicator.Left;
            //int yoffset = picturePositionIndicator.Top;
            int num_toolbar = ToolBarsOnPanel;
            int h_toolbar = 0;
            if ( num_toolbar > 0 ) {
                h_toolbar = m_toolbar_edit.ClientSize.Height;
            }

            int h_position_indicator = picturePositionIndicator.Height;
            int h_hscroll = hScroll.Height;
            int h_track_selector = trackSelector.Height;
            int h_piano_roll = height - (h_toolbar + h_position_indicator + h_hscroll + h_track_selector);
            if ( h_piano_roll <= 0 ) {
                h_piano_roll = pictureBox1.Height;
            }

            using ( Bitmap position_indicator = new Bitmap( width, h_position_indicator ) )
            using ( Bitmap piano_roll = new Bitmap( width - vScroll.Width, h_piano_roll ) )
            using ( Bitmap track_selector = new Bitmap( width, h_track_selector ) ) {
                Bitmap toolbar = null;
                if ( h_toolbar > 0 ) {
                    toolbar = new Bitmap( width, h_toolbar );
                    using ( Graphics g = Graphics.FromImage( toolbar ) ) {
                        g.Clear( Color.Black );
                        int xposi = 0;
                        if ( m_toolbar_measure.Enabled && !m_toolbar_measure.IsWindowMode ) {
                            m_toolbar_measure.DrawTo(
                                g,
                                new Point( xposi, 0 ) );
                            m_toolbar_measure.VirtualLocation = new Point( xposi, 0 );
                            xposi += m_toolbar_measure.ClientSize.Width;
                        }
                        if ( m_toolbar_edit.Enabled && !m_toolbar_edit.IsWindowMode ) {
                            m_toolbar_edit.DrawTo(
                                g,
                                new Point( xposi, 0 ) );
                            m_toolbar_edit.VirtualLocation = new Point( xposi, 0 );
                            xposi += m_toolbar_edit.ClientSize.Width;
                        }
                        if ( m_toolbar_transport.Enabled && !m_toolbar_transport.IsWindowMode ) {
                            m_toolbar_transport.DrawTo(
                                g,
                                new Point( xposi, 0 ) );
                            m_toolbar_transport.VirtualLocation = new Point( xposi, 0 );
                            xposi += m_toolbar_transport.ClientSize.Width;
                        }
                    }
                }
                using ( Graphics g = Graphics.FromImage( position_indicator ) ) {
                    g.Clear( Color.Black );
                    picturePositionIndicatorDrawTo( g );
                }

                using ( Graphics g = Graphics.FromImage( piano_roll ) ) {
                    pictureBox1_Paint( null, new PaintEventArgs( g, new Rectangle( new Point( 0, 0 ), piano_roll.Size ) ) );
                    //DrawTo( g, piano_roll.Size, new Point( 0, 0 ) );
                }

                using ( Graphics g = Graphics.FromImage( track_selector ) ) {
                    trackSelector.DrawTo( g, track_selector.Size );
                }

                using ( Graphics g2 = Graphics.FromImage( res ) ) {
                    if ( h_toolbar > 0 ) {
                        g2.DrawImage( toolbar, 0, 0 );
                    }
                    g2.DrawImage( position_indicator, 0, h_toolbar );
                    g2.DrawImage( piano_roll, 0, h_toolbar + h_position_indicator );
                    g2.FillRectangle(
                        new SolidBrush( Color.FromArgb( 194, 194, 194 ) ),
                        new Rectangle( 0, h_toolbar + h_position_indicator + h_piano_roll, width, hScroll.Height ) );

                    VEScrollBar.DrawTo( g2,
                        hScroll,
                        new Rectangle(
                            hScroll.Left,
                            h_toolbar + h_position_indicator + h_piano_roll,
                            width - hScroll.Left - (width - trackBar.Left),
                            hScroll.Height
                        )
                    );

                    VEScrollBar.DrawTo(
                        g2,
                        vScroll,
                        new Rectangle(
                            width - vScroll.Width,
                            h_toolbar + h_position_indicator,
                            vScroll.Width,
                            h_piano_roll
                        )
                    );
                    trackBar.DrawTo( g2, new Point( width - vScroll.Width - trackBar.Width, h_toolbar + h_position_indicator + h_piano_roll ) );
                    g2.DrawImage( track_selector, 0, h_toolbar + h_position_indicator + h_piano_roll + h_hscroll );
                }
                return res;
            }
        }*/

        /// <summary>
        /// pがrcの中にあるかどうかを判定します
        /// </summary>
        /// <param name="p"></param>
        /// <param name="rc"></param>
        /// <returns></returns>
        static bool IsInRect( Point p, Rectangle rc ) {
            if ( rc.X <= p.X ) {
                if ( p.X <= rc.X + rc.Width ) {
                    if ( rc.Y <= p.Y ) {
                        if ( p.Y <= rc.Y + rc.Height ) {
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// クオンタイズ時の音符の最小単位を、クロック数に換算したものを取得します
        /// </summary>
        /// <param name="qm"></param>
        /// <returns></returns>
        private static int GetUnitClocksOfQuantize( QuantizeMode qm, bool triplet ) {
            int ret = 1;
            switch ( qm ) {
                case QuantizeMode.p4:
                    ret = 480;
                    break;
                case QuantizeMode.p8:
                    ret = 240;
                    break;
                case QuantizeMode.p16:
                    ret = 120;
                    break;
                case QuantizeMode.p32:
                    ret = 60;
                    break;
                case QuantizeMode.p64:
                    ret = 30;
                    break;
                default:
                    return 1;
            }
            if ( triplet ) {
                ret = ret * 2 / 3;
            }
            return ret;
        }

        /// <summary>
        /// マウス位置におけるIDを返します。該当するIDが無ければnullを返します
        /// rectには、該当するIDがあればその画面上での形状を、該当するIDがなければ、
        /// 画面上で最も近かったIDの画面上での形状を返します
        /// </summary>
        /// <param name="mouse_position"></param>
        /// <returns></returns>
        VsqEvent GetItemAtClickedPosition( Point mouse_position, out Rectangle rect ) {
            rect = new Rectangle();
            if ( 68 <= mouse_position.X && mouse_position.X <= pictPianoRoll.Width - 18 ) {
                if ( 0 <= mouse_position.Y && mouse_position.Y <= pictPianoRoll.Height - 18 ) {
                    int selected = m_manager.Selected;
                    if ( selected >= 1 ) {
                        for ( int j = 0; j < m_manager.VsqFile.Tracks[selected].Events.Count; j++ ) {
                            int timesig = m_manager.VsqFile.Tracks[selected].Events[j].Clock;
                            int internal_id = m_manager.VsqFile.Tracks[selected].Events[j].InternalID;
                            // イベントで指定されたIDがLyricであった場合
                            if ( m_manager.VsqFile.Tracks[selected].Events[j].ID.type == VsqIDType.Anote &&
                                m_manager.VsqFile.Tracks[selected].Events[j].ID.LyricHandle != null ) {
                                // 発音長を取得
                                int length = m_manager.VsqFile.Tracks[selected].Events[j].ID.Length;
                                int note = m_manager.VsqFile.Tracks[selected].Events[j].ID.Note;
                                int x = XCoordFromClocks( timesig );
                                int y = YCoordFromNote( note );
                                int lyric_width = (int)(length * ScaleX);
                                if ( x + lyric_width < 0 ) {
                                    continue;
                                } else if ( pictPianoRoll.Width < x ) {
                                    break;
                                }
                                if ( x <= mouse_position.X && mouse_position.X <= x + lyric_width ) {
                                    if ( y + 1 <= mouse_position.Y && mouse_position.Y <= y + AppManager.EditorConfig.PxTrackHeight ) {
                                        rect = new Rectangle( x, y + 1, lyric_width, AppManager.EditorConfig.PxTrackHeight );
                                        return m_manager.VsqFile.Tracks[selected].Events[j];
                                    }
                                }
                            }
                        }
                    }
                }
            }
            return null;
        }

        private void OpenVsqCore( string file ) {
            m_manager.ReadVsq( file );
            if ( m_manager.VsqFile.Tracks.Count >= 2 ) {
                bool found = false;
                foreach ( MidiEvent item in m_manager.VsqFile.Tracks[0].MidiEvent ) {
                    if ( item.type == MidiEventType.tempo && !found ) {
                        AppManager.BaseTempo = item.intValue[0];
                        found = true;
                    }
                }
                SetHScrollRange( (int)m_manager.VsqFile.TotalClocks );
            }
            AppManager.EditorConfig.PushRecentFiles( file );
            UpdateRecentFileMenu();
            Edited = false;
            m_manager.ClearCommandBuffer();
            m_mixer_dlg.UpdateStatus();
#if DEBUG
            for ( int i = 1; i < m_manager.VsqFile.Tracks.Count; i++ ) {
                using ( StreamWriter sw = new StreamWriter( "note_length_" + i + ".txt" ) ) {
                    foreach ( VsqEvent ev in m_manager.VsqFile.Tracks[i].Events )
                        if ( ev.ID.type == VsqIDType.Anote ) {
                            sw.WriteLine( ev.Clock + "\t" + ev.ID.LyricHandle.L0.Phrase + "\t" + ev.ID.LyricHandle.L0.PhoneticSymbol + "\t" + ev.ID.Length );
                        }
                }
            }
#endif
        }

        private void UpdateMenuFonts( string font_name ) {
            if ( font_name == "" ) {
                return;
            }
            Font font = new Font( font_name, menuFile.Font.Size, menuFile.Font.Unit );
            this.Font = font;
            foreach ( Control c in this.Controls ) {
                Boare.Lib.AppUtil.Misc.ApplyFontRecurse( c, font );
            }
        }

        private void picturePositionIndicatorDrawTo( Graphics g ) {
            //using ( SolidBrush BRS_FONT_NORMAL = new SolidBrush( Color.Black ) )
            using ( Font SMALL_FONT = new Font( AppManager.EditorConfig.ScreenFontName, 8 ) ) {
                int width = picturePositionIndicator.Width;
                int height = picturePositionIndicator.Height;

                #region 小節ごとの線
                IEnumerable<VsqBarLineType> enumerable = m_manager.VsqFile.GetEnumerator( ClockFromXCoord( width ) );
                int dashed_line_step = GetUnitClocksOfQuantize( AppManager.EditorConfig.PositionQuantize, AppManager.EditorConfig.PositionQuantizeTriplet );
                foreach ( VsqBarLineType blt in enumerable ) {
                    int local_clock_step = 480 * 4 / blt.LocalDenominator;
                    int x = XCoordFromClocks( blt.Clock );
                    if ( blt.IsSeparator ) {
                        int current = blt.BarCount - m_manager.VsqFile.PreMeasure + 1;
                        g.DrawLine( s_pen_105_105_105,
                                    new Point( x, 3 ),
                                    new Point( x, 46 ) );
                        // 小節の数字
                        g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
                        g.DrawString( current.ToString(),
                                      SMALL_FONT,
                                      Brushes.Black,
                                      new PointF( x + 4, 6 ) );
                        g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.Default;
                    } else {
                        g.DrawLine( s_pen_105_105_105,
                                    new Point( x, 11 ),
                                    new Point( x, 16 ) );
                        g.DrawLine( s_pen_105_105_105,
                                    new Point( x, 26 ),
                                    new Point( x, 31 ) );
                        g.DrawLine( s_pen_105_105_105,
                                    new Point( x, 41 ),
                                    new Point( x, 46 ) );
                    }
                    if ( dashed_line_step > 1 && m_manager.GridVisible ) {
                        int numDashedLine = local_clock_step / dashed_line_step;
                        for ( int i = 1; i < numDashedLine; i++ ) {
                            int x2 = XCoordFromClocks( blt.Clock + i * dashed_line_step );
                            g.DrawLine( s_pen_065_065_065,
                                        new Point( x2, 9 + 5 ),
                                        new Point( x2, 14 + 3 ) );
                            g.DrawLine( s_pen_065_065_065,
                                        new Point( x2, 24 + 5 ),
                                        new Point( x2, 29 + 3 ) );
                            g.DrawLine( s_pen_065_065_065,
                                        new Point( x2, 39 + 5 ),
                                        new Point( x2, 44 + 3 ) );
                        }
                    }
                }
                #endregion

                if ( m_manager.VsqFile != null ) {
                    #region 拍子の変更
                    for ( int i = 0; i < m_manager.VsqFile.TimeSigTable.Count; i++ ) {
                        int clock = m_manager.VsqFile.TimeSigTable[i].Clock;
                        int barcount = m_manager.VsqFile.TimeSigTable[i].BarCount;
                        int x = XCoordFromClocks( clock );
                        if ( width < x ) {
                            break;
                        }
                        string s = m_manager.VsqFile.TimeSigTable[i].Numerator + "/" + m_manager.VsqFile.TimeSigTable[i].Denominator;
                        if ( m_manager.SelectedTimesig.ContainsKey( barcount ) ) {
                            g.DrawString( s,
                                          SMALL_FONT,
                                          AppManager.HilightBrush,
                                          new PointF( x + 4, 36 ) );
                        } else {
                            g.DrawString( s,
                                          SMALL_FONT,
                                          Brushes.Black,
                                          new PointF( x + 4, 36 ) );
                        }

                        if ( m_timesig_dragging ) {
                            if ( m_manager.SelectedTimesig.ContainsKey( barcount ) ) {
                                int edit_clock_x = XCoordFromClocks( m_manager.VsqFile.GetClockFromBarCount( m_manager.SelectedTimesig[barcount].Editing.BarCount ) );
                                g.DrawLine( s_pen_187_187_255,
                                            new Point( edit_clock_x - 1, 32 ),
                                            new Point( edit_clock_x - 1, picturePositionIndicator.Height - 1 ) );
                                g.DrawLine( s_pen_007_007_151,
                                            new Point( edit_clock_x, 32 ),
                                            new Point( edit_clock_x, picturePositionIndicator.Height - 1 ) );
                            }
                        }
                    }
                    #endregion

                    #region テンポの変更
                    for ( int i = 0; i < m_manager.VsqFile.TempoTable.Count; i++ ) {
                        int clock = m_manager.VsqFile.TempoTable[i].Clock;
                        int x = XCoordFromClocks( clock );
                        if ( width < x ) {
                            break;
                        }
                        string s = (60e6 / (float)m_manager.VsqFile.TempoTable[i].Tempo).ToString( "#.00" );
                        if ( m_manager.SelectedTempo.ContainsKey( clock ) ) {
                            g.DrawString( s,
                                          SMALL_FONT,
                                          AppManager.HilightBrush,
                                          new PointF( x + 4, 21 ) );
                        } else {
                            g.DrawString( s,
                                          SMALL_FONT,
                                          Brushes.Black,
                                          new PointF( x + 4, 21 ) );
                        }

                        if ( m_tempo_dragging ) {
                            if ( m_manager.SelectedTempo.ContainsKey( clock ) ) {
                                int edit_clock_x = XCoordFromClocks( m_manager.SelectedTempo[clock].Editing.Clock );
                                g.DrawLine( s_pen_187_187_255,
                                            new Point( edit_clock_x - 1, 18 ),
                                            new Point( edit_clock_x - 1, 32 ) );
                                g.DrawLine( s_pen_007_007_151,
                                            new Point( edit_clock_x, 18 ),
                                            new Point( edit_clock_x, 32 ) );
                            }
                        }
                    }
                    #endregion
                }

                #region 外枠
                /* 左(外側) */
                g.DrawLine( new Pen( Color.FromArgb( 160, 160, 160 ) ),
                            new Point( 0, 0 ),
                            new Point( 0, height - 1 ) );
                /* 左(内側) */
                g.DrawLine( new Pen( Color.FromArgb( 105, 105, 105 ) ),
                            new Point( 1, 1 ),
                            new Point( 1, height - 2 ) );
                /* 中(上側) */
                g.DrawLine( new Pen( Color.FromArgb( 160, 160, 160 ) ),
                            new Point( 1, 47 ),
                            new Point( width - 2, 47 ) );
                /* 中(下側) */
                g.DrawLine( new Pen( Color.FromArgb( 105, 105, 105 ) ),
                            new Point( 2, 48 ),
                            new Point( width - 3, 48 ) );
                // 右(外側)
                g.DrawLine( new Pen( Color.White ),
                            new Point( width - 1, 0 ),
                            new Point( width - 1, height - 1 ) );
                // 右(内側)
                g.DrawLine( new Pen( Color.FromArgb( 241, 239, 226 ) ),
                            new Point( width - 2, 1 ),
                            new Point( width - 2, height - 1 ) );
                #endregion

                #region TEMPO & BEAT
                // TEMPO BEATの文字の部分。小節数が被っている可能性があるので、塗り潰す
                g.FillRectangle(
                    new SolidBrush( picturePositionIndicator.BackColor ),
                    new Rectangle( 2, 3, 65, 45 ) );
                // 横ライン上
                g.DrawLine( new Pen( Color.FromArgb( 104, 104, 104 ) ),
                            new Point( 2, 17 ),
                            new Point( width - 3, 17 ) );
                // 横ライン中央
                g.DrawLine( new Pen( Color.FromArgb( 104, 104, 104 ) ),
                            new Point( 2, 32 ),
                            new Point( width - 3, 32 ) );
                // 横ライン下
                g.DrawLine( new Pen( Color.FromArgb( 104, 104, 104 ) ),
                            new Point( 2, 47 ),
                            new Point( width - 3, 47 ) );
                // 縦ライン
                g.DrawLine( new Pen( Color.FromArgb( 104, 104, 104 ) ),
                            new Point( 67, 2 ),
                            new Point( 67, 46 ) );
                /* TEMPO&BEATとピアノロールの境界 */
                g.DrawLine( new Pen( Color.FromArgb( 104, 104, 104 ) ),
                            new Point( 67, 48 ),
                            new Point( width - 18, 48 ) );
                g.DrawString( "TEMPO",
                              SMALL_FONT,
                              Brushes.Black,
                              new PointF( 11, 20 ) );
                g.DrawString( "BEAT",
                              SMALL_FONT,
                              Brushes.Black,
                              new PointF( 11, 35 ) );
                g.DrawLine( new Pen( Color.FromArgb( 172, 168, 153 ) ),
                            new Point( 0, 0 ),
                            new Point( width, 0 ) );
                g.DrawLine( new Pen( Color.FromArgb( 113, 111, 100 ) ),
                            new Point( 1, 1 ),
                            new Point( width - 1, 1 ) );

                #endregion

                #region 現在のマーカー
                float xoffset = 73 - StartToDrawX;
                int marker_x = (int)(m_manager.CurrentClock * m_scalex + xoffset);
                if ( 67 <= marker_x && marker_x <= width ) {
                    g.DrawLine(
                        new Pen( Color.White, 2f ),
                        new Point( marker_x, 2 ),
                        new Point( marker_x, height ) );
                }
                if ( m_manager.StartMarkerEnabled ) {
                    int x = XCoordFromClocks( m_manager.StartMarker );
                    g.DrawImage(
                        Properties.Resources.start_marker, new Point( x, 3 ) );
                }
                if ( m_manager.EndMarkerEnabled ) {
                    int x = XCoordFromClocks( m_manager.EndMarker ) - 6;
                    g.DrawImage(
                        Properties.Resources.end_marker, new Point( x, 3 ) );
                }
                #endregion
            }
        }
    }
}

namespace Boare.VECaptureScript {
    /*public class test {
        public static void Edit( VsqFile vsq ) {
            String tmp = @"C:\test.wav";//            Path.GetTempFileName();
            int track = m_manager.Selected;
            String reference = @"C:\vocal_first_mono.wav";
            Wave ref_wave = new Wave( reference );

            const double _VOLUME_WINDOW_SEC = 2048.0 / 44100.0;
            int ref_wnd_size = (int)(_VOLUME_WINDOW_SEC * ref_wave.SampleRate);
            double[] ref_wnd = new double[ref_wnd_size];
            for( int i = 0; i < ref_wnd_size; i++ ){
                ref_wnd[i] = math.wnd_hann( i / (double)ref_wnd_size );
            }

            int target_wnd_size = 0;
            double[] target_wnd = null;
            for( int i = 0; i < 1; i++ ){
                Boare.Cadencii.FormSynthesize fs = new Boare.Cadencii.FormSynthesize( vsq, 500, new int[] { track }, new String[] { tmp }, (int)vsq.TotalClocks + 960 );
                DialogResult dr = fs.ShowDialog();
                VsqBPList original = (VsqBPList)vsq.Tracks[track][VsqCurveType.DYN].Clone();
                vsq.Tracks[track][VsqCurveType.DYN].Clear();
                using ( Wave wv = new Wave( tmp ) ) {
                    if ( target_wnd_size == 0 ) {
                        target_wnd_size = (int)(_VOLUME_WINDOW_SEC * wv.SampleRate);
                        target_wnd = new double[target_wnd_size];
                        for ( int j = 0; j < target_wnd_size; j++ ) {
                            target_wnd[j] = math.wnd_hann( j / (double)target_wnd_size );
                        }
                    }
                    for ( int j = 0; j < vsq.Tracks[track].Events.Count; j++ ) {
                        VsqEvent ve = vsq.Tracks[track].Events[j];
                        if ( ve.ID.type == VsqIDType.Anote ) {
                            int start = ve.Clock;
                            int end = ve.ID.Length + start;
                            for ( int n = start; n <= end; n++ ) {
                                int target_dyn = original[n];
                                double time = vsq.SecFromClock( n );
                                double reference_volume = ref_wave.GetVolume( (int)(time * ref_wave.SampleRate), ref_wnd );
                                double target_volume = wv.GetVolume( (int)(time * wv.SampleRate), target_wnd );
                                int new_dyn = (int)(reference_volume / target_volume * target_dyn);
                                if ( new_dyn < 0 ) {
                                    new_dyn = 0;
                                } else if ( 127 < new_dyn ) {
                                    new_dyn = 127;
                                }
                                vsq.Tracks[track][VsqCurveType.DYN].Add( n, new_dyn );
                            }
                        }
                    }
                }
            }
        }
    }*/
/*        public static void Edit( VsqFile Vsq ) {
            System.Diagnostics.Debug.WriteLine( "--------------------------------------------------------" );
            int track = Boare.Cadencii.FormMain.Instance.CommonConfig.Selected;
            Vsq.Tracks[track][VsqCurveType.PIT].Clear();
            Vsq.Tracks[track][VsqCurveType.PBS].Clear();
            List<double> listf0 = new List<double>();
            using ( StreamReader sr = new StreamReader( @"C:\vocal_first_mono.f0" ) ) {
                string line;
                while ( (line = sr.ReadLine()) != null ) {
                    string[] spl = line.Split( " ".ToCharArray(), StringSplitOptions.RemoveEmptyEntries );
                    System.Diagnostics.Debug.WriteLine( spl[0] );
                    listf0.Add( double.Parse( spl[0] ) );
                }
            }
            double[] f0 = listf0.ToArray();
            listf0.Clear();
            listf0 = null;

            const double _DT = 0.01;
            // とりあえずPBSは24にFIX
            Vsq.Tracks[track][VsqCurveType.PBS].Add( Vsq.PreMeasureClocks, 24 );
            for ( int i = 0; i < Vsq.Tracks[track].Events.Count; i++ ) {
                if ( Vsq.Tracks[track].Events[i].ID.type == VsqIDType.Anote ) {
                    VsqEvent ve = Vsq.Tracks[track].Events[i];
                    int start = ve.Clock;
                    int end = ve.ID.Length + start;
                    double sec_start = Vsq.SecFromClock( start );
                    double sec_end = Vsq.SecFromClock( end ) - 1;
                    int n_start = (int)(sec_start / _DT) + 1;
                    int n_end = (int)(sec_end / _DT);
                    int note = ve.ID.Note;
                    System.Diagnostics.Debug.WriteLine( "n_start,n_end=" + n_start + "," + n_end );
                    double base_freq = Math.Pow( 2.0, (note - 69) / 12.0 ) * 440.0;
                    for ( int n = n_start; n <= n_end; n++ ) {
                        double hz = f0[n];
                        if ( hz > 0.0 ) {
                            int pit = (int)(4096.0 * Math.Log( hz / base_freq, 2.0 ));
                            double t = n * _DT;
                            int t_clock = (int)Vsq.ClockFromSec( (float)t );
                            Vsq.Tracks[track][VsqCurveType.PIT].Add( t_clock, pit );
                        }
                    }
                }
            }
            System.Diagnostics.Debug.WriteLine( "--------------------------------------------------------" );
        }*/

}