﻿/*
 * KeySoundPlayer.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.IO;
using System.Media;
using System.Threading;
using System.Windows.Forms;

using Boare.Lib.Media;

namespace Boare.Cadencii {

    class KeySoundPlayer {
        /// <summary>
        /// 鍵盤を押した時に音を鳴らすためのプレイヤー
        /// </summary>
        private static SoundPlayer[] m_sound_previewer = new SoundPlayer[48];
        private static SoundPlayer m_temp_player = null;
        private static Thread m_initialization = null;
        private static bool m_initialized = false;
        private static bool[] m_prepared = new bool[127];
        private static bool s_abort_required = false;
        private static byte s_program_change = 0x00;

        static KeySoundPlayer() {
            for ( int i = 0; i < 127; i++ ) {
                m_prepared[i] = false;
            }
        }

        public static void RefreshPreviewSound() {
            AbortInit();
            for ( int i = 0; i < m_prepared.Length; i++ ) {
                m_prepared[i] = false;
            }
            if ( m_temp_player != null ) {
                m_temp_player.Stop();
                m_temp_player.Dispose();
                m_temp_player = null;
            }
            for ( int i = 0; i < m_sound_previewer.Length; i++ ) {
                if ( m_sound_previewer[i] != null ) {
                    m_sound_previewer[i].Stop();
                    m_sound_previewer[i].Dispose();
                    m_sound_previewer[i] = null;
                }
            }
            for ( int i = 0; i < 127; i++ ) {
                string path = Path.Combine( Path.Combine( Application.StartupPath, "cache" ), i + ".wav" );
                if ( File.Exists( path ) ) {
                    try {
                        File.Delete( path );
                    } catch {
                    }
                }
            }
            Init();
        }

        public static byte SingerProgramChange {
            get {
                return s_program_change;
            }
            set {
                s_program_change = value;
            }
        }

        public static bool IsBusy {
            get {
#if DEBUG
                Console.WriteLine( "KeySoundPlayer.IsBusy" );
#endif
                bool ret = false;
                if ( !m_initialized ) {
                    ret = true;
                } else if ( m_initialization != null ) {
                    if ( m_initialization.IsAlive ) {
                        ret = true;
                    } else {
                        m_initialization = null;
                        ret = false;
                    }
                } else {
                    ret = false;
                }
#if DEBUG
                Console.WriteLine( "    return " + ret );
#endif
                return ret;
            }
        }

        public static void Play( int note ) {
            if ( note < 0 || 127 <= note ) {
                return;
            }
            if ( !m_prepared[note] ) {
                return;
            }
            if ( 36 <= note && note <= 83 ) {
                if ( m_sound_previewer[note - 36] != null ) {
                    try {
                        m_sound_previewer[note - 36].Play();
                    } catch {
                        string path = Path.Combine( Path.Combine( Application.StartupPath, "cache" ), note + ".wav" );
                        try {
                            VSTiProxy.GenerateSinglePhone( note, s_program_change, path );
                            Wave wv = new Wave( path );
                            wv.TrimSilence();
                            wv.Monoralize();
                            wv.Write( path );
                            m_sound_previewer[note - 36].Dispose();
                            m_sound_previewer[note - 36] = new SoundPlayer( path );
                        } catch {
                        }
                    }
                }
            } else {
                if ( m_temp_player == null ) {
                    m_temp_player = new SoundPlayer();
                }
                string path = Path.Combine( Path.Combine( Application.StartupPath, "cache" ), note + ".wav" );
                if ( File.Exists( path ) ) {
                    m_temp_player.SoundLocation = path;
                    m_temp_player.Play();
                }
            }
        }

        public static void Finalize() {
            AbortInit();
        }

        public static void AbortInit() {
            if ( m_initialization != null ) {
                s_abort_required = true;
                while ( IsBusy ) {
                    Application.DoEvents();
                }
            }
        }

        private static void InitCore() {
#if DEBUG
            Common.DebugWriteLine( "KeySoundPlayer+InitCore()" );
#endif
            s_abort_required = false;
            // 音源を準備
            string cache_path = Path.Combine( Application.StartupPath, "cache" );
            if ( !Directory.Exists( cache_path ) ) {
                Directory.CreateDirectory( cache_path );
            }
#if DEBUG
            Common.DebugWriteLine( "    VSTiProxy.CurrentUser=" + VSTiProxy.CurrentUser );
            Common.DebugWriteLine( "    s_abort_required=" + s_abort_required );
#endif
            if ( VSTiProxy.CurrentUser != "" ) {
                return;
            } else {
                VSTiProxy.CurrentUser = "KeySoundPlayer";
            }
            for ( int i = 0; i < 127; i++ ) {
                if ( s_abort_required ) {
                    s_abort_required = false;
                    break;
                }
                string path = Path.Combine( cache_path, i + ".wav" );
                if ( !File.Exists( path ) ) {
                    try {
                        VSTiProxy.GenerateSinglePhone( i, s_program_change, path );
                        if ( File.Exists( path ) ) {
                            try {
                                Wave wv = new Wave( path );
                                wv.TrimSilence();
                                wv.Monoralize();
                                wv.Write( path );
                                m_prepared[i] = true;
                            } catch {
                                m_prepared[i] = false;
                            }
                        }
                    } catch {
                    }
                } else {
                    m_prepared[i] = true;
                }
                if ( 36 <= i && i <= 83 ) {
                    try {
                        m_sound_previewer[i - 36] = new SoundPlayer( path );
                    } catch {
                    }
                }
            }
            VSTiProxy.CurrentUser = "";
            m_initialized = true;
        }

        public static void Init() {
            m_initialization = new Thread( new ThreadStart( InitCore ) );
            m_initialization.Start();
        }
    }

}
