﻿/*
 * VSTiProxy.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Threading;
using System.ComponentModel;
using System.Security;

using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    enum VSTiProxyTrackType {
        TempoTrack,
        MainTrack,
    }

    [StructLayout( LayoutKind.Sequential, Pack = 1 )]
    struct VstEvent : IComparable<VstEvent> {
        public Int32 type;
        public Int32 byteSize;
        public Int32 deltaFrames;
        public Int32 flags;
        byte data0;
        byte data1;
        byte data2;
        byte data3;
        byte data4;
        byte data5;
        byte data6;
        byte data7;
        byte data8;
        byte data9;
        byte data10;
        byte data11;
        byte data12;
        byte data13;
        byte data14;
        byte data15;
        
        public int CompareTo( VstEvent item ) {
            return deltaFrames - item.deltaFrames;
        }
        
        public byte this[int index] {
            get {
                switch ( index ) {
                    case 0:
                        return data0;
                    case 1:
                        return data1;
                    case 2:
                        return data2;
                    case 3:
                        return data3;
                    case 4:
                        return data4;
                    case 5:
                        return data5;
                    case 6:
                        return data6;
                    case 7:
                        return data7;
                    case 8:
                        return data8;
                    case 9:
                        return data9;
                    case 10:
                        return data10;
                    case 11:
                        return data11;
                    case 12:
                        return data12;
                    case 13:
                        return data13;
                    case 14:
                        return data14;
                    case 15:
                        return data15;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }
            set {
                switch ( index ) {
                    case 0:
                        data0 = value;
                        break;
                    case 1:
                        data1 = value;
                        break;
                    case 2:
                        data2 = value;
                        break;
                    case 3:
                        data3 = value;
                        break;
                    case 4:
                        data4 = value;
                        break;
                    case 5:
                        data5 = value;
                        break;
                    case 6:
                        data6 = value;
                        break;
                    case 7:
                        data7 = value;
                        break;
                    case 8:
                        data8 = value;
                        break;
                    case 9:
                        data9 = value;
                        break;
                    case 10:
                        data10 = value;
                        break;
                    case 11:
                        data11 = value;
                        break;
                    case 12:
                        data12 = value;
                        break;
                    case 13:
                        data13 = value;
                        break;
                    case 14:
                        data14 = value;
                        break;
                    case 15:
                        data15 = value;
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }
        }
    }

    unsafe struct MIDI_EVENT {
        public uint clock;
        public MIDI_EVENT* pNext;
        public uint dwDataSize;
        public byte dwOffset;
        public byte* pMidiEvent;
    }

    /// <summary>
    /// C++製dll，VSTi2.dllを使った，VSTiホストへのプラクシー
    /// </summary>
    static class VSTiProxy {
        private static bool s_loaded = false;
        private static string s_dll_path = "C:\\Program Files\\Steinberg\\VSTplugins\\VOCALOID2\\VOCALOID2.dll";
        private static int s_error_counter = 0;
        private static int s_sample_rate = 44100;
        private static string s_current_user = "";

        const int NUM_ERR_THRESHOLD = 10;
        const int _BLOCK_SIZE = 44100;
        const int _SAMPLE_RATE = 44100;
/*      /// ファイル名を指定してプラグインを読み込む
        int __declspec(dllexport) LoadPluginA( char *name, int blockSize, int sampleRate );
        /// ファイル名を指定してプラグインを読み込む(ユニコード版)
        int __declspec(dllexport) LoadPluginW( WCHAR *name, int blockSize, int sampleRate );
        /// プラグインを開放する
        int __declspec(dllexport) FreePlugin();
        /// プラグインに送るmidiデータを指定する。
        int __declspec(dllexport) SendEvent( unsigned char *src, int *deltaFrames, int numEventsm, int targetTrack );
        /// 音符1個分のレンダリングを進める。続きのレンダリングはRenderContinueで行う。戻り値は、残りのフレーム数
        int __declspec(dllexport) RenderFirst( int *outL, int *outR, int sampleFrames );
        /// RenderFirstでレンダリング出来なかった残りの部分のレンダリングを行う。戻り値は、残りのフレーム数
        int __declspec(dllexport) RenderContinue( int *outL, int *outR, int sampleFrames );
        int __declspec(dllexport) RenderToWave( TCHAR *path );*/

        [DllImport( "VSTi2.dll", CharSet = CharSet.Ansi, EntryPoint = "LoadPluginA" )]
        private static extern int LoadPluginA_( string name, int blockSize, int sampleRate );

        [DllImport( "VSTi2.dll", CharSet = CharSet.Unicode, EntryPoint = "LoadPluginW" )]
        private static extern int LoadPluginW_( string name, int blockSize, int sampleRate );

        /*[System.Security.SuppressUnmanagedCodeSecurity]
        [DllImport( "VSTi2.dll", CharSet = CharSet.Unicode, EntryPoint = "LoadPluginW", CallingConvention = CallingConvention.StdCall )]
        private static extern int UnsafeLoadPluginWStdCall( string name, int blockSize, int sampleRate );*/

        [SuppressUnmanagedCodeSecurity]
        [DllImport( "VSTi2.dll", CharSet = CharSet.Unicode, EntryPoint = "LoadPluginW", CallingConvention = CallingConvention.Cdecl )]
        private static extern int UnsafeLoadPluginWCdecl( string name, int blockSize, int sampleRate );

        [DllImport( "VSTi2.dll", EntryPoint = "SendEvent" )]
        private static unsafe extern int SendEvent_( byte* src, int* deltaFrames, int numEvents, int targetTrack );

        [DllImport( "VSTi2.dll", CharSet = CharSet.Unicode, EntryPoint = "RenderToWaveW" )]
        private static extern int RenderToWaveW_( string name, uint eof_clock );

        [DllImport( "VSTi2.dll", EntryPoint = "FreePlugin" )]
        private static extern int FreePlugin_();

        [DllImport( "VSTi2.dll", EntryPoint="SetSampleRate" )]
        private static extern int SetSampleRate_( float sample_rate );

        [DllImport( "VSTi2.dll", EntryPoint="MainsChanged" )]
        private static extern int MainsChanged_( int status );

        [DllImport( "VSTi2.dll", EntryPoint = "SetBlockSize" )]
        private static extern int SetBlockSize_( int block_size );

        [DllImport( "VSTi2.dll", EntryPoint = "ProcessEvents")]
        private static unsafe extern int ProcessEvents_( int numEvents, VstEvent* events );

        [DllImport( "VSTi2.dll", EntryPoint = "ProcessReplacing" )]
        private static unsafe extern void ProcessReplacing_( float** out_buffer, int frames );

        [DllImport( "VSTi2.dll", EntryPoint = "AbortRendering" )]
        private static extern void AbortRendering_();

        [DllImport( "VSTi2.dll", EntryPoint = "GetProgress" )]
        private static extern double GetProgress_();

        /// <summary>
        /// このクラスを使っているMainFormを特定するIDを取得、または設定します。""なら、どのFormMainのインスタンスもコイツを使ってない
        /// </summary>
        public static string CurrentUser {
            get {
                return s_current_user;
            }
            set {
                s_current_user = value;
            }
        }

        public static bool ReloadPlugin() {
            if ( s_loaded ) {
                if ( FreePlugin() ) {
                    s_loaded = false;
                } else {
                    return false;
                }
            }
            return LoadPlugin( _BLOCK_SIZE, _SAMPLE_RATE );
        }


        public static double GetProgress() {
            return GetProgress_();
        }

        public static int SampleRate {
            get {
                return s_sample_rate;
            }
            set {
                s_sample_rate = value;
            }
        }

        private static double totalMilliSec_from_timeCode( uint timeCode, TempoTableEntry[] tempo ) {
            const int _TIME_FORMAT = 480;
            const int DEF_TEMPO = 500000;           // デフォルトのテンポ．
            double ret = 0.0;
            int index = -1;
            for ( int i = 0; i < tempo.Length; i++ ) {
                if ( timeCode <= tempo[i].Clock ) {
                    break;
                }
                index = i;
            }
            if ( index >= 0 ) {
                ret = tempo[index].Time + (timeCode - tempo[index].Clock) * (double)tempo[index].Tempo / (1000.0 * _TIME_FORMAT);
            } else {
                ret = timeCode * (double)DEF_TEMPO / (1000.0 * _TIME_FORMAT);
            }
            return ret;
        }

        public static unsafe int NEW_RenderToWave( NrpnData[] nrpn, TempoTableEntry[] tempo, int clEos, string file ) {
            if ( !s_loaded ) {
                s_loaded = LoadPlugin( 44100, 44100 );
            }
            /*WAVEFORMATEX wf;
            wf.nChannels       = 2;
            wf.wFormatTag      = WAVE_FORMAT_PCM;
            wf.wBitsPerSample  = 16;
            wf.nBlockAlign     = wf.nChannels*wf.wBitsPerSample/8;
            wf.nSamplesPerSec  = g_sampleRate;// sampleFrames;
            wf.nAvgBytesPerSec = wf.nSamplesPerSec*wf.nBlockAlign;

            // Create WAV
            HMMIO hmmio = mmioOpen( path, NULL, MMIO_CREATE | MMIO_WRITE | MMIO_EXCLUSIVE );
            MMCKINFO mmckRiff, mmckData, mmckFmt;
            mmckRiff.fccType = mmioStringToFOURCC( TEXT("WAVE"), 0 );
            mmioCreateChunk( hmmio, &mmckRiff, MMIO_CREATERIFF );
            mmckFmt.ckid = mmioStringToFOURCC( TEXT("fmt "), 0 );
            mmioCreateChunk( hmmio, &mmckFmt, 0 );
            mmioWrite( hmmio, (char*)&wf, sizeof(WAVEFORMATEX) );
            mmioAscend( hmmio, &mmckFmt, 0 );

            mmckData.ckid = mmioStringToFOURCC( TEXT("data"), 0 );
            mmioCreateChunk( hmmio, &mmckData, 0 );*/

            //MIDI_EVENT* lpEvents = merge_events( g_ppTrackEvents[0], g_ppTrackEvents[1] );
            //MIDI_EVENT* current = lpEvents;
            List<VstEvent> proc = new List<VstEvent>();
            for ( int i = 0; i < nrpn.Length; i++ ) {
                VstEvent ve = new VstEvent();
                ve.deltaFrames = nrpn[i].Clock;
                ve[0] = 0xb0;
                ve[1] = nrpn[i].Parameter;
                ve[2] = nrpn[i].Value;
                ve.byteSize = 3;
                proc.Add( ve );
            }
            for ( int i = 0; i < tempo.Length; i++ ) {
                VstEvent ve = new VstEvent();
                ve.deltaFrames = tempo[i].Clock;
                ve[0] = 0xff;
                ve[1] = 0x51;
                ve[2] = 0x03;
                uint t = (uint)tempo[i].Tempo;
                ve[3] = (byte)(t >> 16);
                t = (uint)(t - (ve[3] << 16));
                ve[4] = (byte)(t >> 8);
                ve[5] = (byte)(t - (ve[4] << 8));
                ve.byteSize = 6;
                proc.Add( ve );
            }
            proc.Sort();

            float[] left_ch_src = new float[s_sample_rate];
            float[] right_ch_src = new float[s_sample_rate];
            //float** out_buffer = { left_ch, right_ch };

            //DWORD* wave_bufs;
            //wave_bufs = new DWORD[g_sampleRate];

            SetSampleRate_( (float)s_sample_rate );
            MainsChanged_( 1 );
            SetBlockSize_( s_sample_rate );

            int cur_buf = 0;

            uint dwTempo = 500000;

            int delay = 0;
            int duration = 0;
            uint dwNow = 0;
            uint dwPrev = 0;
            uint dwDelta;
            uint dwDelay = 0;
            uint dwDeltaDelay = 0;

            int addr_msb = 0, addr_lsb = 0;
            int data_msb = 0, data_lsb = 0;

#if DEBUG
            Console.WriteLine( "***********************************************************************" );
            int total_processed = 0;
#endif

            //MIDI_EVENT* pWork = g_ppTrackEvents[1];
            int eof_clock = clEos;
            /*while ( !dwDelay && pWork ) {
                // Delayの取得
                if ( (pWork->pMidiEvent[0] & 0xf0) == 0xb0 ) {
                    switch ( pWork->pMidiEvent[1] ) {
                        case 0x63:
                            addr_msb = pWork->pMidiEvent[2];
                            addr_lsb = 0;
                            break;
                        case 0x62:
                            addr_lsb = pWork->pMidiEvent[2];
                            break;
                        case 0x06:
                            data_msb = pWork->pMidiEvent[2];
                            break;
                        case 0x26:
                            data_lsb = pWork->pMidiEvent[2];
                            if ( addr_msb == 0x50 && addr_lsb == 0x01 ) {
                                delay = data_msb << 7 | data_lsb;
#if DEBUG
                                Console.WriteLine( "delay=" + delay );
#endif
                                dwDelay = (uint)(delay * (double)m_sample_rate / 1000.0);
                            }
                            break;
                    }
                } else if ( pWork->pMidiEvent[0] == 0xff && pWork->pMidiEvent[1] == 0x2f && pWork->pMidiEvent[2] == 0x00 ) {
                    eof_clock = pWork->clock;
#if DEBUG
                    Console.WriteLine( "eof_clock=" + eof_clock );
#endif
                }
                pWork = pWork->pNext;
            }*/

            int current_index = 0;
            while ( true ) {
#if DEBUG
                Console.WriteLine( "-----------------------------------------------------------------------" );
#endif
                //MIDI_EVENT* pProcessEvent = current;
                int nEvents = 0;

                while ( proc[current_index].deltaFrames == dwNow ) {
                    /*if ( current->pMidiEvent[0] == 0xFF ) {
                        //メタイベントの処理
                        switch ( current->pMidiEvent[1] ) {
                            case 0x51:
                                dwTempo = (uint)(current->pMidiEvent[5] | (current->pMidiEvent[4] << 8) | (current->pMidiEvent[3] << 16));
                                break;
                        }
                    }*/
                    // durationを取得
                    if ( proc[current_index][0] == 0xb0 ) {
                        switch ( proc[current_index][1] ) {
                            case 0x63:
                                addr_msb = proc[current_index][2];
                                addr_lsb = 0;
                                break;
                            case 0x62:
                                addr_lsb = proc[current_index][2];
                                break;
                            case 0x06:
                                data_msb = proc[current_index][2];
                                break;
                            case 0x26:
                                data_lsb = proc[current_index][2];
                                // Note Duration in millisec
                                if ( addr_msb == 0x50 && addr_lsb == 0x04 ) {
                                    duration = data_msb << 7 | data_lsb;
#if DEBUG
                                    Console.WriteLine( "duration=" + duration );
#endif
                                }
                                break;
                        }
                    }

                    nEvents++;
                    current_index++;// = current->pNext;
                    if ( current_index >= proc.Count ) {
                        break;
                    }
                }

                if ( current_index >= proc.Count ) {
                    break;
                }

#if DEBUG
                Console.WriteLine( "nEvents=" + nEvents );
#endif
                double msNow = totalMilliSec_from_timeCode( dwNow, tempo );
                double msPrev = totalMilliSec_from_timeCode( dwPrev, tempo );
                //double dt = DeltaToMilliSecond( dwNow - dwPrev, dwTempo );
                double dt = msNow - msPrev;
                dwDelta = (uint)(dt * s_sample_rate / 1000.0);
#if DEBUG
                Console.WriteLine( "dwNow=" + dwNow );
                Console.WriteLine( "dwPrev=" + dwPrev );
                Console.WriteLine( "dt=" + dt );
                Console.WriteLine( "dwDelta=" + dwDelta );
#endif

                //VstEvents* pVSTEvents = (VstEvents*)malloc( sizeof( VstEvents ) + nEvents * sizeof( VstEvent* ) );
                //pVSTEvents->numEvents = 0;
                //pVSTEvents->reserved = NULL;

                VstEvent[] events = new VstEvent[nEvents];
                for ( int i = 0; i < nEvents; i++ ) {
                    //BYTE event_code = pProcessEvent->pMidiEvent[0];
                    //VstEvent* pVSTEvent = NULL;
                    //VstMidiEvent* pMidiEvent;

                    /*switch ( event_code ) {
                        case 0xff:
                            break;
                        case 0xf0:
                        case 0xf7:
                            break;
                        default:
                            pMidiEvent = (VstMidiEvent*)malloc( sizeof( VstMidiEvent ) + pProcessEvent->dwDataSize * sizeof( BYTE ) );
                            pMidiEvent->byteSize = sizeof( VstMidiEvent );
                            pMidiEvent->deltaFrames = dwDelta;
                            pMidiEvent->detune = 0;
                            pMidiEvent->flags = 1;
                            pMidiEvent->noteLength = 0;
                            pMidiEvent->noteOffset = 0;
                            pMidiEvent->noteOffVelocity = 0;
                            pMidiEvent->reserved1 = 0;
                            pMidiEvent->reserved2 = 0;
                            pMidiEvent->type = kVstMidiType;
                            memcpy( &pMidiEvent->midiData, &pProcessEvent->pMidiEvent[pProcessEvent->dwOffset], pProcessEvent->dwDataSize );
                            pVSTEvents->events[pVSTEvents->numEvents++] = (VstEvent*)pMidiEvent;
                            break;
                    }
                    pProcessEvent = pProcessEvent->pNext;*/
                    int index = current_index - nEvents + i + 1;
                    events[i] = new VstEvent();
                    events[i].flags = 0;
                    events[i].type = 0;
                    for ( int j = 0; j < proc[i].byteSize; j++ ) {
                        events[i][j] = proc[index][j];
                    }
                    events[i].deltaFrames = (int)dwDelta;
                }
                fixed( VstEvent* pVstEvent = &events[0] ){
                    ProcessEvents_( nEvents, pVstEvent );
                }

                /*for ( int i = 0; i < pVSTEvents->numEvents; i++ ) {
                    if ( pVSTEvents->events[i]->type == kVstSysExType ) {
                        VstMidiSysexEvent* pSysexEvent = (VstMidiSysexEvent*)pVSTEvents->events[i];
                        free( pSysexEvent->sysexDump );
                    }
                }*/

                while ( dwDelta > 0 ) {
                    uint dwFrames = dwDelta > s_sample_rate ? (uint)s_sample_rate : dwDelta;
                    fixed ( float* left = &left_ch_src[0] )
                    fixed ( float* right = &right_ch_src[0] ) {
                        float*[] bufa = new float*[] { left, right };
                        fixed ( float** buf = &bufa[0] ) {
                            ProcessReplacing_( buf, (int)dwFrames );
                        }
                    }
#if DEBUG
                    for( int i = 0; i < left_ch_src.Length; i++ ){
                        if ( i >= dwFrames ){
                            break;
                        }
                        //Console.WriteLine( left_ch_src[i] );
                    }
#endif
                    /*for ( size_t j = 0; j < dwFrames; j++ ) {
                        WORD right_chx = (WORD)(32768 * out_buffer[0][j]);
                        WORD left_chx = (WORD)(32768 * out_buffer[1][j]);
                        wave_bufs[j] = MAKELONG( right_chx, left_chx );
                    }*/

                    int iOffset = (int)(dwDelay - dwDeltaDelay);
                    if ( iOffset > dwFrames ) {
                        iOffset = (int)dwFrames;
                    }
                    if ( iOffset == 0 ) {
                        //mmioWrite( hmmio, (char*)wave_bufs, dwFrames * sizeof( DWORD ) );
                        //mmioAscend( hmmio, &mmckData, 0 );
                    } else {
                        dwDeltaDelay += (uint)iOffset;
                    }
#if DEBUG
                    total_processed += (int)dwFrames;
#endif
                    dwDelta -= dwFrames;
                }

                //free( pVSTEvents );

                dwPrev = dwNow;
                dwNow = (uint)nrpn[current_index].Clock; // current->clock;
            }

#if DEBUG
            Console.WriteLine( "dwDeltaDelay=" + dwDeltaDelay );
#endif
            //        dwDelta = g_sampleRate * (double)length / 1000.0 + dwDeltaDelay;
            double msLast = totalMilliSec_from_timeCode( dwNow, tempo );
            double msEof = totalMilliSec_from_timeCode( (uint)eof_clock, tempo );
            double msRemain = msEof - msLast;
            msRemain = (msRemain < 0.0) ? 0.0 : msRemain;
            uint dwDeltaDraft = (uint)(s_sample_rate * msRemain / 1000.0);
            dwDelta = (uint)(s_sample_rate * ((double)duration + (double)delay) / 1000.0 + dwDeltaDelay);
#if DEBUG
            Console.WriteLine( "dwDeltaDraft=" + dwDeltaDraft );
            Console.WriteLine( "dwDelta=" + dwDelta );
#endif
            dwDelta = (dwDelta > dwDeltaDraft) ? dwDelta : dwDeltaDraft;
            while ( dwDelta > 0 ) {
                uint dwFrames = dwDelta > (uint)s_sample_rate ? (uint)s_sample_rate : dwDelta;
                fixed ( float* left = &left_ch_src[0] )
                fixed ( float* right = &right_ch_src[0] ) {
                    float*[] bufa = new float*[] { left, right };
                    fixed ( float** buf = &bufa[0] ) {
                        ProcessReplacing_( buf, (int)dwFrames );
                    }
                }
                /*for ( size_t j = 0; j < dwFrames; j++ ) {
                    WORD right_chx = (WORD)(32768 * out_buffer[0][j]);
                    WORD left_chx = (WORD)(32768 * out_buffer[1][j]);
                    wave_bufs[j] = MAKELONG( right_chx, left_chx );
                }*/
#if DEBUG
                for( int i = 0; i < left_ch_src.Length; i++ ){
                    if ( i >= dwFrames ){
                        break;
                    }
                    //Console.WriteLine( left_ch_src[i] );
                }
#endif

                //mmioWrite( hmmio, (char*)wave_bufs, dwFrames * sizeof( DWORD ) );
                //mmioAscend( hmmio, &mmckData, 0 );

                dwDelta -= dwFrames;
#if DEBUG
                total_processed += (int)dwFrames;
#endif
            }

            MainsChanged_( 0 );//            g_pAEffect->dispatcher( g_pAEffect, effMainsChanged, 0, 0, 0, 0 );

            //mmioAscend( hmmio, &mmckRiff, 0 );
            //mmioClose( hmmio, 0 );

            //free_events( lpEvents );
            //delete[] left_ch;
            //delete[] right_ch;
            //delete[] wave_bufs;
#if DEBUG
            Console.WriteLine( "total_processed=" + total_processed );
#endif

            return 1;
        }

        public static void AbortRendering() {
            try {
                AbortRendering_();
            } catch ( Exception ex ) {
#if DEBUG
                bocoree.debug.PushLog( "VSTiProxy+AbortRendering()" );
                bocoree.debug.PushLog( "    ex=" + ex.ToString() );
#endif
            }
        }

        private static bool LoadPlugin( int blockSize, int sampleRate ) {
            s_dll_path = VsqUtil.VstiDllPath;
#if DEBUG
            bocoree.debug.PushLog( "VSTiProxy+LoadPlugin(String, Int32, Int32)" );
            bocoree.debug.PushLog( "    m_dll_path=" + s_dll_path );
#endif
            if ( s_dll_path == "" ) {
                s_loaded = false;
                return false;
            }
            if ( s_error_counter > NUM_ERR_THRESHOLD ) {
                //throw new ApplicationException();
                return false;
            }
            int ret = 0;
            try {
                ret = LoadPluginW_( s_dll_path, blockSize, sampleRate );
                s_loaded = true;
                return true;
            } catch ( Exception ex ) {
#if DEBUG
                bocoree.debug.PushLog( ex.ToString() );
#endif
                s_error_counter++;
                //throw new ApplicationException();
            }
            if ( ret == 0 ) {
                s_error_counter++;
                //throw new ApplicationException();
            }
            return false;
        }

        private static bool FreePlugin() {
            if ( s_loaded ) {
                try {
                    int ret = FreePlugin_();
                    if ( ret == 0 ) {
                        return false;
                    } else {
                        return true;
                    }
                } catch ( Exception ex ) {
#if DEBUG
                    bocoree.debug.PushLog( "VSTiProxy+FreePlugin()" );
                    bocoree.debug.PushLog( "    ex=" + ex.ToString() );
#endif
                }
            }
            return false;
        }

        public unsafe static void SendEvent( byte* src, int* deltaFrames, int numEvents, VSTiProxyTrackType targetTrack ) {
            if ( s_loaded ) {
                int track = targetTrack == VSTiProxyTrackType.TempoTrack ? 0 : 1;
                try {
                    int ret = SendEvent_( src, deltaFrames, numEvents, track );
                    if ( ret == 0 ) {
                        //throw new ApplicationException();
                    }
                } catch ( Exception ex ) {
#if DEBUG
                    bocoree.debug.PushLog( "VSTiProxy+SendEvent(byte*, int*, int, VSTiProxyTrackType)" );
                    bocoree.debug.PushLog( "    ex=" + ex.ToString() );
#endif
                }
            }
        }

        public static bool RenderToWave( string name, uint eof_clock ) {
            try {
                int ret = RenderToWaveW_( name, eof_clock );
                if ( ret == 0 ) {
                    return false;
                }
                return true;
            } catch ( Exception ex ) {
#if DEBUG
                bocoree.debug.PushLog( "VSTiProxy+RenderToWave(string, uint)" );
                bocoree.debug.PushLog( "    ex=" + ex.ToString() );
#endif
            }
            return false;
        }

        public static void Finalize() {
            AbortRendering();
            if ( s_loaded ) {
                FreePlugin();
                s_loaded = false;
            }
        }

        public static unsafe bool RenderToWave( NrpnData[] nrpn, TempoTableEntry[] tempo, int clEos, string file ) {
#if DEBUG
            Console.WriteLine( "VSTiProxy.RenderToWave( NRPN[], TempoTableEntry[], int, string )" );
            bocoree.debug.PushLog( "VSTiPRoxy+RenderToWave(NrpnData[], TempoTableEntry[], int, string)" );
#endif
            if ( !s_loaded ) {
                s_loaded = LoadPlugin( 44100, 44100 );
            }
            if ( !s_loaded ) {
                return false;
            }
            int tempo_count = tempo.Length;
            byte[] masterEventsSrc = new byte[tempo_count * 3];
            int[] masterClocksSrc = new int[tempo_count];
            int count = -3;
            for ( int i = 0; i < tempo.Length; i++ ) {
                count+=3;
                masterClocksSrc[i] = tempo[i].Clock;
                byte b0 = (byte)(tempo[i].Tempo >> 16);
                uint u0 = (uint)(tempo[i].Tempo - (b0 << 16));
                byte b1 = (byte)(u0 >> 8);
                byte b2 = (byte)(u0 - (u0 << 8));
#if DEBUG
                Console.WriteLine( "    b0-b2=0x" + Convert.ToString( b0, 16 ) + ", 0x" + Convert.ToString( b1, 16 ) + ", 0x" + Convert.ToString( b2, 16 ) );
#endif
                masterEventsSrc[count] = b0;
                masterEventsSrc[count + 1] = b1;
                masterEventsSrc[count + 2] = b2;
            }
            fixed ( byte* masterEvents = &masterEventsSrc[0] )
            fixed ( int* masterClocks = &masterClocksSrc[0] ) {
                SendEvent( masterEvents, masterClocks, tempo_count, VSTiProxyTrackType.TempoTrack );
            }

            int numEvents = nrpn.Length + 1;
            byte[] bodyEventsSrc = new byte[numEvents * 3];
            int[] bodyClocksSrc = new int[numEvents];
            count = -3;
            int last_clock = 0;
#if DEBUG
            string test = Path.Combine( Path.GetDirectoryName( file ), Path.GetFileNameWithoutExtension( file ) + ".txt" );
            StreamWriter sw = new StreamWriter( test, false );
#endif
            for ( int i = 0; i < numEvents - 1; i++ ) {
                count += 3;
                bodyEventsSrc[count] = 0xb0;
                bodyEventsSrc[count + 1] = nrpn[i].Parameter;
                bodyEventsSrc[count + 2] = nrpn[i].Value;
                bodyClocksSrc[i] = nrpn[i].Clock;
                last_clock = nrpn[i].Clock;
#if DEBUG
                sw.WriteLine( nrpn[i].Clock + "\t0x" + Convert.ToString( nrpn[i].Parameter, 16 ) + "\t0x" + Convert.ToString( nrpn[i].Value, 16 ) );
#endif
            }
#if DEBUG
            sw.Close();
#endif
            count += 3;
            bodyEventsSrc[count] = 0xff;
            bodyEventsSrc[count + 1] = 0x2f;
            bodyEventsSrc[count + 2] = 0x00;
            bodyClocksSrc[numEvents - 1] = clEos;
            int index = tempo_count - 1;
            for ( int i = tempo_count - 1; i >= 0; i-- ) {
                if ( tempo[i].Clock < last_clock ) {
                    index = i;
                    break;
                }
            }
            int last_tempo = tempo[index].Tempo;

            fixed ( byte* bodyEvents = &bodyEventsSrc[0] )
            fixed ( int* bodyClocks = &bodyClocksSrc[0] ) {
                SendEvent( bodyEvents, bodyClocks, numEvents, VSTiProxyTrackType.MainTrack );
            }
            return RenderToWave( file, (uint)clEos );
        }

        public static unsafe void GenerateSinglePhone( int note, byte program_change, string file ) {
#if DEBUG
            Common.DebugWriteLine( "VSTiProxy+GenerateSinglePhone" );
            Common.DebugWriteLine( "    program_change=" + program_change );
#endif
            #region pre-set data
            NrpnData[] nrps = new NrpnData[]{
                new NrpnData( 0, 0x63, 0x60 ),
                new NrpnData( 0, 0x62, 0x00 ),
                new NrpnData( 0, 0x06, 0x00 ),
                new NrpnData( 0, 0x26, 0x00 ),
                new NrpnData( 0, 0x63, 0x60 ),
                new NrpnData( 0, 0x62, 0x01 ),
                new NrpnData( 0, 0x06, 0x00 ),
                new NrpnData( 0, 0x26, 0x00 ),
                new NrpnData( 0, 0x63, 0x60 ),
                new NrpnData( 0, 0x62, 0x02 ),
                new NrpnData( 0, 0x06, 0x00 ),
                new NrpnData( 0, 0x63, 0x53 ),
                new NrpnData( 0, 0x62, 0x02 ),
                new NrpnData( 0, 0x06, program_change ),
                new NrpnData( 286, 0x63, 0x50 ),
                new NrpnData( 286, 0x62, 0x01 ),
                new NrpnData( 286, 0x06, 0x02 ),
                new NrpnData( 286, 0x26, 0x47 ),
                new NrpnData( 286, 0x62, 0x02 ),
                new NrpnData( 286, 0x06, (byte)note ),
                new NrpnData( 286, 0x62, 0x03 ),
                new NrpnData( 286, 0x06, 0x40 ),
                new NrpnData( 286, 0x62, 0x04 ),
                new NrpnData( 286, 0x06, 0x03 ),
                new NrpnData( 286, 0x26, 0x74 ),
                new NrpnData( 286, 0x62, 0x05 ),
                new NrpnData( 286, 0x06, 0x03 ),
                new NrpnData( 286, 0x62, 0x12 ),
                new NrpnData( 286, 0x06, 0x01 ),
                new NrpnData( 286, 0x62, 0x13 ),
                new NrpnData( 286, 0x06, 0x61 ),
                new NrpnData( 286, 0x26, 0x00 ),
                new NrpnData( 286, 0x62, 0x4F ),
                new NrpnData( 286, 0x06, 0x7F ),
                new NrpnData( 286, 0x62, 0x50 ),
                new NrpnData( 286, 0x06, 0x04 ),
                new NrpnData( 286, 0x62, 0x51 ),
                new NrpnData( 286, 0x06, 0x08 ),
                new NrpnData( 286, 0x62, 0x52 ),
                new NrpnData( 286, 0x06, 0x14 ),
                new NrpnData( 286, 0x62, 0x53 ),
                new NrpnData( 286, 0x06, 0x1C ),
                new NrpnData( 286, 0x62, 0x54 ),
                new NrpnData( 286, 0x06, 0x18 ),
                new NrpnData( 286, 0x62, 0x55 ),
                new NrpnData( 286, 0x06, 0x0A ),
                new NrpnData( 286, 0x62, 0x56 ),
                new NrpnData( 286, 0x06, 0x0C ),
                new NrpnData( 286, 0x62, 0x57 ),
                new NrpnData( 286, 0x06, 0x0C ),
                new NrpnData( 286, 0x62, 0x58 ),
                new NrpnData( 286, 0x06, 0x00 ),
                new NrpnData( 286, 0x62, 0x59 ),
                new NrpnData( 286, 0x06, 0x32 ),
                new NrpnData( 286, 0x62, 0x5A ),
                new NrpnData( 286, 0x06, 0x32 ),
                new NrpnData( 286, 0x62, 0x7F ),
                new NrpnData( 286, 0x06, 0x7F ),
            };
            #endregion
            if ( !s_loaded ) {
                try {
                    s_loaded = LoadPlugin( 44100, 44100 );
                } catch(Exception ex) {
#if DEBUG
                    bocoree.debug.PushLog( ex.ToString() );
#endif
                    return;
                }
                if ( !s_loaded ) {
                    return;
                }
            }
            byte[] masterEventsSrc = new byte[3];
            masterEventsSrc[0] = 0x07;
            masterEventsSrc[1] = 0xa1;
            masterEventsSrc[2] = 0x20;
            int[] masterClocksSrc = new int[1];
            masterClocksSrc[0] = 0;
            fixed( byte* masterEvents = &masterEventsSrc[0] )
            fixed ( int* masterClocks = &masterClocksSrc[0] ) {
                SendEvent( masterEvents, masterClocks, 1, VSTiProxyTrackType.TempoTrack );
            }

            int numEvents = nrps.Length + 1;
            byte[] bodyEventsSrc = new byte[numEvents * 3];
            int[] bodyClocksSrc = new int[numEvents];
            int count = -3;
            //nrps[17].Value = (byte)note;
            for( int i = 0; i < numEvents - 1; i++ ){
                count += 3;
                bodyEventsSrc[count] = 0xb0;
                bodyEventsSrc[count + 1] = nrps[i].Parameter;
                bodyEventsSrc[count + 2] = nrps[i].Value;
                bodyClocksSrc[i] = nrps[i].Clock;
            }
            count+=3;
            bodyEventsSrc[count] = 0xff;
            bodyEventsSrc[count + 1] = 0x2f;
            bodyEventsSrc[count + 2] = 0x00;
            bodyClocksSrc[numEvents - 1] = 2280;
            fixed( byte* bodyEvents = &bodyEventsSrc[0] )
            fixed ( int* bodyClocks = &bodyClocksSrc[0] ) {
                SendEvent( bodyEvents, bodyClocks, numEvents, VSTiProxyTrackType.MainTrack );
            }

            RenderToWave( file, 2280 );
        }
    }

}
