﻿/*
 * Version.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;

namespace Boare.Cadencii {

    public struct Version : IComparable<Version> {
        private int m_measure;
        private int m_minor;
        private int m_build;
        private string m_comment;
        private bool m_is_build_empty;

        public int Measure {
            get {
                return m_measure;
            }
            set {
                m_measure = value;
            }
        }

        public int Minor {
            get {
                return m_minor;
            }
            set {
                m_minor = value;
            }
        }

        public int Build {
            get {
                return m_build;
            }
            set {
                m_build = value;
            }
        }

        public string Comment {
            get {
                return m_comment;
            }
            set {
                m_comment = value;
            }
        }

        public Version( int measure, int minor, int build ) {
            m_measure = measure;
            m_minor = minor;
            m_build = build;
            m_comment = "";
            m_is_build_empty = false;
        }

        public Version( int measure, int minor, string comment ) {
            m_measure = measure;
            m_minor = minor;
            m_build = 0;
            m_comment = comment;
            m_is_build_empty = true;
        }

        public int CompareTo( Version item ) {
            if ( m_measure != item.m_measure ) {
                return m_measure - item.m_measure;
            }
            if ( m_minor != item.m_minor ) {
                return m_minor - item.m_minor;
            }
            if ( m_build != item.m_build ) {
                return m_build - item.m_build;
            }
            if ( m_comment == null ) {
                m_comment = "";
            }
            if ( item.m_comment == null ) {
                item.m_comment = "";
            }
            if ( m_comment == "" && item.m_comment != "" ) {
                return 1;
            } else if ( m_comment != "" && item.m_comment == "" ) {
                return -1;
            } else if ( m_comment != "" && item.m_comment != "" ) {
                return m_comment.CompareTo( item.m_comment );
            }
            return 0;
        }

        public override string ToString() {
            if ( m_is_build_empty ) {
                return m_measure + "." + m_minor + m_comment;
            } else {
                return m_measure + "." + m_minor + "." + m_build;
            }
        }
    }

}
