﻿/*
 * VsqFileEx.cs
 * Copyright (c) 2008-2009 kbinani
 *
 * This file is part of Boare.Cadencii.
 *
 * Boare.Cadencii is free software; you can redistribute it and/or
 * modify it under the terms of the BSD License.
 *
 * Boare.Cadencii is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
using System;
using System.Collections.Generic;
using System.Drawing;

using Boare.Lib.Vsq;

namespace Boare.Cadencii {

    public class BezierCurves {
        public BezierChain[] Dynamics;
        public BezierChain[] Brethiness;
        public BezierChain[] Brightness;
        public BezierChain[] Clearness;
        public BezierChain[] Opening;
        public BezierChain[] GenderFactor;
        public BezierChain[] PortamentoTiming;
        public BezierChain[] PitchBend;
        public BezierChain[] PitchBendSensitivity;

        public BezierCurves() {
            Dynamics = new BezierChain[0];
            Brethiness = new BezierChain[0];
            Brightness = new BezierChain[0];
            Clearness = new BezierChain[0];
            Opening = new BezierChain[0];
            GenderFactor = new BezierChain[0];
            PortamentoTiming = new BezierChain[0];
            PitchBend = new BezierChain[0];
            PitchBendSensitivity = new BezierChain[0];
        }

        public BezierCurves( AttachedCurve item ) {
            VsqCurveType work = VsqCurveType.DYN;
            Dynamics = new BezierChain[item[work].Count];
            int i = -1;
            foreach( int key in item[work].Keys ){
                i++;
                Dynamics[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.BRE;
            Brethiness = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                Brethiness[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.BRI;
            Brightness = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                Brightness[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.CLE;
            Clearness = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                Clearness[i] = (BezierChain)item[work][key].Clone();
            }
            work = VsqCurveType.OPE;
            Opening = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                Opening[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.GEN;
            GenderFactor = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                GenderFactor[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.POR;
            PortamentoTiming = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                PortamentoTiming[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.PIT;
            PitchBend = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                PitchBend[i] = (BezierChain)item[work][key].Clone();
            }

            work = VsqCurveType.PBS;
            PitchBendSensitivity = new BezierChain[item[work].Count];
            i = -1;
            foreach ( int key in item[work].Keys ) {
                i++;
                PitchBendSensitivity[i] = (BezierChain)item[work][key].Clone();
            }
        }
    }

    [Serializable]
    public class VsqFileEx : VsqFile, ICloneable {
        private List<AttachedCurve> m_attached_curve;

        public List<AttachedCurve> AttachedCurves {
            get {
                return m_attached_curve;
            }
            set {
                m_attached_curve = value;
            }
        }

        public new object Clone(){
            VsqFileEx ret = new VsqFileEx( 0, 1, 4, 4, 5000 );
            ret.m_tracks = new List<VsqTrack>();
            for ( int i = 0; i < m_tracks.Count; i++ ) {
                ret.m_tracks.Add( (VsqTrack)m_tracks[i].Clone() );
            }
            ret.m_tempo_table = new List<TempoTableEntry>();
            for ( int i = 0; i < m_tempo_table.Count; i++ ) {
                ret.m_tempo_table.Add( (TempoTableEntry)m_tempo_table[i].Clone() );
            }
            ret.m_timesig_table = new List<TimeSigTableEntry>();
            for ( int i = 0; i < m_timesig_table.Count; i++ ) {
                ret.m_timesig_table.Add( (TimeSigTableEntry)m_timesig_table[i].Clone() );
            }
            ret.m_tpq = m_tpq;
            ret.m_total_clocks = m_total_clocks;
            ret.m_base_tempo = m_base_tempo;
            ret.m_master = (VsqMaster)m_master.Clone();
            ret.m_mixer = (VsqMixer)m_mixer.Clone();
            ret.m_premeasure_clocks = m_premeasure_clocks;
            ret.m_attached_curve.Clear();
            for ( int i = 0; i < m_attached_curve.Count; i++ ) {
                ret.m_attached_curve.Add( (AttachedCurve)m_attached_curve[i].Clone() );
            }
            return ret;
        }

        public static CadenciiCommand GCommandAddBezierChain( int track, VsqCurveType curve_type, int chain_id, int clock_resolution, BezierChain chain ) {
            CadenciiCommand ret = new CadenciiCommand();
            ret.Type = CadenciiCommandType.AddBezierChain;
            ret.Args = new object[5];
            ret.Args[0] = track;
            ret.Args[1] = curve_type;
            ret.Args[2] = (BezierChain)chain.Clone();
            ret.Args[3] = clock_resolution;
            ret.Args[4] = chain_id;
            return ret;
        }

        public static CadenciiCommand GCommandDeleteBezierChain( int track, VsqCurveType curve_type, int chain_id, int clock_resolution ) {
            CadenciiCommand ret = new CadenciiCommand();
            ret.Type = CadenciiCommandType.DeleteBezierChain;
            ret.Args = new object[4];
            ret.Args[0] = track;
            ret.Args[1] = curve_type;
            ret.Args[2] = chain_id;
            ret.Args[3] = clock_resolution;
            return ret;
        }

        public static CadenciiCommand GCommandReplaceBezierChain( int track, VsqCurveType curve_type, int chain_id, BezierChain chain, int clock_resolution ) {
            CadenciiCommand ret = new CadenciiCommand();
            ret.Type = CadenciiCommandType.ReplaceBezierChain;
            ret.Args = new object[5];
            ret.Args[0] = track;
            ret.Args[1] = curve_type;
            ret.Args[2] = chain_id;
            ret.Args[3] = chain;
            ret.Args[4] = clock_resolution;
            return ret;
        }

        public CadenciiCommand Execute( CadenciiCommand command ) {
#if DEBUG
            Common.DebugWriteLine( "VsqFileEx.Execute" );
#endif
            CadenciiCommand ret = null;
            if ( command.Type == CadenciiCommandType.VsqCommand ) {
                ret = new CadenciiCommand();
                ret.Type = CadenciiCommandType.VsqCommand;
                ret.VsqCommand = base.Execute( command.VsqCommand );
            } else {
                if ( command.Type == CadenciiCommandType.AddBezierChain ) {
                    #region AddBezierChain
#if DEBUG
                    Common.DebugWriteLine( "    AddBezierChain" );
#endif
                    int track = (int)command.Args[0];
                    VsqCurveType curve_type = (VsqCurveType)command.Args[1];
                    BezierChain chain = (BezierChain)command.Args[2];
                    int clock_resolution = (int)command.Args[3];
                    int added_id = (int)command.Args[4];
                    AttachedCurves[track - 1].AddBezierChain( curve_type, chain, added_id );
                    ret = GCommandDeleteBezierChain( track, curve_type, added_id, clock_resolution );
                    if ( chain.Count > 1 ) {
                        int min = chain.points[0].Base.X;
                        int max = min;
                        for ( int i = 1; i < chain.points.Count; i++ ) {
                            min = Math.Min( min, chain.points[i].Base.X );
                            max = Math.Max( max, chain.points[i].Base.X );
                        }
                        int max_value = VsqCurveTypeUtil.GetMaxValue( curve_type );
                        int min_value = VsqCurveTypeUtil.GetMinValue( curve_type );
                        if ( min < max ) {
                            List<BPPair> edit = new List<BPPair>();
                            int last_value = int.MaxValue;
                            for ( int clock = min; clock < max; clock += clock_resolution ) {
                                int value = (int)chain.GetValue( (float)clock );
                                if ( value < min_value ) {
                                    value = min_value;
                                } else if ( max_value < value ) {
                                    value = max_value;
                                }
                                if ( value != last_value ) {
                                    edit.Add( new BPPair( clock, value ) );
                                    last_value = value;
                                }
                            }
                            int value2 = Tracks[track][curve_type][max];
#if DEBUG
                            Common.DebugWriteLine( "        " + max + "," + value2 );
#endif
                            edit.Add( new BPPair( max, value2 ) );
                            command.VsqCommand = VsqCommand.GCommandTrackEditCurve( track, curve_type, edit );
                        }
                    }
                    #endregion
                } else if ( command.Type == CadenciiCommandType.DeleteBezierChain ) {
                    #region DeleteBezierChain
                    int track = (int)command.Args[0];
                    VsqCurveType curve_type = (VsqCurveType)command.Args[1];
                    int chain_id = (int)command.Args[2];
                    int clock_resolution = (int)command.Args[3];
                    BezierChain chain = (BezierChain)AttachedCurves[track - 1][curve_type][chain_id].Clone();
                    AttachedCurves[track - 1][curve_type].Remove( chain_id );
                    ret = GCommandAddBezierChain( track, curve_type, chain_id, clock_resolution, chain );
                    if ( command.VsqCommand != null && ret != null ) {
                        ret.VsqCommand = base.Execute( command.VsqCommand );
                    }
                    #endregion
                } else if ( command.Type == CadenciiCommandType.ReplaceBezierChain ) {
                    #region ReplaceBezierChain
                    int track = (int)command.Args[0];
                    VsqCurveType curve_type = (VsqCurveType)command.Args[1];
                    int chain_id = (int)command.Args[2];
                    BezierChain chain = (BezierChain)command.Args[3];
                    int clock_resolution = (int)command.Args[4];
                    BezierChain target = (BezierChain)AttachedCurves[track - 1][curve_type][chain_id].Clone();
                    AttachedCurves[track - 1][curve_type][chain_id] = chain;
                    ret = GCommandReplaceBezierChain( track, curve_type, chain_id, target, clock_resolution );
                    if ( chain.Count == 1 ) {
                        int ex_min = chain.points[0].Base.X;
                        int ex_max = ex_min;
                        if ( target.points.Count > 1 ) {
                            for ( int i = 1; i < target.points.Count; i++ ) {
                                ex_min = Math.Min( ex_min, target.points[i].Base.X );
                                ex_max = Math.Max( ex_max, target.points[i].Base.X );
                            }
                            if ( ex_min < ex_max ) {
                                int default_value = VsqCurveTypeUtil.GetDefaultValue( curve_type );
                                List<BPPair> edit = new List<BPPair>();
                                edit.Add( new BPPair( ex_min, default_value ) );
                                edit.Add( new BPPair( ex_max, default_value ) );
                                command.VsqCommand = VsqCommand.GCommandTrackEditCurve( track, curve_type, edit );
                            }
                        }
                    } else if ( chain.Count > 1 ) {
                        int min = chain.points[0].Base.X;
                        int max = min;
                        for ( int i = 1; i < chain.points.Count; i++ ) {
                            min = Math.Min( min, chain.points[i].Base.X );
                            max = Math.Max( max, chain.points[i].Base.X );
                        }
                        int ex_min = min;
                        int ex_max = max;
                        if ( target.points.Count > 0 ) {
                            ex_min = target.points[0].Base.X;
                            ex_max = ex_min;
                            for ( int i = 1; i < target.points.Count; i++ ) {
                                ex_min = Math.Min( ex_min, target.points[i].Base.X );
                                ex_max = Math.Max( ex_max, target.points[i].Base.X );
                            }
                        }
                        int max_value = VsqCurveTypeUtil.GetMaxValue( curve_type );
                        int min_value = VsqCurveTypeUtil.GetMinValue( curve_type );
                        int default_value = VsqCurveTypeUtil.GetDefaultValue( curve_type );
                        List<BPPair> edit = new List<BPPair>();
                        if ( ex_min < min ) {
                            edit.Add( new BPPair( ex_min, default_value ) );
                            /*if ( !Tracks[track][curve_type].ContainsKey( ex_min ) ) {
                                edit.Add( new BPPair( ex_min, Tracks[track][curve_type][ex_min] ) );
                            }
                            foreach ( int key_clock in Tracks[track][curve_type].GetKeyClockEnumerator() ) {
                                if ( ex_min <= key_clock && key_clock < min ) {
                                    edit.Add( new BPPair( key_clock, Tracks[track][curve_type][key_clock] ) );
                                } else if ( min <= key_clock ) {
                                    break;
                                }
                            }*/
                        }
                        if ( min < max ) {
                            int last_value = int.MaxValue;
                            for ( int clock = min; clock < max; clock += clock_resolution ) {
                                int value = (int)chain.GetValue( (float)clock );
                                if ( value < min_value ) {
                                    value = min_value;
                                } else if ( max_value < value ) {
                                    value = max_value;
                                }
                                if ( last_value != value ) {
                                    edit.Add( new BPPair( clock, value ) );
                                    last_value = value;
                                }
                            }
                            int value2 = 0;
                            if ( max < ex_max ) {
                                value2 = (int)chain.GetValue( max );
                                if ( value2 < min_value ) {
                                    value2 = min_value;
                                } else if ( max_value < value2 ) {
                                    value2 = max_value;
                                }
                            } else {
                                value2 = Tracks[track][curve_type][max];
                            }
                            edit.Add( new BPPair( max, value2 ) );
                        }
                        if ( max < ex_max ) {
                            /*foreach ( int key_clock in Tracks[track][curve_type].GetKeyClockEnumerator() ) {
                                if ( max < key_clock && key_clock < ex_max ) {
                                    edit.Add( new BPPair( key_clock, Tracks[track][curve_type][key_clock] ) );
                                } else if ( ex_max <= key_clock ) {
                                    break;
                                }
                            }
                            edit.Add( new BPPair( ex_max, Tracks[track][curve_type][ex_max] ) );*/
                            if ( max + 1 < ex_max ) {
                                edit.Add( new BPPair( max + 1, default_value ) );
                            }
                            edit.Add( new BPPair( ex_max, default_value ) );
                        }
                        if ( edit.Count > 0 ) {
                            command.VsqCommand = VsqCommand.GCommandTrackEditCurve( track, curve_type, edit );
                        }
                    }
                    #endregion
                }
                if ( command.VsqCommand != null && ret != null ) {
                    ret.VsqCommand = base.Execute( command.VsqCommand );
                }
            }
            return ret;
        }

        public VsqFileEx( int default_singer_program_change, int pre_measure, int numerator, int denominator, int tempo ) :
            base( default_singer_program_change, pre_measure, numerator, denominator, tempo ) {
            m_attached_curve = new List<AttachedCurve>();
            for ( int i = 1; i < Tracks.Count; i++ ) {
                m_attached_curve.Add( new AttachedCurve() );
            }
        }

        public VsqFileEx( string _fpath ) :
            base( _fpath ){
            m_attached_curve = new List<AttachedCurve>();
            for ( int i = 1; i < Tracks.Count; i++ ) {
                m_attached_curve.Add( new AttachedCurve() );
            }
        }
    }

}
